#include "nbc_def.h"
#include "nbc_ext.h"

void allocate_buffer()
{
  buf_base = (char *)malloc(sizeof(char)*MAX_OUT_LEN);
  buf_sys = (char *)malloc(sizeof(char)*MAX_OUT_LEN);
  buf_delim = (char *)malloc(sizeof(char)*MAX_OUT_LEN);
  buf_out = (char *)malloc(sizeof(char)*MAX_OUT_LEN);
  if (buf_base == NULL  || buf_sys == NULL ||
      buf_delim == NULL || buf_out == NULL)
    die("Could't allocate memory for buffers\n");
}

void release_buffer()
{
  free(buf_base);  buf_base = NULL;
  free(buf_sys);   buf_base = NULL;
  free(buf_delim); buf_base = NULL;
  free(buf_out);   buf_base = NULL;
}

void allocate_log_likelihood(void)
{
  log_likelihood = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*MAX_MAX_ITERATE);
  lambda = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*MAX_MAX_ITERATE);
  if (log_likelihood == NULL || lambda == NULL)
    die("Couldn't allocate memory for log_likelihood or lambda\n");

  if (mode_smooth) {
    log_prior = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*MAX_MAX_ITERATE);
    if (log_prior == NULL)
      die("Couldn't allocate memory for log_likelihood or lambda\n");
  }
}

void allocate_log_likelihood_best(void)
{
  int n;
  if (num_init > 1) {
    n = (max_init_iterate > 0) ? max_init_iterate : MAX_MAX_ITERATE;
  } else {
    n = (max_iterate > 0) ? max_iterate : MAX_MAX_ITERATE;
  }

  log_likelihood_best = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*n);
  lambda_best = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*n);
  if (log_likelihood_best == NULL || lambda_best == NULL)
    die("Couldn't allocate memory for {log_likelihood,lambda}_best\n");

  if (mode_smooth) {
    log_prior_best = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*n);
    if (log_prior_best == NULL)
      die("Couldn't allocate memory for log_prior_best\n");
  }
}

void allocate_prob(void)
{
  int k;

  prob_comp = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*num_class);
  prob_incomp = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_entry);
  prob_cond = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*num_class);
  if (prob_comp == NULL || prob_incomp == NULL || prob_cond == NULL)
    die("Couldn't allocate memory for prob_{comp,incomp,cond}\n");

  for (k=0; k<num_class; k++) {
    prob_comp[k] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_entry);
    prob_cond[k] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_entry);
    if (prob_comp[k] == NULL || prob_cond[k] == NULL)
      die("Couldn't allocate memory for prob_{comp,cond}[k]\n");
  }
}

void allocate_ic(void)
{
  int num_search;
  
  num_search = (kmax-kmin)/kstep + 1;

  ic = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_search);
  bic = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_search);
  autoclass = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_search);

  if (ic == NULL || bic == NULL || autoclass == NULL)
    die("Couldn't allocate memory for ic, bic or autoclass\n");
}

void allocate_seed(void)
{
  int u,num_search;

  num_search = (mode_search_num_class) ? ((kmax-kmin)/kstep + 1): 1;

  log_seed = (unsigned int **)malloc(sizeof(unsigned int *)*num_search);
  log_num_iterate = (int **)malloc(sizeof(int *)*num_search);
  log_final_lambda = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*num_search);
  log_r_best = (int *)malloc(sizeof(int)*num_search);
  if (log_seed == NULL || log_num_iterate == NULL ||
      log_final_lambda == NULL || log_r_best == NULL)
    die("Couldn't allocate memory for log_{seed,num_iterate,final_lambda,r_best}\n");

  for (u = 0; u < num_search; u++) {
    log_seed[u] = (unsigned int *)malloc(sizeof(unsigned int)*num_init);
    log_num_iterate[u] = (int *)malloc(sizeof(int)*num_init);
    log_final_lambda[u] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_init);
    if (log_seed == NULL || log_num_iterate == NULL || log_final_lambda == NULL)
      die("Couldn't allocate memory for log_{seed,num_iterate,final_lambda}[u]\n");
  }
}

void release_prob(void)
{
  int k;
  for (k=0; k<num_class; k++) {
    free(prob_comp[k]); prob_comp[k] = NULL;
    free(prob_cond[k]); prob_cond[k] = NULL;
  }
  free(prob_comp);   prob_comp = NULL;
  free(prob_incomp); prob_incomp = NULL;
  free(prob_cond);   prob_cond = NULL;
}

void release_log_likelihood(void)
{
  free(log_likelihood); log_likelihood = NULL;
  free(lambda); lambda = NULL;
  if (mode_smooth) {
    free(log_prior); log_prior = NULL;
  }
}

void release_log_likelihood_best(void)
{
  free(log_likelihood_best); log_likelihood_best = NULL;
  free(lambda_best); lambda_best = NULL;
  if (mode_smooth) {
    free(log_prior_best); log_prior_best = NULL;
  }
}

void release_ic(void)
{
  free(ic);  ic = NULL;
  free(bic); bic = NULL;
  free(autoclass); autoclass = NULL;
}

void release_seed(void)
{
  int u, num_search;

  num_search = (mode_search_num_class) ? ((kmax-kmin)/kstep + 1): 1;

  for (u=0; u < num_search; u++) {
    free(log_seed[u]);         log_seed[u] = NULL;
    free(log_num_iterate[u]);  log_num_iterate[u] = NULL;
    free(log_final_lambda[u]); log_final_lambda[u] = NULL;
  }
  free(log_seed);         log_seed = NULL;
  free(log_num_iterate);  log_num_iterate = NULL;
  free(log_final_lambda); log_final_lambda = NULL;
}

void allocate_param(void)
{
  int k,v;

  param_c = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
  param_w = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class);
  if (param_c == NULL || param_w == NULL)
    die("Couldn't allocate memory for param_c or param_w\n");

  for (k=0; k<num_class; k++) {
    param_w[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (param_w[k] == NULL)
      die("Couldn't allocate memory for param_w[k]\n");
    for (v=0; v<tuple_size; v++) {
      param_w[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      if (param_w[k][v] == NULL)
        die("Couldn't allocate memory for param_w[k][v]\n");
    }
  }
}

void allocate_param_best(void)
{
  int k,v;

  param_c_best = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
  param_w_best = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class);
  if (param_c_best == NULL || param_w_best == NULL)
    die("Couldn't allocate memory for param_c_best or param_w_best\n");

  for (k=0; k<num_class; k++) {
    param_w_best[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (param_w_best[k] == NULL)
      die("Couldn't allocate memory for param_w_best[k]\n");
    for (v=0; v<tuple_size; v++) {
      param_w_best[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      if (param_w_best[k][v] == NULL)
        die("Couldn't allocate memory for param_w_best[k][v]\n");
    }
  }
}

void allocate_hparam(void)
{
  int k,v;

  hparam_c = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
  hparam_w = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class);
  if (hparam_c == NULL || hparam_w == NULL)
    die("Couldn't allocate memory for hparam_{c,w}\n");

  for (k=0; k<num_class; k++) {
    hparam_w[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (hparam_w[k] == NULL)
      die("Couldn't allocate memory for hparam_w[k]\n");
    for (v=0; v<tuple_size; v++) {
      hparam_w[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      if (hparam_w[k][v] == NULL)
        die("Couldn't allocate memory for hparam_w[k][v]\n");
    }
  }
}

void release_hparam(void)
{
  int k,v;

  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      free(hparam_w[k][v]); hparam_w[k][v] = NULL;
    }
    free(hparam_w[k]); hparam_w[k] = NULL;
  }
  free(hparam_w); hparam_w = NULL;
  free(hparam_c); hparam_c = NULL;
}

void allocate_expect(void)
{
  int k,v;

  expect_c = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
  tmp_expect_c = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
  expect_w = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class);
  if (expect_c == NULL || tmp_expect_c == NULL || expect_w == NULL)
    die("Couldn't allocate memory for {,tmp_}expect_c or expect_w\n");

  for (k=0; k<num_class; k++) {
    expect_w[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (expect_w[k] == NULL)
      die("Couldn't allocate memory for expect_w[k]\n");
    for (v=0; v<tuple_size; v++) {
      expect_w[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      if (expect_w[k][v] == NULL)
        die("Couldn't allocate memory for expect_w[k][v]\n");
    }
  }
}

void allocate_prob_mem(void)
{
  int k,v;

  prob_case_mem = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*num_class);
  prob_att_mem = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class);
  if (prob_case_mem == NULL || prob_att_mem == NULL)
    die("Couldn't allocate memory for prob_{case,att}_mem\n");

  for (k=0; k<num_class; k++) {
    prob_case_mem[k] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_entry);
    prob_att_mem[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (prob_case_mem[k] == NULL || prob_att_mem[k] == NULL)
      die("Couldn't allocate memory for prob_{case,att}_mem[k]\n");
    for (v=0; v<tuple_size; v++) {
      prob_att_mem[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      if (prob_att_mem[k][v] == NULL)
        die("Couldn't allocate memory for prob_att_mem[k][v]\n");
    }
  }

  prob_w = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
  if (prob_w == NULL)
    die("Couldn't allocate memory for prob_w\n");
  for (v=0; v<tuple_size; v++) {
    prob_w[v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
    if (prob_w[v] == NULL)
      die("Couldn't allocate memory for prob_w[v]\n");
  }
}

void release_param(void)
{
  int k,v;
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      free(param_w[k][v]); param_w[k][v] = NULL;
    }
    free(param_w[k]); param_w[k] = NULL;
  }
  free(param_w); param_w = NULL;
  free(param_c); param_c = NULL;
}

void release_param_best(void)
{
  int k,v;
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      free(param_w_best[k][v]); param_w_best[k][v] = NULL;
    }
    free(param_w_best[k]); param_w_best[k] = NULL;
  }
  free(param_w_best); param_w_best = NULL;
  free(param_c_best); param_c_best = NULL;
}

void release_param_opt(void)
{
  int k,v;

  for (k=0; k<num_class_opt; k++) {
    for (v=0; v<tuple_size; v++) {
      free(param_w_opt[k][v]); param_w_opt[k][v] = NULL;
    }
    free(param_w_opt[k]); param_w_opt[k] = NULL;
  }
  free(param_w_opt); param_w_opt = NULL;
  free(param_c_opt); param_c_opt = NULL;
}

void release_expect(void)
{
  int k,v;
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      free(expect_w[k][v]); expect_w[k][v] = NULL;
    }
    free(expect_w[k]); expect_w[k] = NULL;
  }
  free(expect_w); expect_w = NULL;
  free(expect_c); expect_c = NULL;
}

void release_prob_mem(void)
{
  int k,v;
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      free(prob_att_mem[k][v]); prob_att_mem[k][v] = NULL;
    }
    free(prob_att_mem[k]);  prob_att_mem[k] = NULL;
    free(prob_case_mem[k]); prob_case_mem[k] = NULL;
  }
  free(prob_att_mem);  prob_att_mem = NULL;
  free(prob_case_mem); prob_case_mem = NULL;

  for (v=0; v<tuple_size; v++) {
    free(prob_w[v]); prob_w[v] = NULL;
  }
  free(prob_w); prob_w = NULL;
}

void allocate_read_param(int max_num_class, int *max_num_word)
{
  int k,v;

  read_param_c = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*max_num_class);
  
  if (read_param_c == NULL)
    die("Couldn't allocate memory for read_param_c\n");

  read_param_w = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*max_num_class);
  if (read_param_w == NULL)
    die("Couldn't allocate memory for read_param_w\n");
  
  for (k=0; k<max_num_class; k++) {
    read_param_w[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (read_param_w[k] == NULL)
      die("Couldn't allocate memory for read_param_w[k]\n");
    for (v=0; v<tuple_size; v++) {
      read_param_w[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*max_num_word[v]);
      if (read_param_w[k][v] == NULL)
	die("Couldn't allocate memory for read_param_w[k][v]\n");
    }
  }
}

void release_read_param(int max_num_class)
{
  int k,v;

  free(read_param_c); read_param_c = NULL;

  for (k=0; k<max_num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      free(read_param_w[k][v]); read_param_w[k][v] = NULL;
    }
    free(read_param_w[k]); read_param_w[k] = NULL;
  }

  free(read_param_w); read_param_w = NULL;
}

void expand_read_param_c(int old_max_num_class, int new_max_num_class)
{
  int i;
  FLOAT_TYPE *new_read_param_c;

  new_read_param_c = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*new_max_num_class);
  if (new_read_param_c == NULL)
    die("Couldn't allocate memory for new_read_param_c\n");

  for (i=0; i<old_max_num_class; i++) {
    new_read_param_c[i] = read_param_c[i];
  }

  free(read_param_c);
  read_param_c = new_read_param_c;
}

void expand_read_param_w(int old_max_num_class, int new_max_num_class,
			 int *old_max_num_word, int *new_max_num_word)
{
  int k,i,v;
  FLOAT_TYPE ***new_read_param_w;

  new_read_param_w = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*new_max_num_class);
  if (new_read_param_w == NULL)
    die("Couldn't allocate memory for new_read_param_w\n");

  for (k=0; k<new_max_num_class; k++) {
    new_read_param_w[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (new_read_param_w[k] == NULL)
      die("Couldn't allocate memory for new_read_param_w[k]\n");
    
    for (v=0; v<tuple_size; v++) {
      new_read_param_w[k][v] =
	(FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*new_max_num_word[v]);
      if (new_read_param_w[k][v] == NULL)
	die("Couldn't allocate memory for new_read_param_w[k][v]\n");
    }
  }

  for (k=0; k<old_max_num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<old_max_num_word[v]; i++) {
	new_read_param_w[k][v][i] = read_param_w[k][v][i];
      }
    }
  }

  for (k=0; k<old_max_num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      free(read_param_w[k][v]); read_param_w[k][v] = NULL;
    }
    free(read_param_w[k]); read_param_w[k] = NULL;
  }
  free(read_param_w);

  read_param_w = new_read_param_w;
}

void allocate_cluster(void)
{
  int k,v;

  att_cluster_pred = (int **)malloc(sizeof(int *)*tuple_size);
  if (att_cluster_pred == NULL)
    die("Couldn't allocate memory for att_cluster_pred\n");
  for (v=0; v<tuple_size; v++) {
    att_cluster_pred[v] = (int *)malloc(sizeof(int)*num_word[v]);
    if (att_cluster_pred[v] == NULL)
      die("Couldn't allocate memory for att_cluster_pred[v]\n");
  }

  att_cluster_freq = (int **)malloc(sizeof(int *)*tuple_size);
  if (att_cluster_freq == NULL)
    die("Couldn't allocate memory for att_cluster_freq\n");
  for (v=0; v<tuple_size; v++) {
    att_cluster_freq[v] = (int *)malloc(sizeof(int)*num_class);
    if (att_cluster_freq[v] == NULL)
      die("Couldn't allocate memory for att_cluster_freq[v]\n");
  }

  att_cluster = (int ***)malloc(sizeof(int **)*tuple_size);
  if (att_cluster == NULL)
    die("Couldn't allocate memory for att_cluster\n");

  for (v=0; v<tuple_size; v++) {
    att_cluster[v] = (int **)malloc(sizeof(int *)*num_class);
    if (att_cluster[v] == NULL)
      die("Couldn't allocate memory for att_cluster[v]\n");

    for (k=0; k<num_class; k++) {
      att_cluster[v][k] = (int *)malloc(sizeof(int)*num_word[v]);
      if (att_cluster[v][k] == NULL)
	die("Couldn't allocate memory for att_cluster[v][k]\n");
    }
  }

  case_cluster_pred = (int *)malloc(sizeof(int)*num_entry);  /* redundant, but safe */
  if (case_cluster_pred == NULL)
    die("Couldn't allocate memory for case_cluster_pred\n");

  case_cluster_freq = (int *)malloc(sizeof(int)*num_class);
  if (case_cluster_freq == NULL)
    die("Couldn't allocate memory for case_cluster_freq\n");

  case_cluster = (int **)malloc(sizeof(int *)*num_class);
  if (case_cluster == NULL)
    die("Couldn't allocate memory for case_cluster[v]\n");
  
  for (k=0; k<num_class; k++) {
    case_cluster[k] = (int *)malloc(sizeof(int)*num_entry);  /* redundant, but safe */
    if (case_cluster[k] == NULL)
      die("Couldn't allocate memory for case_cluster[k]\n");
  }

  if (cl_type == CL_WODD) {
    wodd_att = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class);
    wodd_case = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*num_class);
    if (wodd_att == NULL || wodd_case == NULL)
      die("Couldn't allocate memory for wodd_{att,case}\n");
    
    for (k=0; k<num_class; k++) {
      wodd_att[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
      wodd_case[k] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_entry);
      if (wodd_att[k] == NULL || wodd_case[k] == NULL)
	die("Couldn't allocate memory for wodd_{att,case}[k]\n");

      for (v=0; v<tuple_size; v++) {
	wodd_att[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
	if (wodd_att[k][v] == NULL)
	  die("Couldn't allocate memory for wodd_att[k][v]\n");
      }
    }
  }
}

void release_cluster(void)
{
  int k,v;

  for (v=0; v<tuple_size; v++) {
    for (k=0; k<num_class; k++) {
      free(att_cluster[v][k]); att_cluster[v][k] = NULL;
    }
    free(att_cluster[v]); att_cluster[v] = NULL;
  }
  free(att_cluster); att_cluster = NULL;

  for (v=0; v<tuple_size; v++) {
    free(att_cluster_pred[v]); att_cluster_pred[v] = NULL;
    free(att_cluster_freq[v]); att_cluster_freq[v] = NULL;
  }
  free(att_cluster_pred); att_cluster_pred = NULL;
  free(att_cluster_freq); att_cluster_freq = NULL;

  for (k=0; k<num_class; k++) {
    free(case_cluster[k]); case_cluster[k] = NULL;
  }
  free(case_cluster); case_cluster = NULL;

  free(case_cluster_pred); case_cluster_pred = NULL;
  free(case_cluster_freq); case_cluster_freq = NULL;

  if (cl_type == CL_WODD) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size; v++) {
	free(wodd_att[k][v]); wodd_att[k][v] = NULL;
      }
      free(wodd_att[k]); wodd_att[k] = NULL;
      free(wodd_case[k]); wodd_case[k] = NULL;
    }
    free(wodd_att); wodd_att = NULL;
    free(wodd_case); wodd_case = NULL;
  }
}

void allocate_rank(void)
{
  int k,v;

  rank_class = (int *)malloc(sizeof(int)*num_class);
  rank_att_mem = (int ***)malloc(sizeof(int **)*num_class);
  rank_case_mem = (int **)malloc(sizeof(int **)*num_class);
  if (rank_class == NULL || rank_att_mem == NULL || rank_case_mem == NULL)
    die("Couldn't allocate memory for rank_class or rank_{case,att}_mem\n");

  for (k=0; k<num_class; k++) {
    rank_case_mem[k] = (int *)malloc(sizeof(int)*num_entry);
    if (rank_case_mem[k] == NULL)
      die("Couldn't allocate memory for rank_case_mem[k]\n");

    rank_att_mem[k] = (int **)malloc(sizeof(int *)*tuple_size);
    if (rank_att_mem[k] == NULL)
      die("Couldn't allocate memory for rank_att_mem[k]\n");
    for (v=0; v<tuple_size; v++) {
      rank_att_mem[k][v] = (int *)malloc(sizeof(int)*num_word[v]);
      if (rank_att_mem[k][v] == NULL)
	die("Couldn't allocate memory for rank_att_mem[k][v]\n");
    }
  }

  if (rel_m == REL_WODD) {
    wodd_case = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*num_class);
    wodd_att = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class);
    if (wodd_att == NULL || wodd_case == NULL)
      die("Couldn't allocate memory for wodd_{att,case}\n");

    for (k=0; k<num_class; k++) {
      wodd_case[k] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_entry);
      wodd_att[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
      if (wodd_case[k] == NULL || wodd_att[k] == NULL)
	die("Couldn't allocate memory for wodd_{case,att}[k]\n");

      for (v=0; v<tuple_size; v++) {
	wodd_att[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
	if (wodd_att[k][v] == NULL)
	  die("Couldn't allocate memory for wodd_att[k][v]\n");
      }
    }
  }

}

void release_rank(void)
{
  int k,v;

  free(rank_class); rank_class = NULL;

  for (k=0; k<num_class; k++) {
    free(rank_case_mem[k]); rank_case_mem[k] = NULL;
  }
  free(rank_case_mem); rank_case_mem = NULL;

  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      free(rank_att_mem[k][v]); rank_att_mem[k][v] = NULL;
    }
    free(rank_att_mem[k]); rank_att_mem[k] = NULL;
  }
  free(rank_att_mem); rank_att_mem = NULL;
  
  if (rel_m == REL_WODD) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size; v++) {
	free(wodd_att[k][v]); wodd_att[k][v] = NULL;
      }
      free(wodd_att[k]); wodd_att[k] = NULL;
      free(wodd_case[k]); wodd_case[k] = NULL;
    }
    free(wodd_att); wodd_att = NULL;
    free(wodd_case); wodd_case = NULL;
  }
}

void allocate_rank_sm(void)
{
  int k,l,v;

  rank_att_mem_sm = (int ***)malloc(sizeof(int **)*num_class);
  rank_att_mem_part = (int ***)malloc(sizeof(int **)*3);
  if (rank_att_mem_sm == NULL || rank_att_mem_part == NULL)
    die("Couldn't allocate memory for rank_att_mem_{sm,part}\n");
  
  for (k=0; k<num_class; k++) {
    rank_att_mem_sm[k] = (int **)malloc(sizeof(int *)*tuple_size);
    if (rank_att_mem_sm[k] == NULL)
      die("Couldn't allocate memory for rank_att_mem_sm[k]\n");
    for (v=0; v<tuple_size; v++) {
      rank_att_mem_sm[k][v] = (int *)malloc(sizeof(int)*num_word[v]);
      if (rank_att_mem_sm[k][v] == NULL)
	die("Couldn't allocate memory for rank_att_mem_sm[k][v]\n");
    }
  }

  for (l=0; l<3; l++) {
    rank_att_mem_part[l] = (int **)malloc(sizeof(int *)*tuple_size);
    if (rank_att_mem_part[l] == NULL)
      die("Couldn't allocate memory for rank_att_mem_part[l]\n");
    for (v=0; v<tuple_size; v++) {
      rank_att_mem_part[l][v] = (int *)malloc(sizeof(int)*num_word[v]);
      if (rank_att_mem_part[l][v] == NULL)
	die("Couldn't allocate memory for rank_att_mem_part[l][v]\n");
    }
  }
}

void release_rank_sm(void)
{
  int k,l,v;

  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      free(rank_att_mem_sm[k][v]); rank_att_mem_sm[k][v] = NULL;
    }
    free(rank_att_mem_sm[k]); rank_att_mem_sm[k] = NULL;
  }
  for (l=0; l<3; l++) {
    for (v=0; v<tuple_size; v++) {
      free(rank_att_mem_part[l][v]); rank_att_mem_part[l][v] = NULL;
    }
    free(rank_att_mem_part[l]); rank_att_mem_part[l] = NULL;
  }
  free(rank_att_mem_sm); rank_att_mem_sm = NULL;
  free(rank_att_mem_part); rank_att_mem_part = NULL;
}

void allocate_log_likelihood_sm(void)
{
  int n,t;

  n = (max_iterate > 0) ? max_iterate : MAX_MAX_ITERATE;

  /* We need one extra entry to record the values for the first full EM */
  log_likelihood_sm = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*(num_sm+1));
  log_prior_sm = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*(num_sm+1));
  lambda_sm = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*(num_sm+1));
  if (log_likelihood_sm == NULL || log_prior_sm == NULL || lambda_sm == NULL)
    die("Couldn't allocate memory for log_likelihood_sm, log_prior_sm or lambda_sm\n");

  for (t=0; t<(num_sm+1); t++) {  /* Be careful for the terminal condition */
    /* At maximum, we have n partial EM iterations and n full EM iterations: */
    log_likelihood_sm[t] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*(2*n));
    log_prior_sm[t] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*(2*n));
    lambda_sm[t] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*(2*n));
    if (log_likelihood_sm[t] == NULL || log_prior_sm[t] == NULL || lambda_sm[t] == NULL)
      die("Couldn't allocate memory for log_likelihood_sm[t], log_prior_sm[t] or lambda_sm[t]\n");
  }
}

void allocate_jmerge()
{
  int k;
  
  jmerge = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*num_class);
  if (jmerge == NULL)
    die("Couldn't allocate memory for jmerge\n");

  for (k=0; k<num_class; k++) {
    jmerge[k] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
    if (jmerge[k] == NULL)
      die("Couldn't allocate memory for jmerge[k]\n");
  }
}

void allocate_jsplit()
{
  jsplit = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
  if (jsplit == NULL)
    die("Couldn't allocate memory for jsplit\n");
}

void allocate_candidate()
{
  int h;
  
  sm_candidate = (int **)malloc(sizeof(int *)*num_candidate);
  if (sm_candidate == NULL)
    die("Couldn't allocate memory for sm_candidate\n");
 
  for (h=0; h<num_candidate; h++) {
    sm_candidate[h] = (int *)malloc(sizeof(int)*3);
    if (sm_candidate[h] == NULL)
      die("Couldn't allocate memory for sm_candidate[h]\n");
  }
}

void allocate_prob_part(void)
{
  int l;

  prob_incomp_part = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_entry);
  prob_cond_part = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE*)*3);
  if (prob_incomp_part == NULL || prob_cond_part == NULL)
    die("Couldn't allocate memory for prob_{incomp,cond}_part\n");

  for (l=0; l<3; l++) {
    prob_cond_part[l] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_entry);
    if (prob_cond_part[l] == NULL)
      die("Couldn't allocate memory for prob_cond_part[l]\n");
  }
}

void allocate_num_iterate_sm()
{
  num_iterate_sm_full = (int *)malloc(sizeof(int)*(num_sm+1));
  num_iterate_sm_part = (int *)malloc(sizeof(int)*(num_sm+1));
  if (num_iterate_sm_full == NULL || num_iterate_sm_part == NULL)
    die("Couldn't allocate memory for num_iterate_sm_{full,part}\n");
}

void allocate_num_result_sm()
{
  num_result_rejected = (int *)malloc(sizeof(int)*num_sm);
  num_result_skipped = (int *)malloc(sizeof(int)*num_sm);
  if (num_result_rejected == NULL || num_result_skipped == NULL)
    die("Couldn't allocate memory for num_result_{rejected,skipped}\n");
}

void release_jmerge()
{
  int k;
  for (k=0; k<num_class; k++) {
    free(jmerge[k]); jmerge[k] = NULL;
  }
  free(jmerge); jmerge = NULL;
}

void release_jsplit()
{
  free(jsplit); jsplit = NULL;
}

void release_candidate()
{
  int h;

  for (h=0; h<num_candidate; h++) {
    free(sm_candidate[h]); sm_candidate[h] = NULL;
  }
  free(sm_candidate); sm_candidate = NULL;
}

void release_prob_part(void)
{
  int l;

  for (l=0; l<3; l++) {
    free(prob_cond_part[l]); prob_cond_part[l] = NULL;
  }
  free(prob_cond_part);   prob_cond_part = NULL;
  free(prob_incomp_part); prob_incomp_part = NULL;
}

void release_log_likelihood_sm(void)
{
  int t;

  for (t=0; t<(num_sm+1); t++) {
    free(log_likelihood_sm[t]); log_likelihood_sm[t] = NULL;
    free(log_prior_sm[t]); log_prior_sm[t] = NULL;
    free(lambda_sm[t]); lambda_sm[t] = NULL;
  }
  free(log_likelihood_sm); log_likelihood_sm = NULL;
  free(log_prior_sm); log_prior_sm = NULL;
  free(lambda_sm); lambda_sm = NULL;
}

void release_num_iterate_sm()
{
  free(num_iterate_sm_full); num_iterate_sm_full = NULL;
  free(num_iterate_sm_part); num_iterate_sm_part = NULL;
}

void release_num_result_sm()
{
  free(num_result_rejected); num_result_rejected = NULL;
  free(num_result_skipped);  num_result_skipped = NULL;
}

void allocate_param_part(void)
{
  param_c_part = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*3);
  if (param_c_part == NULL)
    die("Couldn't allocate memory for param_c_part\n");
}

void  release_param_part(void)
{
  free(param_c_part); param_c_part = NULL;
}

void allocate_param_sm(void)
{
  int k,v;

  param_c_sm = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
  param_w_sm = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class);
  if (param_c_sm == NULL || param_w_sm == NULL)
    die("Couldn't allocate memory for param_{c,w}_sm\n");

  for (k=0; k<num_class; k++) {
    param_w_sm[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (param_w_sm[k] == NULL)
      die("Couldn't allocate memory for param_w_sm[k]\n");
    for (v=0; v<tuple_size; v++) {
      param_w_sm[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      if (param_w_sm[k][v] == NULL)
	die("Couldn't allocate memory for param_w_sm[k][v]\n");
    }
  }
}

void release_param_sm(void)
{
  int k,v;

  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      free(param_w_sm[k][v]); param_w_sm[k][v] = NULL;
    }
    free(param_w_sm[k]); param_w_sm[k] = NULL;
  }
  free(param_w_sm); param_w_sm = NULL;
  free(param_c_sm); param_c_sm = NULL;
}

void allocate_prob_att_mem_sm(void)
{
  int k,l,v;

  prob_w_sm = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
  if (prob_w_sm == NULL)
    die("Couldn't allocate memory for prob_w_sm\n");
  for (v=0; v<tuple_size; v++) {
    prob_w_sm[v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
    if (prob_w_sm[v] == NULL)
      die("Couldn't allocate memory for prob_w_sm[v]\n");
  }

  prob_att_mem_sm = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class);
  prob_att_mem_part = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*3);
  if (prob_att_mem_sm == NULL || prob_att_mem_part == NULL)
    die("Couldn't allocate memory for prob_att_mem_{sm,part}\n");

  for (k=0; k<num_class; k++) {
    prob_att_mem_sm[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (prob_att_mem_sm[k] == NULL)
      die("Couldn't allocate memory for prob_att_mem_sm[k]\n");
    for (v=0; v<tuple_size; v++) {
      prob_att_mem_sm[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      if (prob_att_mem_sm[k][v] == NULL)
	die("Couldn't allocate memory for prob_att_mem_sm[k][v]\n");
    }
  }

  for (l=0; l<3; l++) {
    prob_att_mem_part[l] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (prob_att_mem_part[l] == NULL)
      die("Couldn't allocate memory for prob_att_mem{_part[l]\n");
    for (v=0; v<tuple_size; v++) {
      prob_att_mem_part[l][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      if (prob_att_mem_part[l][v] == NULL)
	die("Couldn't allocate memory for prob_att_mem_part[l][v]\n");
    }
  }
}

void release_prob_att_mem_sm(void)
{
  int k,l,v;

  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      free(prob_att_mem_sm[k][v]); prob_att_mem_sm[k][v] = NULL;
    }
    free(prob_att_mem_sm[k]); prob_att_mem_sm[k] = NULL;
  }
  for (l=0; l<3; l++) {
    for (v=0; v<tuple_size; v++) {
      free(prob_att_mem_part[l][v]); prob_att_mem_part[l][v] = NULL;
    }
    free(prob_att_mem_part[l]); prob_att_mem_part[l] = NULL;
  }
  free(prob_att_mem_sm); prob_att_mem_sm = NULL;
  free(prob_att_mem_part); prob_att_mem_part = NULL;

  for (v=0; v<tuple_size; v++) {
    free(prob_w_sm[v]); prob_w_sm[v] = NULL;
  }
  free(prob_w_sm); prob_w_sm = NULL;
}
