#include "nbc_def.h"
#include "nbc_ext.h"
#include "vnbc_def.h"
#include "vnbc_ext.h"

void allocate_free_energy(void)
{
  free_energy = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*MAX_MAX_ITERATE);
  if (free_energy == NULL)
    die("Couldn't allocate memory for free_energy\n");
}

void allocate_free_energy_best(void)
{
  free_energy_best = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*MAX_MAX_ITERATE);
  if (free_energy_best == NULL)
    die("Couldn't allocate memory for free_energy_best\n");
}

void allocate_seed_vb(void)
{
  int u,num_search;

  num_search = (mode_search_num_class) ? ((kmax-kmin)/kstep + 1): 1;

  log_final_free_energy = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*num_search);
  if (log_final_free_energy == NULL)
    die("Couldn't allocate memory for log_final_free_energy\n");

  for (u = 0; u < num_search; u++) {
    log_final_free_energy[u] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_init);
    if (log_final_free_energy[u] == NULL)
      die("Couldn't allocate memory for log_final_free_energy[u]\n");
  }
}

void allocate_hparam_best(void)
{
  int k,v;

  hparam_c_best = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
  hparam_c_upd_best = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
  hparam_w_best = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class);
  hparam_w_upd_best = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class);
  if (hparam_c_best == NULL || hparam_w_best == NULL ||
      hparam_c_upd_best == NULL || hparam_w_upd_best == NULL)
    die("Couldn't allocate memory for hparam_{c,w}_best{,_upd}\n");

  for (k=0; k<num_class; k++) {
    hparam_w_best[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    hparam_w_upd_best[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (hparam_w_best[k] == NULL || hparam_w_upd_best[k] == NULL)
      die("Couldn't allocate memory for hparam_w_best{,_upd}[k]\n");
    for (v=0; v<tuple_size; v++) {
      hparam_w_best[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      hparam_w_upd_best[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      if (hparam_w_best[k][v] == NULL)
        die("Couldn't allocate memory for hparam_w_best{,_upd}[k][v]\n");
    }
  }
}

void allocate_hparam_opt(void)
{
  int k,v;
  
  hparam_c_opt = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
  hparam_c_upd_opt = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
  hparam_w_opt = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class);
  hparam_w_upd_opt = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class);
  if (hparam_c_opt == NULL || hparam_w_opt == NULL ||
      hparam_c_upd_opt == NULL || hparam_w_upd_opt == NULL)
     die("Couldn't allocate memory for hparam_{c,w}_opt{,_upd}\n");

  for (k=0; k<num_class; k++) {
    hparam_w_opt[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    hparam_w_upd_opt[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (hparam_w_opt[k] == NULL || hparam_w_upd_opt[k] == NULL)
      die("Couldn't allocate memory for hparam_w_opt{,_upd}[k]\n");
    for (v=0; v<tuple_size; v++) {
      hparam_w_opt[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      hparam_w_upd_opt[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      if (hparam_w_opt[k][v] == NULL || hparam_w_upd_opt[k][v] == NULL)
        die("Couldn't allocate memory for hparam_w_opt{,_upd}[k][v]\n");
    }
  }
}

void allocate_pi(void)
{
  int k,v;

  pi_c = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
  pi_w = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class);
  if (pi_c == NULL || pi_w == NULL)
    die("Couldn't allocate memory for pi_c or pi_w\n");

  for (k=0; k<num_class; k++) {
    pi_w[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (pi_w[k] == NULL)
      die("Couldn't allocate memory for pi_w[k]\n");
    for (v=0; v<tuple_size; v++) {
      pi_w[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      if (pi_w[k][v] == NULL)
        die("Couldn't allocate memory for pi_w[k][v]\n");
    }
  }
}

void release_free_energy(void)
{
  free(free_energy); free_energy = NULL;
}

void release_free_energy_best(void)
{
  free(free_energy_best); free_energy_best = NULL;
}

void release_seed_vb(void)
{
  int u, num_search;

  num_search = (mode_search_num_class) ? ((kmax-kmin)/kstep + 1): 1;

  for (u=0; u < num_search; u++) {
    free(log_final_free_energy[u]); log_final_free_energy[u] = NULL;
  }
  free(log_final_free_energy); log_final_free_energy = NULL;
}

void allocate_hparam_upd(void)
{
  int k,v;

  hparam_c_upd = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
  hparam_w_upd = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class);
  if (hparam_c_upd == NULL || hparam_w_upd == NULL)
    die("Couldn't allocate memory for hparam_{c,w}_upd\n");

  for (k=0; k<num_class; k++) {
    hparam_w_upd[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (hparam_w_upd[k] == NULL)
      die("Couldn't allocate memory for hparam_w_upd[k]\n");
    for (v=0; v<tuple_size; v++) {
      hparam_w_upd[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      if (hparam_w_upd[k][v] == NULL)
        die("Couldn't allocate memory for hparam_w_upd[k][v]\n");
    }
  }
}

void release_hparam_upd(void)
{
  int k,v;

  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      free(hparam_w_upd[k][v]); hparam_w_upd[k][v] = NULL;
    }
    free(hparam_w_upd[k]); hparam_w_upd[k] = NULL;
  }
  free(hparam_w_upd); hparam_w_upd = NULL;
  free(hparam_c_upd); hparam_c_upd = NULL;
}

void release_hparam_best(void)
{
  int k,v;
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      free(hparam_w_best[k][v]);     hparam_w_best[k][v] = NULL;
      free(hparam_w_upd_best[k][v]); hparam_w_upd_best[k][v] = NULL;
    }
    free(hparam_w_best[k]);     hparam_w_best[k] = NULL;
    free(hparam_w_upd_best[k]); hparam_w_upd_best[k] = NULL;
  }
  free(hparam_w_best);     hparam_w_best = NULL;
  free(hparam_w_upd_best); hparam_w_upd_best = NULL;
  free(hparam_c_best);     hparam_c_best = NULL;
  free(hparam_c_upd_best); hparam_c_upd_best = NULL;
}

void release_hparam_opt(void)
{
  int k,v;

  for (k=0; k<num_class_opt; k++) {
    for (v=0; v<tuple_size; v++) {
      free(hparam_w_opt[k][v]);     hparam_w_opt[k][v] = NULL;
      free(hparam_w_upd_opt[k][v]); hparam_w_upd_opt[k][v] = NULL;
    }
    free(hparam_w_opt[k]);     hparam_w_opt[k] = NULL;
    free(hparam_w_upd_opt[k]); hparam_w_upd_opt[k] = NULL;
  }
  free(hparam_w_opt);     hparam_w_opt = NULL;
  free(hparam_w_upd_opt); hparam_w_upd_opt = NULL;
  free(hparam_c_opt);     hparam_c_opt = NULL;
}

void release_pi(void)
{
  int k,v;
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      free(pi_w[k][v]); pi_w[k][v] = NULL;
    }
    free(pi_w[k]); pi_w[k] = NULL;
  }
  free(pi_w); pi_w = NULL;
  free(pi_c); pi_c = NULL;
}

void allocate_free_energy_sm(void)
{
  int n,t;

  n = (max_iterate > 0) ? max_iterate : MAX_MAX_ITERATE;

  /* We need one extra entry to record the first full EM: */
  free_energy_sm = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*(num_sm + 1));
  if (free_energy_sm == NULL)
    die("Couldn't allocate memory for free_energy_sm\n");

  for (t=0; t<(num_sm+1); t++) {  /* be careful for the terminal condition */
    /* n times for partial EM, and n times for full EM, at maximum */
    free_energy_sm[t] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*(2*n));
    if (free_energy_sm[t] == NULL)
      die("Couldn't allocate memory for free_energy_sm[t]\n");
  }
}

void release_free_energy_sm(void)
{
  int t;

  for (t=0; t<(num_sm+1); t++) {
    free(free_energy_sm[t]); free_energy_sm[t] = NULL;
  }
  free(free_energy_sm); free_energy_sm = NULL;
}

void allocate_hparam_part(void)
{
  hparam_c_upd_part = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*3);
  if (hparam_c_upd_part == NULL)
    die("Couldn't allocate memory for hparam_c_upd_part\n");
}

void allocate_hparam_sm(void)
{
  int k,v;

  hparam_c_sm = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
  hparam_w_sm = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class);
  if (hparam_c_sm == NULL || hparam_w_sm == NULL)
    die("Couldn't allocate memory for hparam_{c,w}_sm\n");

  hparam_c_upd_sm = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
  hparam_w_upd_sm = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class);
  if (hparam_c_upd_sm == NULL || hparam_w_upd_sm == NULL)
    die("Couldn't allocate memory for hparam_{c,w}_upd_sm\n");

  for (k=0; k<num_class; k++) {
    hparam_w_sm[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    hparam_w_upd_sm[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (hparam_w_sm[k] == NULL || hparam_w_upd_sm[k] == NULL)
      die("Couldn't allocate memory for hparam_w{,_upd}_sm[k]\n");
    for (v=0; v<tuple_size; v++) {
      hparam_w_sm[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      hparam_w_upd_sm[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      if (hparam_w_sm[k][v] == NULL || hparam_w_upd_sm[k][v] == NULL)
	die("Couldn't allocate memory for hparam_w{,_upd}_sm[k][v]\n");
    }
  }
}

void release_hparam_sm(void)
{
  int k,v;

  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      free(hparam_w_sm[k][v]); hparam_w_sm[k][v] = NULL;
      free(hparam_w_upd_sm[k][v]); hparam_w_upd_sm[k][v] = NULL;
    }
    free(hparam_w_sm[k]); hparam_w_sm[k] = NULL;
    free(hparam_w_upd_sm[k]); hparam_w_upd_sm[k] = NULL;
  }
  free(hparam_c_sm); hparam_c_sm = NULL;
  free(hparam_c_upd_sm); hparam_c_upd_sm = NULL;
  free(hparam_w_sm); hparam_w_sm = NULL;
  free(hparam_w_upd_sm); hparam_w_upd_sm = NULL;
}

void  release_hparam_part(void)
{
  free(hparam_c_upd_part); hparam_c_upd_part = NULL;
}
