#include "nbc_def.h"
#include "nbc_ext.h"

void cluster_main(void)
{
  FILE *fp_cluster;

  allocate_cluster();

  switch (cl_type) {
  case CL_MP:
    compute_att_cluster_ml();
    break;
  case CL_WODD:
    compute_att_cluster_wodd();
    break;
  default:
    die("Internal error: Unsupported clustering method\n");
  }

  switch (cl_type) {
  case CL_MP:
    compute_case_cluster_ml();
    break;
  case CL_WODD:
    compute_case_cluster_wodd();
    break;
  default:
    die("Internal error: Unsupported clustering method\n");
  }

  print_cluster_freq(fp_log);

  fp_cluster = fopen(fname_cluster,"wb");
  if (fp_cluster == NULL)
    die("Couldn't open %s\n",fname_cluster);
  print_cluster(fp_cluster);
  fclose(fp_cluster);

  release_cluster();
}

void compute_att_cluster_ml(void)
{
  int *tmp_count;
  int k,i,v;
  int k_ml;
  FLOAT_TYPE p_max;

  /* Allocate the temporary areas */
  tmp_count = (int *)malloc(sizeof(int)*num_class);
  if (tmp_count == NULL) die("Couldn't allocate memory for tmp_count\n");

  for (v=0; v<tuple_size; v++) 
    for (k=0; k<num_class; k++)
      att_cluster_freq[v][k] = 0;

  for (v=0; v<tuple_size; v++) {
    for (i=0; i<num_word[v]; i++) {
      k_ml = 0;
      p_max = 0.0;
      for (k=0; k<num_class; k++) {
	if (prob_att_mem[k][v][i] > p_max) {
	  k_ml = k;
	  p_max = prob_att_mem[k][v][i];
	}      
      }
      att_cluster_pred[v][i] = k_ml;
      att_cluster_freq[v][k_ml]++;  /* Count up the att-vals whose most likely class is k_ml */
    }
  }

  for (v=0; v<tuple_size; v++) {
    for (k=0; k<num_class; k++)
      tmp_count[k] = 0;

    for (i=0; i<num_word[v]; i++) {
      k_ml = att_cluster_pred[v][i];
      att_cluster[v][k_ml][tmp_count[k_ml]] = i;
      tmp_count[k_ml]++;
    }

    if (mode_lex_order) {
      tmp_word_list = word_list[v];
      for (k=0; k<num_class; k++) {
        qsort(att_cluster[v][k],att_cluster_freq[v][k],sizeof(int),
              (int(*)(const void *,const void *))compare_word);
      }
    }
  }

  free(tmp_count); tmp_count = NULL;
}

void compute_att_cluster_wodd(void)
{
  int *tmp_count;
  int k_ml;
  FLOAT_TYPE p_max;
  FLOAT_TYPE p_neg;
  int k,i,v;

  /** We first compute the weighted log-odds ratios **/

  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
        p_neg =
          (param_c[k] + TINY_FLOAT) *
          (1.0 - param_w[k][v][i] + TINY_FLOAT) / 
          (1.0 - prob_w[v][i] + TINY_FLOAT);
	wodd_att[k][v][i] =
          prob_att_mem[k][v][i] *
          MATH_LOG((prob_att_mem[k][v][i] + TINY_FLOAT)/p_neg);
      }
    }
  }

  /** The followings are almost the same as the lines in compute_att_cluster_ml() **/

  tmp_count = (int *)malloc(sizeof(int)*num_class);
  if (tmp_count == NULL) die("Couldn't allocate memory for tmp_count\n");

  for (v=0; v<tuple_size; v++) 
    for (k=0; k<num_class; k++)
      att_cluster_freq[v][k] = 0;

  for (v=0; v<tuple_size; v++) {
    for (i=0; i<num_word[v]; i++) {
      k_ml = 0;
      p_max = 0.0;
      for (k=0; k<num_class; k++) {
	if (wodd_att[k][v][i] > p_max) {
	  k_ml = k;
	  p_max = wodd_att[k][v][i];
	}      
      }
      att_cluster_pred[v][i] = k_ml;
      att_cluster_freq[v][k_ml]++;  /* Count up the att-vals whose most likely class is k_ml */
    }
  }

  for (v=0; v<tuple_size; v++) {
    for (k=0; k<num_class; k++)
      tmp_count[k] = 0;

    for (i=0; i<num_word[v]; i++) {
      k_ml = att_cluster_pred[v][i];
      att_cluster[v][k_ml][tmp_count[k_ml]] = i;
      tmp_count[k_ml]++;
    }

    if (mode_lex_order) {
      tmp_word_list = word_list[v];
      for (k=0; k<num_class; k++) {
        qsort(att_cluster[v][k],att_cluster_freq[v][k],sizeof(int),
              (int(*)(const void *,const void *))compare_word);
      }
    }
  }

  free(tmp_count); tmp_count = NULL;
}

void compute_case_cluster_ml(void)
{
  int *tmp_count;
  int k,e;
  int k_ml;
  FLOAT_TYPE p_max;

  tmp_count = (int *)malloc(sizeof(int)*num_class);
  if (tmp_count == NULL)
    die("Couldn't allocate memory for tmp_count\n");

  for (k=0; k<num_class; k++)
    case_cluster_freq[k] = 0;
  
  for (e=0; e<num_entry; e++) {
    k_ml = 0;
    p_max = 0.0;
    for (k=0; k<num_class; k++) {
      if (prob_case_mem[k][e] > p_max) {
        k_ml = k;
        p_max = prob_case_mem[k][e];
      }      
    }
    case_cluster_pred[e] = k_ml;
    case_cluster_freq[k_ml]++;
  }

  for (k=0; k<num_class; k++)
    tmp_count[k] = 0;
  
  for (e=0; e<num_entry; e++) {
    k_ml = case_cluster_pred[e];
    case_cluster[k_ml][tmp_count[k_ml]] = e;
    tmp_count[k_ml]++;
  }
  
  free(tmp_count); tmp_count = NULL;
}

void compute_case_cluster_wodd(void)
{
  int *tmp_count;
  int k_ml;
  FLOAT_TYPE p_max;
  FLOAT_TYPE p_neg;
  int k,v,e;
  FLOAT_TYPE *p_c;
  FLOAT_TYPE p_ic;

  tmp_count = (int *)malloc(sizeof(int)*num_class);
  p_c = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
  if (tmp_count == NULL || p_c == NULL)
    die("Couldn't allocate memory for tmp_count, p_c\n");

  for (e=0; e<num_entry; e++) {
    for (v=0; v<tuple_size; v++)
      freq_w_id[v] = freq_list[e].id[v];

    p_ic = 0.0;
    for (k=0; k<num_class; k++) {
      p_c[k] = param_c[k];
      for (v=0; v<tuple_size; v++)
	p_c[k] *= param_w[k][v][freq_w_id[v]];
      p_ic += p_c[k];
    }

    for (k=0; k<num_class; k++) {
      p_neg =
        (param_c[k] - p_c[k] + TINY_FLOAT) /
        (1.0 - p_ic + TINY_FLOAT);
      wodd_case[k][e] =
        prob_case_mem[k][e] *
        MATH_LOG((prob_case_mem[k][e] + TINY_FLOAT)/p_neg);
    }
  }

  for (k=0; k<num_class; k++)
    case_cluster_freq[k] = 0;

  for (e=0; e<num_entry; e++) {
    k_ml = 0;
    p_max = 0.0;
    for (k=0; k<num_class; k++) {
      if (wodd_case[k][e] > p_max) {
        k_ml = k;
        p_max = wodd_case[k][e];
      }      
    }
    case_cluster_pred[e] = k_ml;
    case_cluster_freq[k_ml]++;
  }

  for (k=0; k<num_class; k++)
    tmp_count[k] = 0;

  for (e=0; e<num_entry; e++) {
    k_ml = case_cluster_pred[e];
    case_cluster[k_ml][tmp_count[k_ml]] = e;
    tmp_count[k_ml]++;
  }

  free(tmp_count); tmp_count = NULL;
  free(p_c); p_c = NULL;
}

