#include "nbc_def.h"
#include "nbc_ext.h"

void dist_main(void)
{
  FILE *fp_dist;

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  stime_kl_dist = current_sec();
  stime_lap_kl_dist = elapsed_sec();
#endif
  fp_dist = fopen(fname_dist,"wb"); /* binary mode */
  if (fp_dist == NULL)
    die("Couldn't open %s\n", fname_dist);
  compute_print_case_kl_dist(fp_dist);
  compute_print_att_kl_dist(fp_dist);
  fclose(fp_dist);
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  etime_kl_dist = current_sec();
  etime_lap_kl_dist = elapsed_sec();
#endif
}

void compute_print_case_kl_dist(FILE *fp)
{
  int k,e0,e1;
  FLOAT_TYPE d,d1=0.0;

  print_case_kl_dist_header(fp);
  for (e0=0; e0<num_entry; e0++) {
    for (e1=0; e1<num_entry; e1++) {
      d = 0.0;
      for (k=0; k<num_class; k++) {
	if (prob_case_mem[k][e0] > 0.0 && prob_case_mem[k][e1] > 0.0) {
	  d1 =
	    prob_case_mem[k][e0] *
	    (MATH_LOG(prob_case_mem[k][e0]) - MATH_LOG(prob_case_mem[k][e1]));
	}
	else if (prob_case_mem[k][e0] > 0.0) {  /* prob_case_mem[k][e1] == 0 */
	  die("Couldn't compute KL div. since P(Class=%d|case-ID=%d) = 0"
	      " (smoothing may be required)\n", k, e1);
	}
	else {
	  d1 = 0.0;
	}
	d += d1;
      }
      print_case_kl_dist(fp,e0,e1,d);
    }
  }
  
  fflush(fp);
}

void compute_print_att_kl_dist(FILE *fp)
{
  int k,v,i,j,i0,j0;
  FLOAT_TYPE d;
  FLOAT_TYPE d1=0.0;

  for (v=0; v<tuple_size; v++) {
    print_att_kl_dist_header(fp,v);
    for (i=0; i<num_word[v]; i++) {
      i0 = print_order[v][i];
      for (j=0; j<num_word[v]; j++) {
        j0 = print_order[v][j];
        d = 0.0;
        for (k=0; k<num_class; k++) {
          if (prob_att_mem[k][v][i0] > 0.0 && prob_att_mem[k][v][j0] > 0.0) {
            d1 =
	      prob_att_mem[k][v][i0] *
	      (MATH_LOG(prob_att_mem[k][v][i0]) - MATH_LOG(prob_att_mem[k][v][j0]));
          }
          else if (prob_att_mem[k][v][i0] > 0.0) {  /* prob_att_mem[k][v][j0] == 0 */
            die("Couldn't compute KL div. since P(Class=%d|%s) = 0"
                " (smoothing may be required)\n", k, word_list[v][j0]);
          }
          else {
            d1 = 0.0;
          }
          d += d1;
        }
	print_att_kl_dist(fp,v,i0,j0,d);
      }
    }
  }
  
  fflush(fp);
}

