#include "nbc_def.h"
#include "nbc_ext.h"

int hash(char *s)
{
  unsigned v;
  for (v=0; *s != '\0'; s++) {
    v = ((v << CHAR_BIT) + *s) % HASH_SIZE;
  }
  return (int)v;
}

void init_hashtable(HASHTABLE *tp)
{
  int i;

  tp->serial_id = 0;
  tp->table = (HASH_NODE_PTR *)malloc(sizeof(HASH_NODE_PTR)*HASH_SIZE);
  if (tp->table == NULL) die("Couldn't allocate memory for tp->table\n");

  for (i=0; i<HASH_SIZE; i++) tp->table[i] = NULL;
}

void release_hashtable(HASHTABLE *tp)
{
  int i;
  HASH_NODE_PTR q,r;

  for (i=0; i<HASH_SIZE; i++) {
    for (q = tp->table[i]; q != NULL; q = r) {
      r = q->next;
      free(q->key); q->key = NULL;
      free(q); q = NULL;
    }
  }

  free(tp->table);
  tp->table = NULL;
}

int register_hashtable(HASHTABLE *tp, char *entry, int check_only)
{
  int cmp;
  HASH_NODE_PTR *p, q;

  p = tp->table + hash(entry);

  while ((*p) != NULL) {
    if ((cmp = strcmp(entry, (*p)->key)) == 0) {
      return (*p)->id;
    } else {
      p = &((*p)->next);
    }
  }

  if (check_only) return -1;

  q = (struct HASH_NODE *)malloc(sizeof(struct HASH_NODE));
  if (q == NULL) die("Couldn't allocate memory for q\n");

  q->key = (char *)malloc(sizeof(char)*(MAX_WORD_LEN + 1));
  if (q->key == NULL) die("Couldn't allocate memory for q->key\n");

  strcpy(q->key, entry);
  q->id = tp->serial_id;
  q->next = NULL;

  tp->serial_id++;
  *p = q;

  return q->id;
}

char **make_entry_list(HASHTABLE *tp)
{
  int i;
  HASH_NODE_PTR q;
  char **el;

  el = (char **)malloc(sizeof(char *)*tp->serial_id);
  if (el == NULL) die("Couldn't allocate memory for el\n");

  for (i=0; i<HASH_SIZE; i++) {
    q = tp->table[i];
    while (q != NULL) {
      el[q->id] = q->key;
      q = q->next;
    }
  }

  return el;
}
