#include "nbc_def.h"
#include "nbc_ext.h"

FLOAT_TYPE compute_bic(void)
{
  FLOAT_TYPE ll;
  int num_param;
  int v;

  if (mode_nomem) {
    ll = compute_log_likelihood_nomem();
  } else {
    compute_complete_prob();
    compute_incomplete_prob();
    ll = compute_log_likelihood();
  }

  num_param = (num_class - 1);
  for (v=0; v<tuple_size; v++) {
    num_param += num_class*(num_word[v] - 1);
  }

  return ll - 0.5*num_param*MATH_LOG(total_freq);
}

FLOAT_TYPE compute_autoclass(void)
{
  FLOAT_TYPE ll;
  FLOAT_TYPE l0,l1,l2;

  if (mode_nomem) {
    ll = compute_expect_nomem();
  } else {
    compute_complete_prob();
    compute_incomplete_prob();
    compute_conditional_prob();
    compute_expect();
    ll = compute_log_likelihood();
  }

  l0 = compute_complete_marginal_log_likelihood();
  l1 = compute_complete_log_likelihood();
  l2 = ll;
  
  /*
  fprintf(fp_msg,conv_msg("l0 = %.12Lf\n"),l0);
  fprintf(fp_msg,conv_msg("l1 = %.12Lf\n"),l1);
  fprintf(fp_msg,conv_msg("l2 = %.12Lf\n"),l2);
  */

  return (l0 - l1 + l2);
}

/*
 * Compute the marginal likelihood with complete data (Cooper and Herskovits, 1992)
 *
 * -- expect_c[num_class],
 *    expect_w1[num_class][num_word]
 *    expect_w2[num_class][num_word]
 *      are assumed to be already computed
 */
FLOAT_TYPE compute_complete_marginal_log_likelihood(void)
{
  FLOAT_TYPE l;

  l =
    (mode_read_hparam) ?
    compute_complete_marginal_log_likelihood1() :
    compute_complete_marginal_log_likelihood2();
  
  return l;
}

FLOAT_TYPE compute_complete_marginal_log_likelihood1(void)
{
  int k,v,i;
  FLOAT_TYPE l;
  FLOAT_TYPE a_sum0, a_sum1;

  a_sum0 = 0.0;
  a_sum1 = 0.0;
  for (k=0; k<num_class; k++) {
    a_sum0 += hparam_c[k] - 1.0;
    a_sum1 += expect_c[k] + hparam_c[k] - 1.0;
  }

  l = LGAMMA(a_sum0) - LGAMMA(a_sum1);
  for (k=0; k<num_class; k++) {
    l += LGAMMA(expect_c[k] + hparam_c[k] - 1.0) - LGAMMA(hparam_c[k] - 1.0);
  }

  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      a_sum0 = 0.0;
      a_sum1 = 0.0;
      for (i=0; i<num_word[v]; i++) {
	a_sum0 += hparam_w[k][v][i] - 1.0;
	a_sum1 += expect_w[k][v][i] + hparam_w[k][v][i] - 1.0;
      }
      l += LGAMMA(a_sum0) - LGAMMA(a_sum1);
      for (i=0; i<num_word[v]; i++) {
	l +=
	  LGAMMA(expect_w[k][v][i] + hparam_w[k][v][i] - 1.0)
	  - LGAMMA(hparam_w[k][v][i] - 1.0);
      }
    }
  }

  return l;
}

FLOAT_TYPE compute_complete_marginal_log_likelihood2(void)
{
  int k,v,i;
  FLOAT_TYPE a_c;
  FLOAT_TYPE *a_w;
  FLOAT_TYPE t1,t2,t3,t4;

  a_c = smooth_c + 1.0;
  a_w = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*tuple_size);
  if (a_w == NULL)
    die("Couldn't allocate memory for a_w\n");

  for (v=0; v<tuple_size; v++) {
    a_w[v] = smooth_w[v] + 1.0;
  }

  t1 =
    LGAMMA(num_class * a_c)
    - LGAMMA(total_freq + num_class * a_c)
    - num_class * LGAMMA(a_c);

  /*
   *  In the case of tuple_size == 2: 
   *    t2 = num_class
   *           * (LGAMMA(num_word1 * a_w1) + LGAMMA(num_word2 * a_w2)
   *              - (num_word1 * LGAMMA(a_w1) + num_word2 * LGAMMA(a_w2)));
   */
  t2 = 0;
  for (v=0; v<tuple_size; v++) {
    t2 += LGAMMA(num_word[v] * a_w[v]);
  }
  for (v=0; v<tuple_size; v++) {
    t2 -= num_word[v] * LGAMMA(a_w[v]);
  }
  t2 *= num_class;

  t3 = 0.0;
  for (k=0; k<num_class; k++) {
    t3 += LGAMMA(expect_c[k] + a_c);
  }
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      t3 -= LGAMMA(expect_c[k] + num_word[v] * a_w[v]);
    }
  }

  t4 = 0.0;
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
        t4 += LGAMMA(expect_w[k][v][i] + a_w[v]);
      }
    }
  }

  free(a_w);

  return (t1 + t2 + t3 + t4);
}

/* Compute the likelihood with the pseudo complete data */
FLOAT_TYPE compute_complete_log_likelihood(void)
{
  int k,v,i;
  FLOAT_TYPE ll;

  ll = 0.0;
  for (k=0; k<num_class; k++) {
    ll += expect_c[k] * MATH_LOG(param_c[k]);
  }
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
        ll += expect_w[k][v][i] * MATH_LOG(param_w[k][v][i]);
      }
    }
  }

  return ll;
}

