#include "nbc_def.h"
#include "nbc_ext.h"

/* open the log files -- hereafter we can use die() */
void open_log()
{
  fp_log = fopen(fname_log,"wb");
  if (fp_log == NULL) {
    fprintf(stderr,"Couldn't open %s\n", fname_log);
    exit(1);
  }

  if (mode_msg) {
    fp_msg = fopen(fname_msg,"wb");
    if (fp_msg == NULL) {
      fprintf(stderr,"Couldn't open %s\n", fname_msg);
      exit(1);
    }
  }
  else {
    fp_msg = stderr;
  }
}

#ifdef USE_ICONV
void initialize_iconv_sys(void)
{
#ifdef CHARCODE
  cd_sys_in = iconv_open(CHARCODE_INTERNAL,CHARCODE);
  if (cd_sys_in == (iconv_t)-1)
    die("iconv_open failed (in: %s) -- run 'iconv -l' for available character codes\n",
	CHARCODE);

  cd_sys_out = iconv_open(CHARCODE,CHARCODE_INTERNAL);
  if (cd_sys_out == (iconv_t)-1)
    die("iconv_open failed (out: %s) -- run 'iconv -l' for available character codes\n",
	CHARCODE);
#endif
}
#endif

#ifdef USE_ICONV
void initialize_iconv()
{
#ifdef CHARCODE
  cd_sys_in = iconv_open(CHARCODE_INTERNAL,CHARCODE);
  if (cd_sys_in == (iconv_t)-1)
    die("iconv_open failed (in: %s) -- run 'iconv -l' for available character codes\n",
	CHARCODE);

  cd_sys_out = iconv_open(CHARCODE,CHARCODE_INTERNAL);
  if (cd_sys_out == (iconv_t)-1)
    die("iconv_open failed (out: %s) -- run 'iconv -l' for available character codes\n",
	CHARCODE);
#endif

  cd_in = iconv_open(CHARCODE_INTERNAL,charcode_in);
  if (cd_in == (iconv_t)-1)
    die("iconv_open failed (in: %s) -- run 'iconv -l' for available character codes\n",
	charcode_in);

  cd_out = iconv_open(charcode_out,CHARCODE_INTERNAL);
  if (cd_out == (iconv_t)-1)
    die("iconv_open failed (out: %s) -- run 'iconv -l' for available character codes\n",
	charcode_out);
}
#endif

/* parse the specification for the number of classes */
void get_num_class(char *s)
{
  char *endptr;
  int i;
  char ks0[10], ks1[10], ks2[10];
  char *s0,*k0,*k1,*k2;
  int state;

  if (*s == '\0') {
    print_usage(stderr);
    exit(1);
  }
  i = strtol(s,&endptr,10);
  /* we have just an integer -- now consider this as the number of classes */
  if (*endptr == '\0') {
    num_class = i;
    return;
  }
  
  /* try to parse the string in the form of Kmin:Kmax:Kstep */
  state = 0;
  k0 = ks0;
  k1 = ks1;
  k2 = ks2;
  s0 = s;
  while (*s0 != '\0') {
    if (*s0 == ':' || *s0 == ',' || *s0 == '-') {
      state++;
    } else if (state == 0) {
      *k0 = *s0;
      k0++;
    } else if (state == 1) {
      *k1 = *s0;
      k1++;
    } else if (state == 2) {
      *k2 = *s0;
      k2++;
    } else {
      fprintf(stderr,"Internal error (get_num_class)\n");
      exit(1);
    }
    s0++;
  }
 
  if (state != 2) {
    fprintf(stderr,"Illegal value for -k option -- %s",s);
    exit(1);
  }

  *k0 = *k1 = *k2 = '\0';

  kmin = get_int(ks0);
  kmax = get_int(ks1);
  kstep = get_int(ks2);
  mode_search_num_class = 1;
}

FLOAT_TYPE get_smooth_c(char *s)
{
  FLOAT_TYPE f = 0.0;

  if (strcmp(s,SMOOTH_AUTO) == 0) {
    f = -1.0;
    mode_auto_smooth = 1;
    mode_auto_smooth_c = MODE_SMOOTH_AUTO;
  }
  else {
    f = get_float(s);
    if (f < 0) {
      fprintf(stderr,"Unexpected value -- %s\n",s);
      print_usage(stderr);
      exit(1);
    }
    mode_auto_smooth_c = MODE_SMOOTH_NONE;
  }
  
  return f;
}

/* parse the specification for attribute pseudo counts 'u1,u2,u3,..' */
int get_smooth_w(char *s)
{
  char buf[MAX_FLOAT_LEN];  /* MAX_FLOAT_LEN is assumed to be positive */
  char *s0;
  int n,k,num_smooth_w;

  /* We first count up the number of commas: */
  n = 0;
  s0 = s;
  while (*s0 != '\0') {
    if (*s0 == ',') n++;
    s0++;
  }
  num_smooth_w = n + 1;

  smooth_w = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_smooth_w);
  if (smooth_w == NULL) {
    fprintf(stderr,"Couldn't allocate memory for smooth_w in get_smooth_w()\n");
    exit(1);
  }
  mode_auto_smooth_w =
    (enum smooth_auto_type *)malloc(sizeof(enum smooth_auto_type)*num_smooth_w);
  if (mode_auto_smooth_w == NULL) {
    fprintf(stderr,"Couldn't allocate memory for mode_auto_smooth_w\n");
    exit(1);
  }

  n = 0;
  k = 0;
  s0 = s;
  while (*s0 != '\0') {
    if (*s0 == ',') {
      buf[k] = '\0';
      if (strcmp(buf,SMOOTH_AUTO) == 0) {
        smooth_w[n] = -1.0;
        mode_auto_smooth = 1;
	mode_auto_smooth_w[n] = MODE_SMOOTH_AUTO;
      }
      else {
        smooth_w[n] = get_float(buf);
	mode_auto_smooth_w[n] = MODE_SMOOTH_NONE;
      }
      n++;
      k = 0;
    } else {
      buf[k] = *s0;
      k++;
    }
    if (k >= MAX_FLOAT_LEN && k > strlen(SMOOTH_AUTO)) {
      fprintf(stderr,"Too long input for a floating point number: %d\n",k);
      fprintf(stderr,"Unexpected value -- %s\n",s);
      print_usage(stderr);
      exit(1);
    }
    s0++;
  }
  buf[k] = '\0';
  if (strcmp(buf,SMOOTH_AUTO) == 0) {
    smooth_w[n] = -1.0;
    mode_auto_smooth_w[n] = MODE_SMOOTH_AUTO;
  }
  else {
    smooth_w[n] = get_float(buf);
    mode_auto_smooth_w[n] = MODE_SMOOTH_NONE;
  }

  return num_smooth_w;
}

/* parse a string representation of integer */
int get_int(char *s)
{
  char *endptr;
  int i;

  if (*s == '\0') {
    print_usage(stderr);
    exit(1);
  }
  i = strtol(s, &endptr, 10);
  if (*endptr != '\0') {
    fprintf(stderr,"Unexpected value -- %s\n",s);
    print_usage(stderr);
    exit(1);
  }

  return i;
}

/* parse a string representation of unsigned integer */
unsigned int get_uint(char *s)
{
  char *endptr;
  unsigned int i;

  if (*s == '\0') {
    print_usage(stderr);
    exit(1);
  }
  i = strtoul(s, &endptr, 10);
  if (*endptr != '\0') {
    fprintf(stderr,"Unexpected value -- %s\n",s);
    print_usage(stderr);
    exit(1);
  }

  return i;
}

/* parse a string representation of floating-point numbers */
FLOAT_TYPE get_float(char *s)
{
  char *endptr;
  FLOAT_TYPE d;

  if (*s == '\0') {
    print_usage(stderr);
    exit(1);
  }
  d = strtod(s, &endptr);
  if (*endptr != '\0') {
    fprintf(stderr,"Unexpected value -- %s\n",s);
    print_usage(stderr);
    exit(1);
  }
  return d;
}

/* get a character */
char get_char(char *s)
{
  if (*s == '\0') {
    print_usage(stderr);
    exit(1);
  }
  return s[0];
}

/* set the random seed */
void initialize_rand()
{
  srand(rand_seed);
}

/* set the pseudo counts based on the current number of classes */
void set_auto_smooth(void)
{
  int v;
  int max_num_word;
  FLOAT_TYPE max_alpha;

  if (mode_bdeu) {
    max_num_word = -1;
    for (v=0; v<tuple_size; v++) {
      max_num_word = (num_word[v] > max_num_word) ? num_word[v] : max_num_word;
    }
    max_alpha = (1.0 + bdeu_min_smooth) * num_class * max_num_word;
    
    smooth_c = max_alpha / num_class - 1.0;
    for (v=0; v<tuple_size; v++) {
      smooth_w[v] = max_alpha / (num_class * num_word[v]) - 1.0;
    }
  }
  else {
    if (mode_auto_smooth_c == MODE_SMOOTH_AUTO) {
      smooth_c = 1.0 / num_class;
    }
    for (v=0; v<tuple_size; v++) {
      if (mode_auto_smooth_w[v] == MODE_SMOOTH_AUTO) {
	smooth_w[v] = 1.0 / num_word[v];
      }
    }
  }

  if (mode_auto_smooth || mode_bdeu) print_smooth(fp_log, num_class);

  mode_smooth = (smooth_c > 0) ? 1 : 0;
  for (v=0; v<tuple_size; v++) {
    mode_smooth |= (smooth_w[v] > 0) ? 1 : 0;
  }
  mode_smooth_strict = (smooth_c > 0) ? 1 : 0;
  for (v=0; v<tuple_size; v++) {
    mode_smooth_strict &= (smooth_w[v] > 0) ? 1 : 0;
  }

  if (!mode_smooth_strict && ic_type == AUTOCLASS) {
    fprintf(stderr,"Couldn't apply AUTOCLASS unless all smoothing constant > 0\n");
    print_usage(stderr);
    exit(1);
  }
}
