#include "nbc_def.h"
#include "nbc_ext.h"

void die(char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
  vfprintf(fp_log,fmt,ap); fflush(fp_log);
  va_end(ap);

  va_start(ap,fmt);
  vfprintf(fp_msg,fmt,ap); fflush(fp_msg);
  va_end(ap);

  fclose(fp_log);
  exit(1);
}

void die_s(int line_c, char *buf)
{
  die("Illegal line %d: %s",line_c,conv_msg(buf));
}

void die_line(int line_c, int id, char *buf)
{
  die("[%d] Illegal line %d: %s",id,line_c,conv_msg(buf));
}

void log_msg(char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
  vfprintf(fp_msg,fmt,ap);
  va_end(ap);
  fflush(fp_msg);
}

/* CR/LF conversion and %.nnLf -> %.nnf */
char *conv_base(char *buf_in)
{
  char *bi, *bm;
  int l;
  int state; /* state for %.nnLf -> %.nnf conversion */

  *buf_base = '\0';
  bi = buf_in;
  bm = buf_base;
  l = 0;
  state = 0;
#ifdef USE_CRLF
  while (*bi != '\0') {
    if (l >= MAX_OUT_LEN) die("Internal error (conv_base): buffer overflow\n");
    if (*bi == '\n') {
      *bm = '\r';   /* insert CR */
      *(bm+1) = '\n';
      bm += 2;
      l += 2;
    } else if (*bi == '%') {
      state = 1;
      *bm = *bi;  bm++;  l++;
    } else if (state == 1 && *bi == '.') {
      state = 2;
      *bm = *bi;  bm++;  l++;
    } else if ((state == 1 || state == 2) && isdigit(*bi)) {
      state = 3;
      *bm = *bi;  bm++;  l++;
    } else if (state == 3 && isdigit(*bi)) {
      state = 3;
      *bm = *bi;  bm++;  l++;
    } else if (state == 3 && *bi == 'L') {
      state = 4;
#ifndef NO_LDOUBLE
      *bm = *bi;  bm++;  l++;
#endif
    } else if (state == 4 && *bi == 'f') {
      state = 0;
      *bm = *bi;  bm++;  l++;
    } else {
      *bm = *bi;  bm++;  l++;
    }
    bi++;
  }
#else
  /* Only convert %.nnLf -> %.nnf */
  while (*bi != '\0') {
    if (l >= MAX_OUT_LEN) die("Internal error (conv_base): buffer overflow\n");
    if (*bi == '%') {
      state = 1;
      *bm = *bi;  bm++;  l++;
    } else if (state == 1 && *bi == '.') {
      state = 2;
      *bm = *bi;  bm++;  l++;
    } else if ((state == 1 || state == 2) && isdigit(*bi)) {
      state = 3;
      *bm = *bi;  bm++;  l++;
    } else if (state == 3 && isdigit(*bi)) {
      state = 3;
      *bm = *bi;  bm++;  l++;
    } else if (state == 3 && *bi == 'L') {
      state = 4;
#ifndef NO_LDOUBLE
      *bm = *bi;  bm++;  l++;
#endif
    } else if (state == 4 && *bi == 'f') {
      state = 0;
      *bm = *bi;  bm++;  l++;
    } else {
      *bm = *bi;  bm++;  l++;
    }
    bi++;
  }
#endif
  *bm = '\0';

  return buf_base;
}

/* Conversion from the internal charcode to the system's default charcode */
char *conv_sys(char *buf_in)
{
#if defined(USE_ICONV) && defined(CHARCODE)
  int buf_len_in0;
  size_t buf_len_in, buf_len_out;
  char *b;
#ifdef ICONV_CONST
  const char *bi;
#else
  char *bi;
#endif

  if (!mode_sys_out) return buf_in;

  *buf_sys = '\0';
  b = buf_sys;
  bi = buf_in;
  buf_len_in0 = strlen(buf_in);
  buf_len_in = (size_t)buf_len_in0;
  buf_len_out = (size_t)MAX_OUT_LEN;
  if (iconv(cd_sys_out,&bi,&buf_len_in,&b,&buf_len_out) == (size_t)-1)
    die("Couldn't convert: %s (errno=%d)\n",buf_in,errno);
  buf_sys[MAX_OUT_LEN - buf_len_out] = '\0';
#else
  return buf_in;
#endif

  return buf_sys;
}

char *conv_msg(char *buf_in)
{
  return conv_sys(conv_base(buf_in));
}

/* Conversion of the delimiters: ',' -> ',' or '\t' */
char *conv_delim(char *buf_in)
{
  char *bi,*bm;
  int l;

  *buf_delim = '\0';
  bi = buf_in;
  bm = buf_delim;
  l = 0;
  while (*bi != '\0') {
    if (l >= MAX_OUT_LEN) die("Internal error (conv_delim): buffer overflow\n");
    if (*bi == '~') {
      bi++;
      if (*bi == '\0' || *bi == '\n')
        die("Internal error (conv_delim): illegal escape sequence\n");
      *bm = *bi;
      bm++; l++;
    }
    else if (*bi == ',') {
      *bm = (mode_csv_out) ? *bi : '\t';
      bm++; l++;
    }
    else {
      *bm = *bi;
      bm++; l++;
    }
    bi++;
  }
  *bm = '\0';

  return buf_delim;
}

/* Conversion from the internal charcode to the charcode specified by -q flag */
char *conv_out(char *buf_in)
{
#ifdef USE_ICONV
  int buf_len_in0;
  size_t buf_len_in, buf_len_out;
  char *b;
#  ifdef ICONV_CONST
  const char *bi;
#  else
  char *bi;
#  endif
#endif

#ifdef USE_ICONV
  if (!mode_conv_out) return buf_in;

  *buf_out = '\0';
  b = buf_out;
  bi = buf_in;
  buf_len_in0 = strlen(buf_in);
  buf_len_in = (size_t)buf_len_in0;
  buf_len_out = (size_t)MAX_OUT_LEN;
  if (iconv(cd_out,&bi,&buf_len_in,&b,&buf_len_out) == (size_t)-1)
    die("Couldn't convert: %s (errno=%d)\n",buf_in,errno);
  buf_out[MAX_OUT_LEN - buf_len_out] = '\0';
#else
  return buf_in;
#endif

  return buf_out;
}

char *conv_data(char *buf_in)
{
  return conv_out(conv_delim(conv_base(buf_in)));
}

void conv_fprintf(FILE *fp, char *fmt, ...)
{
  va_list ap;

  va_start(ap,fmt);
  vfprintf(fp,conv_data(fmt),ap);
  va_end(ap);
}

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)

/* get the current time in seconds */
FLOAT_TYPE current_sec()
{
  struct rusage ru;
  FLOAT_TYPE t;

  getrusage(RUSAGE_SELF, &ru);

  t =
    (FLOAT_TYPE)(ru.ru_utime.tv_sec) +
    ((FLOAT_TYPE)(ru.ru_utime.tv_usec))/1.0e+6;

  return t;
}

/* get the elasped time in seconds */
FLOAT_TYPE elapsed_sec()
{
  struct timeval tv;
  FLOAT_TYPE t;

  gettimeofday(&tv, NULL);

  t = (FLOAT_TYPE)tv.tv_sec + ((FLOAT_TYPE)tv.tv_usec)/1.0e+6;
  return t;
}
#endif

void print_dot_message(int m)
{
  if (m % 10 == 0) {
    if (m % 100 == 0) fprintf(fp_msg,"%d",m);
    else fprintf(fp_msg,".");
    if (fp_msg == stderr) fflush(fp_msg);
  }
}
