#ifndef NBC_DEF_H
#define NBC_DEF_H

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <limits.h>
#include <math.h>
#ifdef USE_ICONV
#include <iconv.h>
#endif
#include <errno.h>
#include <stdarg.h>
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
#  include <sys/time.h>
#  include <sys/resource.h>
#  include <unistd.h>
#endif
#ifdef USE_FREE_GETOPT
#  include "getopt/getopt.h"
#else
#  ifdef HAVE_GETOPT_H
#    include <getopt.h>
#  endif
#endif
#ifdef _OPENMP
#  include <omp.h>
#endif

#define TOOL_NAME "NBCT"
#ifndef VERSION
#  define VERSION "unknown"
#endif
#define DEFAULT_CLASS_SMOOTH (0)
#define DEFAULT_WORD_SMOOTH (0)
#ifdef NO_LDOUBLE
#  define DEFAULT_EPSILON (1.0e-5)
#else
#  define DEFAULT_EPSILON (1.0e-5L)
#endif
#define DEFAULT_MAX_ITERATE (0)
#define DEFAULT_RAND_SEED ((unsigned int)time(NULL))
#ifdef NO_LDOUBLE
#  define ACCEPT_SMALL_NUMBER      (1.0e-9)
#else
#  define ACCEPT_SMALL_NUMBER      (1.0e-9L)
#endif
#define HASH_SIZE (999991)
#define MAX_WORD_LEN (1024)
#define MAX_FREQ_LEN (10)
#define MAX_INT_LEN (10)
#define MAX_FLOAT_LEN (30)
#define MAX_LINE_LEN (65536)
#define MAX_OUT_LEN (65536)
#define ENUM_TYPE_LEN (64)
#define CHARCODE_LEN (100)
#define MAX_MAX_ITERATE (10000)
#define NOISE_RATIO (1.0e-2)
#define IGNORE_JSPLIT (1.0e-6)
#define MAX_FILENAME (1024)
#define MAX_FILENAME_AUX (2 * MAX_FILENAME + 32)
#define SMALL_FLOAT (1.0e-6)
#define TINY_FLOAT (1.0e-15)
#define INIT_TMP_CLASS (50)
#define INIT_TMP_WORD (1000)
#define CHARCODE_INTERNAL "UTF-8"
#define FSUFFIX_FREQ    "freq"
#define FSUFFIX_PARAM   "param"
#define FSUFFIX_SMOOTH  "smooth"
#define FSUFFIX_DIST    "dissim"
#define FSUFFIX_MEMP    "memp"
#define FSUFFIX_LOG     "log"
#define FSUFFIX_MSG     "msg"
#define FSUFFIX_CLUSTER "cluster"
#define FSUFFIX_RANK    "rank"
#define FSUFFIX_SMEM    "smem"
#define SMOOTH_AUTO     "auto"
#define MIN_SMEM_IMPROVE_RATIO (10.0)

#ifdef NO_LDOUBLE
typedef double FLOAT_TYPE;
#else
typedef long double FLOAT_TYPE;
#endif

typedef struct HASH_NODE {
  char *key;
  int id;
  struct HASH_NODE *next;
} *HASH_NODE_PTR;

typedef struct _HASHTABLE {
  HASH_NODE_PTR *table;
  int serial_id;
} HASHTABLE;

typedef struct FREQ {
  int *id;
  int freq;
  int line_no;
  int ans_id;
} *FREQ_PTR;

enum info_criterion_type {BIC, AUTOCLASS, VB};
enum cluster_type {CL_MP, CL_WODD};
enum initialization_type_for_class {RANDOM_C, NOISY_UNIFORM_C };
enum initialization_type_for_word {RANDOM_W, NOISY_UNIFORM_W, WORD_FREQ};
enum relevance_measure
  {REL_MP, REL_WODD};
enum smooth_auto_type {MODE_SMOOTH_NONE, MODE_SMOOTH_AUTO};

void parse_args(int, char **);
void initialize_modes(void);

void get_num_class(char *);
FLOAT_TYPE get_smooth_c(char *);
int get_smooth_w(char *);
int get_int(char *);
unsigned int get_uint(char *);
FLOAT_TYPE get_float(char *);
char get_char(char *);
void open_log(void);
#ifdef USE_ICONV
void initialize_iconv(void);
void initialize_iconv_sys(void);
#endif
void initialize_rand(void);
void set_auto_smooth(void);

void em_main(void);
void no_em_main(void);
void em_iterate(int);
int em_iterate_body(int,int,int);
int em_iterate_body_nomem(int,int,int);
void compute_incomplete_prob(void);
FLOAT_TYPE compute_log_likelihood(void);
int is_converged(int,int,int);
int is_decreased(int,int,int);
void initialize_param(void);
void compute_complete_prob(void);
void compute_conditional_prob(void);
FLOAT_TYPE compute_log_likelihood_nomem(void);
FLOAT_TYPE compute_log_prior(void);
FLOAT_TYPE compute_expect_nomem(void);
void compute_expect(void);
void update_param(void);
void check_param(void);
void compute_member_prob(void);
void save_log_likelihood_best(int);
void restore_log_likelihood_best(int);
void save_param_opt(void);
void save_param_best(void);
void restore_param_opt(void);
void restore_param_best(void);
void print_dot_message(int);

void smem_iterate(int);
int smem_iterate_full(int,int,int);
int smem_iterate_part(int,int,int,int);
void compute_jmerge(void);
void sort_candidate(void);
int compare_candidate(const int **, const int **);
void compute_incomplete_prob_part(int *);
void compute_conditional_prob_part(int *);
void save_incomplete_prob_part(int *);
void save_conditional_prob_part(int *);
void save_log_likelihood_best_sm(int,int);
void restore_log_likelihood_best_sm(int,int);
void compute_jsplit(void);
void initialize_param_part(int *);
void compute_complete_prob_part(int *);
void compute_expect_part(int *);
void check_expect_part(int *);
void compute_member_prob_sm(void);
void compute_member_prob_part(int *);
void update_param_part(int *);
void check_param_part(int *);
void save_param_part(int *);
void save_param_sm(void);
void restore_param_sm(void);

FLOAT_TYPE compute_info_criterion(enum info_criterion_type);
FLOAT_TYPE compute_bic(void);
FLOAT_TYPE compute_autoclass(void);
FLOAT_TYPE compute_complete_marginal_log_likelihood(void);
FLOAT_TYPE compute_complete_marginal_log_likelihood1(void);
FLOAT_TYPE compute_complete_marginal_log_likelihood2(void);
FLOAT_TYPE compute_complete_log_likelihood(void);

void cluster_main(void);
void cluster_ml_main(void);
void compute_case_cluster_ml(void);
void compute_case_cluster_wodd(void);
void compute_att_cluster_ml(void);
void compute_att_cluster_wodd(void);

void rank_main(void);
int compare_prob_case_mem(const int *, const int *);
int compare_prob_att_mem(const int *, const int *);
int compare_param_c(const int *, const int *);
void compute_rank(void);
void compute_rank_att(void);
void compute_rank_case(void);
void compute_rank_sm(void);
void compute_rank_part(int *);

void vbem_main(void);
void no_vbem_main(void);

void read_freq_body(void);
int get_number_of_entries(void);
void read_freq(void);
int compare_word(const int *, const int *);
void register_freq(int,char **,int,char *,char *,char *);
void make_word_freq_list(void);

void copy_read_param(void);
void expand_read_param_c(int,int);
void expand_read_param_w(int,int,int *,int *);
int read_param_body(FILE *);
void read_param(void);
void read_hparam(void);
void check_hparam(void);

void print_usage(FILE *);
void print_smooth(FILE *,int);
void print_log_likelihood(FILE *,int,int);
void print_ic(FILE *);
void print_seed(FILE *);
void print_decrease(FILE *,int,int,FLOAT_TYPE,FLOAT_TYPE);
void print_jmerge(FILE *);
void print_jsplit(FILE *);
void print_log_likelihood_sm(FILE *,int);
void print_num_iterate_sm(FILE *,int);
void print_num_result_sm(FILE *,int);
void print_time(FILE *);
void print_settings(FILE *);
void print_att_list(FILE *);
void print_att_freq_list(FILE *);
void print_case_freq_list(FILE *);
void print_ans_class_list(FILE *);
void print_size(FILE *);
void print_param(FILE *);
void print_complete_prob(FILE *);
void print_incomplete_prob(FILE *);
void print_conditional_prob(FILE *);
void print_expectation(FILE *);
void print_member_prob(FILE *);
void print_cluster_ml(FILE *);
void print_ml_class_freq(FILE *);
void print_rank(FILE *);
void print_rank_sm(FILE *,int,int *);
void print_cluster(FILE *);
void print_cluster_freq(FILE *);
void print_case_kl_dist_header(FILE *);
void print_case_kl_dist(FILE *, int, int, FLOAT_TYPE);
void print_att_kl_dist_header(FILE *, int);
void print_att_kl_dist(FILE *, int, int, int, FLOAT_TYPE);

void allocate_buffer(void);
void release_buffer(void);
void allocate_param(void);
void allocate_param_best(void);
void allocate_hparam(void);
void release_hparam(void);
void allocate_expect(void);
void allocate_prob_mem(void);
void allocate_log_likelihood(void);
void allocate_log_likelihood_best(void);
void allocate_prob(void);
void allocate_ic(void);
void allocate_seed(void);
void allocate_log_likelihood_sm(void);
void allocate_jmerge(void);
void allocate_jsplit(void);
void allocate_candidate(void);
void allocate_prob_part(void);
void allocate_num_iterate_sm(void);
void allocate_num_result_sm(void);
void allocate_param_part(void);
void allocate_param_sm(void);
void allocate_prob_att_mem_sm(void);
void allocate_cluster(void);
void release_cluster(void);
void allocate_cluster_ml(void);
void allocate_rank(void);
void allocate_rank_sm(void);
void allocate_read_param(int,int *);
void release_param(void);
void release_param_best(void);
void release_param_opt(void);
void release_expect(void);
void release_prob_mem(void);
void release_prob(void);
void release_log_likelihood(void);
void release_log_likelihood_best(void);
void release_ic(void);
void release_seed(void);
void release_jmerge(void);
void release_jsplit(void);
void release_candidate(void);
void release_prob_part(void);
void release_log_likelihood_sm(void);
void release_num_iterate_sm(void);
void release_num_result_sm(void);
void release_param_part(void);
void release_param_sm(void);
void release_prob_att_mem_sm(void);
void release_cluster_ml(void);
void release_rank(void);
void release_rank_sm(void);
void release_read_param(int);

void dist_main(void);
void compute_print_case_kl_dist(FILE *);
void compute_print_att_kl_dist(FILE *);

int hash(char *);
void init_hashtable(HASHTABLE *);
void release_hashtable(HASHTABLE *);
int register_hashtable(HASHTABLE *, char *, int);
char **make_entry_list(HASHTABLE *);

FLOAT_TYPE current_sec(void);
FLOAT_TYPE elapsed_sec(void);
void die(char *,...);
void die_s(int,char *);
void die_line(int,int,char *);
void log_msg(char *,...);
char *conv_base(char *);
char *conv_sys(char *);
char *conv_msg(char *);
char *conv_delim(char *);
char *conv_out(char *);
char *conv_data(char *);
void conv_fprintf(FILE *,char *,...);

double lngamma(double);
double digamma(double);

#ifdef NO_LDOUBLE
#define MATH_LOG(x) log((x))
#else
#define MATH_LOG(x) logl((x))
#endif

#define LGAMMA(x) ((FLOAT_TYPE)lngamma((double)(x)))
#define DIGAMMA(x) ((FLOAT_TYPE)digamma((double)(x)))

#endif
