#ifndef NBC_VAR_H
#define NBC_VAR_H

#include "nbc_def.h"

/* Global vars. for files */
char fname_base[MAX_FILENAME];         /* basename */
char fname_suffix[MAX_FILENAME];       /* suffix name */
char fname_freq[MAX_FILENAME_AUX];     /* occurrences of cases */
char fname_param[MAX_FILENAME_AUX];    /* parameters */
char fname_smooth[MAX_FILENAME_AUX];   /* pseudo counts */
char fname_dist[MAX_FILENAME_AUX];     /* dissimilarities */
char fname_memp[MAX_FILENAME_AUX];     /* membership probabilities */
char fname_log[MAX_FILENAME_AUX];      /* log (general) */
char fname_cluster[MAX_FILENAME_AUX];  /* clustering */
char fname_rank[MAX_FILENAME_AUX];     /* relevance analysis */
char fname_rank_sm[MAX_FILENAME_AUX];  /* relevance analysis (SMEM) */
char fname_msg[MAX_FILENAME_AUX];      /* messages */
FILE *fp_log = NULL;
FILE *fp_msg = NULL;
FILE *fp_sm  = NULL;

/* Global vars. for code conversion */
#ifdef USE_ICONV
#  ifdef CHARCODE
iconv_t cd_sys_in, cd_sys_out;  /* Discriptor of the converter (system's defalt) */
#  endif
iconv_t cd_in, cd_out;          /* Discriptor of the converter (user) */
#endif
char *buf_base = NULL;
char *buf_sys  = NULL;
char *buf_delim = NULL;
char *buf_out  = NULL;

/* Global vars. for time measuring */
FLOAT_TYPE stime_read;
FLOAT_TYPE stime_em;
FLOAT_TYPE stime_memp;
FLOAT_TYPE stime_kl_dist;
FLOAT_TYPE etime_read;
FLOAT_TYPE etime_em;
FLOAT_TYPE etime_memp;
FLOAT_TYPE etime_kl_dist;
FLOAT_TYPE stime_lap_read;
FLOAT_TYPE stime_lap_em;
FLOAT_TYPE stime_lap_memp;
FLOAT_TYPE stime_lap_kl_dist;
FLOAT_TYPE etime_lap_read;
FLOAT_TYPE etime_lap_em;
FLOAT_TYPE etime_lap_memp;
FLOAT_TYPE etime_lap_kl_dist;

/* Global vars. for probabilistic inferences */
int num_class_opt;  /* optimal number of classes */
FREQ_PTR freq_list; /* occurrences of objects */
int total_freq;     /* total occurrences of objects */
int kmin;
int kmax;
int kstep;
int num_entry;      /* number of distinct objects */
int num_entry_pre;  /* number of distinct objects (preliminary) */
FLOAT_TYPE *param_c;        /* class parameters */
FLOAT_TYPE *read_param_c;   /* temporary class parameters */
FLOAT_TYPE *param_c_opt;    /* class parameters with the optimal number of classes */
FLOAT_TYPE *param_c_best;   /* class parameters used in random restarts */
FLOAT_TYPE *hparam_c;       /* class hyper-parameters */
FLOAT_TYPE ***hparam_w;     /* attribute hyper-parameters */
FLOAT_TYPE **prob_comp;     /* complete probs. used in EM */
FLOAT_TYPE *prob_incomp;    /* incomplete probs. used in EM */
FLOAT_TYPE **prob_cond;     /* conditinal probs. used in EM */
FLOAT_TYPE *expect_c;       /* expected occurrences of classes E[c] */
FLOAT_TYPE *tmp_expect_c;   /* expected occurrences of classes E[c] (temporary area) */
FLOAT_TYPE *log_likelihood; /* log-likelihood */
FLOAT_TYPE *log_prior;      /* log of the prior prob. */
FLOAT_TYPE *lambda;         /* log of the a posteriori prob. (MAP), or the log-likelihood (ML)*/
FLOAT_TYPE *log_likelihood_best;  /* log-likelihood with the best params in random restarts */
FLOAT_TYPE *log_prior_best;       /* log of the prior prob. with the best params */
FLOAT_TYPE *lambda_best;          /* lambda with the best params */
FLOAT_TYPE final_log_likelihood;  /* converged log-likeilhood  */
FLOAT_TYPE final_log_prior;       /* converged log of prior prob */
FLOAT_TYPE final_lambda;          /* converged lambda */
FLOAT_TYPE final_log_likelihood_best;  /* converged log-likeilhood with the best params */
FLOAT_TYPE final_log_prior_best;       /* converged log of prior prob with the best params */
FLOAT_TYPE final_lambda_best;          /* converged lambda with the best params */
FLOAT_TYPE *ic;             /* model score which is specified in the command line */
FLOAT_TYPE *bic;            /* BIC */
FLOAT_TYPE *autoclass;      /* AutoClass */
int *num_word;         /* numbers of distinct attribute values */
char ***word_list;     /* lists of attribute values */
char **ans_class_list; /* list of answer classes */
int **word_freq_list;  /* numbers of occurrences of attribute values */
int **print_order;     /* ordered indexes of word_list to be printed */
int **case_cluster;    /* clusters of cases (by clusters) */
int *case_cluster_pred;   /* clusters of cases (by cases) */
int *case_cluster_freq;   /* sizes of case clusters */
int ***att_cluster;       /* cluster of attribute values (by clusters) */
int **att_cluster_pred;   /* cluster of attribute values (by attribute values) */
int **att_cluster_freq;   /* sizes of att-val clusters */
FLOAT_TYPE ***param_w;        /* attribute parameters */
FLOAT_TYPE ***read_param_w;   /* temporary attribute parameters */
FLOAT_TYPE ***param_w_opt;    /* attribute parameters with the optimal number of classes */
FLOAT_TYPE ***param_w_best;   /* attribute parameters used in random restarts */
FLOAT_TYPE ***expect_w;       /* expected occrrences of att-vals E[x_j|c] */
FLOAT_TYPE **prob_w;          /* unconditional prob for attributes P(x_j) = \sum_c P(c)P(x_j|c) */
FLOAT_TYPE **prob_case_mem;   /* membership probs for cases P(c|x_1,x_2,...,xJ) */
FLOAT_TYPE ***prob_att_mem;   /* membership probs for att-vals P(c|x_j) */
int **rank_case_mem;          /* relevance rankinng of cases */
int ***rank_att_mem;          /* relevance ranking of att-vals */
FLOAT_TYPE **wodd_case;       /* weighted log-odds ratio for cases */
FLOAT_TYPE ***wodd_att;       /* weighted log-odds ratio for att-vals */
FLOAT_TYPE ***param_w_sm;     /* temporary attribute parameters in SMEM */
FLOAT_TYPE **prob_w_sm;       /* temporary attribute probs P(w) (for logging ony) */
FLOAT_TYPE ***prob_att_mem_sm;    /* membership prob (before applying a SM operation) */
FLOAT_TYPE ***prob_att_mem_part;  /* membership prob (after applying a SM operation) */
int ***rank_att_mem_sm;           /* ranking (before applying a SM operation) */
int ***rank_att_mem_part;         /* ranking (after applying a SM operation) */

/* Global vars. for the log on random restarts  */
unsigned int **log_seed;    /* random seeds */
int **log_num_iterate;      /* number of iterations for each trial */
int *log_r_best;            /* ID of the trial that yields the best parameters */
FLOAT_TYPE **log_final_lambda;  /* converged lambda's */

/* Global vars. for SMEM */
int num_sm;                  /* number of split-merge operations */
int max_candidate;           /* maximum number of split-merge candidates */
int max_jsplit;              /* maximum number of split candidates for each merge candidate */
int mode_rank_sm;            /* flag for the logs on relevance */
int num_rank_sm;             /* number of the relevance att-vals to be displayed */
int mode_norm;               /* flag for normalizing Jmerge */
FLOAT_TYPE min_param_c;      /* lower limit of the probabilities of classes to be split  */
FLOAT_TYPE min_smem_improve; /* lower limit of the improvement obtained by a SM operation */
FLOAT_TYPE min_new_param_c;  /* lower limit of the probabilities of new classes */
int **sm_candidate;          /* split-merge candidates */
FLOAT_TYPE **jmerge;         /* Jmerge scores */
FLOAT_TYPE *norm;            /* Euclid norm for normalizing Jmerge scores */
FLOAT_TYPE *jsplit;          /* Jsplit scores */
FLOAT_TYPE *param_c_sm;      /* temporary parameters for a SM operation */
int num_candidate;           /* current number of SM candidates */
int *num_iterate_sm_full;    /* number of full EM iterations */
int *num_iterate_sm_part;    /* number of partial EM iterations */
int *num_result_rejected;    /* number of rejected candidates */
int *num_result_skipped;     /* number of skipped candidates */
FLOAT_TYPE *param_c_part;        /* class parameters used in partial EM */
FLOAT_TYPE *prob_incomp_part;    /* incomplete probs used in partial EM */
FLOAT_TYPE **prob_cond_part;     /* conditional probs used in partial EM */
FLOAT_TYPE **log_likelihood_sm;  /* log-likelihood in partial EM */
FLOAT_TYPE **log_prior_sm;       /* log of prior prob in partial EM */
FLOAT_TYPE **lambda_sm;          /* lambda in partial EM */

/* Global vars. for user configuration */
int mode_verbose;                /* flag: debugging of numerical computation */
int mode_em;                     /* flag: EM learning */
int mode_csv_in, mode_csv_out;   /* flag: input/output in CSV form */
int mode_conv_in, mode_conv_out; /* flag: code conversion for user */
int mode_sys_out;                /* flag: code conversion for system (default code) */
int mode_nomem;                  /* flag: saving memory in EM */
FLOAT_TYPE smooth_c;             /* smoothing constants for class */
FLOAT_TYPE *smooth_w;            /* smoothing constants for attributes */
int mode_smooth;                 /* flag: smoothing mode */
int mode_smooth_strict;          /* flag: full smoothing mode */
int mode_auto_smooth;            /* flag: automatic smoothing configuration (total)  */
enum smooth_auto_type mode_auto_smooth_c;     /* flag:  auto smoothing config. (class) */
enum smooth_auto_type *mode_auto_smooth_w;    /* flags: auto smoothing config. (attr) */
int mode_bdeu;               /* flag: BDeu */
FLOAT_TYPE bdeu_min_smooth;  /* minimum smoothing constant for BDeu */
int mode_memp;               /* flag: membership probability */
int mode_dist;               /* flag: dissimilarities */
int mode_search_num_class;   /* flag: model selection */
int mode_suffix;             /* flag: filename suffix */
int mode_cluster;            /* flag: clustering */
int mode_set_seed;           /* flag: manual setting of random seeds */
int mode_lex_order;          /* flag: displaying att-vals in the lexicographical order */
int mode_rank;               /* flag: relevance analysis */
int mode_as_is;              /* flag: construncting a list of att-vals from the occurrence data*/
int mode_eval;               /* flag: clustering evaluation (not fully supported) */
int mode_read_hparam;        /* flag: reading smoothing constants from a file */
int mode_msg;                /* flag: writing summary messages to a file */
int num_rank;                /* number of relevant cases/att-vals to be output */
int num_class;               /* number of classes */
int num_ans_class;           /* number of answer classes */
int tuple_size;              /* number of attributes */
int max_iterate;             /* max number of EM iterations */
int num_init;                /* number of restarts */
int max_init_iterate;        /* number of preliminary EM iterations in random restarts */
FLOAT_TYPE epsilon;                  /* threshold for judginng convergence */
enum info_criterion_type ic_type;    /* model score */
enum cluster_type cl_type;           /* clustering score */
enum initialization_type_for_class init_type_class; /* initialization method for class params. */
enum initialization_type_for_word init_type_word;   /* initialization method for attr. params. */
enum relevance_measure rel_m;        /* relevance score */
char ic_type_str[ENUM_TYPE_LEN];     /* string representation of model score */
char cl_type_str[ENUM_TYPE_LEN];     /* string representation of clustering score */
char init_type_str_class[ENUM_TYPE_LEN]; /* string representation of class param initialization */
char init_type_str_word[ENUM_TYPE_LEN];  /* string representation of attr. param initialization */
char rel_m_str[ENUM_TYPE_LEN];       /* string representation of relevance score */
unsigned int rand_seed;              /* random seed */
int *rank_class;               /* ranking of classes according to the magnitude of class probs. */
#ifdef USE_ICONV
char charcode_in[CHARCODE_LEN];      /* input character code (iconv) */
char charcode_out[CHARCODE_LEN];     /* output character code (iconv) */
#endif

/* Global vars. for hashtables */
HASHTABLE *ht;         /* array of hashtables for attributes */
HASHTABLE ht_vec;      /* hashtable for attribute vectors */
HASHTABLE ht_class;    /* hashtable for answer classes */

/* Temporary area for EM */
int *freq_w_id;               /* temporary array for att-val IDs */

/* Temporary area for sorting */
FLOAT_TYPE *tmp_param_c;
FLOAT_TYPE *tmp_rel_m;
FLOAT_TYPE **tmp_jmerge;
FLOAT_TYPE *tmp_jsplit;
char **tmp_word_list;

#endif

