#include "nbc_def.h"
#include "nbc_ext.h"

#ifdef LANG_JA

/*
 * It seems that MSVC++ cannot handle the LF character ('\n') well in the source
 * code that includes UTF-8, so we insert a space character between non-ASCII
 * character and LF (though is a bit dirty).
 */

char usage[] =
    "\n<<<< " TOOL_NAME ": Naive Bayes Clustering 用ツールキット (version " VERSION " / ML+MAP) >>>>\n\n"
    "[使用法] nbc -f basename -k #classes -x #attributes [option...]\n"
    "  引数付きオプション:\n"
    "    -f 入出力ファイルの接頭辞               [必須]\n"
    "    -k クラス数 K もしくは Kmin:Kmax:Kstep  [-R あるいは -H オプションがない場合必須]\n"
    "    -x 属性数                               [必須]\n"
    "    -s 出力ファイルの接尾辞                 [省略時: 接尾辞なし]\n"
    "    -c クラス擬似カウント \n"
    "       （auto:AutoClass もしくは正数）      [省略時: 平滑化なし]\n"
    "    -w 属性擬似カウント \n"
    "       （auto:AutoClass もしくは正数）      [省略時: 平滑化なし]\n"
    "    -V BDeu における最小擬似カウント        [省略時: BDeu は不使用]\n"
    "    -I クラスパラメータの初期化方法 \n"
    "       （random:ランダム, noisy_u:ノイズ付き一様）[省略時: noisy_u]\n"
    "    -J 属性パラメータの初期化方法 \n"
    "       （random:ランダム, noisy_u:ノイズ付き一様: v_freq:属性値出現確率)\n"
    "                                                  [省略時: noisy_u]\n"
    "    -e 収束判定の閾値                             [省略時: 1.0e-6]\n"
    "    -m EM 繰り返し数                              [省略時: 収束まで繰り返し]\n"
    "    -t クラスタリング手法 （mp: 帰属分布, wodd: 重み付き対数尤度比） \n"
    "    -g 情報量規準 （bic:BIC, autoclass:AUTOCLASS）[省略時: BIC]\n"
    "       ※ autoclass は平滑化（-c または -w が指定）のときに限り有効 \n"
    "    -r 乱数の種                                   [省略時: time() の返り値]\n"
    "    -n ランダム再出発における試行数               [省略時: 再出発なし]\n"
    "    -l ランダム再出発における予備的 EM 繰り返し数 [省略時: 収束まで繰り返し]\n"
    "    -z 関連事例・属性値の表示数（関連分析を実行） [省略時: 関連分析を無効化] \n"
    "    -Q 関連度の尺度 （mp:P(k|x), wodd:重みつき対数尤度比） [省略時: mp] \n"
    "    -o 分割併合操作の最大数（SMEM を実行）        [省略時: SMEM を無効化]\n"
    "    -a 分割併合候補の最大数                       [省略時: 上限なし]\n"
    "    -G 各併合対に対する分割候補の最大数           [省略時: 上限なし]\n"
    "    -U 分割併合を適用できるクラス確率の下限       [省略時: 上限なし]\n"
    "    -C 分割併合を適用した後のクラス確率の下限     [省略時: 上限なし]\n"
    "    -N 分割併合操作で向上する対数尤度・事後確率の下限 \n"
    "                               [省略時: 収束判定の閾値（-e で指定）]\n"
#ifdef USE_ICONV
#ifdef CHARCODE
    "    -p 入力文字コード集合（利用可能な文字コード集合 -- 'iconv -l' を実行） \n"
    "                                                  [省略時: " CHARCODE "]\n"
    "    -q 出力文字コード集合（利用可能な文字コード集合 -- 'iconv -l' を実行） \n"
    "                                                  [省略時: " CHARCODE "]\n"
#else
    "    -p 入力文字コード集合（利用可能な文字コード集合 -- 'iconv -l' を実行） \n"
    "                                                  [省略時: " CHARCODE_INTERNAL "]\n"
    "    -q 出力文字コード集合（利用可能な文字コード集合 -- 'iconv -l' を実行） \n"
    "                                                  [省略時: " CHARCODE_INTERNAL "]\n"
#endif
#endif
    "  引数なしオプション:\n"
    "    -R EM を無効化し，ファイルからパラメータファイルを読み込み \n"
    "    -H ファイルからハイパーパラメータを読み込み \n"
    "    -E クラスター評価 \n"
    "    -S 出力メッセージをファイルに出力         [省略時: 標準エラー出力]\n"
    "    -D 帰属分布をファイルに出力 \n"
    "    -d 距離計算を実行 \n"
    "    -i 入力が CSV 形式であると仮定            [省略時: TAB区切り形式]\n"
    "    -j CSV 形式で出力                         [省略時: TAB区切り形式]\n"
    "    -v 途中の計算結果をファイルに出力 \n"
    "    -F Jmerge 値を正規化 \n"
    "    -M 通常 EM においてメモリを節約（計算が遅くなる可能性あり） \n"
    "    -Y 入力ファイルでの出現順に属性値を表示   [省略時: 文字コード順]\n"
    "    -h このメッセージを出力 \n\n";

void print_usage(FILE *fp)
{
  fprintf(fp,conv_msg(usage));
}

void print_log_likelihood(FILE *fp, int k, int num_iterate)
{
  int m;

  if (mode_smooth) {
    conv_fprintf(fp,"#### ID:LOGLIKE-K%d\n#### クラス数~, 更新回数~, 対数尤度~, 対数事前確率~, 対数事後確率 \n####\n",k);
    for (m=0; m<num_iterate; m++) {  /* m is the index for an EM iteration */
      conv_fprintf(fp,"%d,%d,%.12Lf,%.12Lf,%.12Lf\n",
		   num_class,m,log_likelihood[m],log_prior[m],lambda[m]);
    }
  } else {
    conv_fprintf(fp,"#### ID:LOGLIKE-K%d\n#### クラス数~, 更新回数~, 対数尤度 \n####\n",k);
    for (m=0; m<num_iterate; m++) {
      conv_fprintf(fp,"%d,%d,%.12Lf\n",num_class,m,log_likelihood[m]);
    }
  }

  fflush(fp);
}

void print_ic(FILE *fp)
{
  int k,u;

  conv_fprintf(fp,"#### ID:SCORE\n#### クラス数~, BIC~, AUTOCLASS\n"
	          "#### 選択した情報量規準=%s~, 最適クラス数=%d\n####\n",
	       ic_type_str,num_class_opt);
  u = 0;
  for (k = kmin; k <= kmax; k += kstep) {
    if (mode_smooth) {
      conv_fprintf(fp,"%d,%.12Lf,%.12Lf\n",k,bic[u],autoclass[u]);
    } else {
      conv_fprintf(fp,"%d,%.12Lf,N/A\n",k,bic[u]);
    }
    u++;
  }

  fflush(fp);
}

void print_seed(FILE *fp)
{
  int k,u,r;

  if (mode_search_num_class) {
    u = 0;
    for (k = kmin; k <= kmax; k += kstep) {
      if (mode_smooth) {
	conv_fprintf(fp,"#### ID:SEED-K%d\n#### クラス数~, 初期パラメータ試行（最良=%d）~, 乱数の種~, EM繰り返し数~, 対数事後確率 \n####\n",k,log_r_best[u]);
      } else {
	conv_fprintf(fp,"#### ID:SEED-K%d\n#### クラス数~, 初期パラメータ試行（最良=%d）~, 乱数の種~, EM繰り返し数~, 対数尤度 \n####\n",k,log_r_best[u]);
      }
      
      for (r = 0; r < num_init; r++) {
	if (mode_set_seed) {
	  conv_fprintf(fp,"%d,%d,N/A,%d,%.12Lf\n",
		       k,r,
		       log_num_iterate[u][r],log_final_lambda[u][r]);
	} else {
	  conv_fprintf(fp,"%d,%d,%u,%d,%.12Lf\n",
		       k,r,
		       log_seed[u][r],log_num_iterate[u][r],log_final_lambda[u][r]);
	}
      }
      u++;
    }

  } else {
    
    if (mode_smooth) {
      conv_fprintf(fp,"#### ID:SEED\n#### クラス数~, 初期パラメータ試行（最良=%d）~, 乱数の種~, EM繰り返し数~, 対数事後確率 \n####\n",log_r_best[0]);
    } else {
      conv_fprintf(fp,"#### ID:SEED\n#### クラス数~, 初期パラメータ試行（最良=%d）~, 乱数の種~, EM繰り返し数~, 対数尤度 \n####\n",log_r_best[0]);
    }
    for (r = 0; r < num_init; r++) {
      if (mode_set_seed) {
	conv_fprintf(fp,"%d,%d,N/A,%d,%.12Lf\n",
		     num_class,r,
		     log_num_iterate[0][r],log_final_lambda[0][r]);
      } else {
	conv_fprintf(fp,"%d,%d,%u,%d,%.12Lf\n",
		     num_class,r,
		     log_seed[0][r],log_num_iterate[0][r],log_final_lambda[0][r]);
      }
    }
  }
  
  fflush(fp);
}

void print_decrease(FILE *fp, int r, int m,
		    FLOAT_TYPE pre_lambda, FLOAT_TYPE post_lambda)
{
  if (mode_smooth) {
    conv_fprintf(fp,"#### ID:ERR_MSG-DECREASE\n#### Log posterior probability decreased!\n####\n");
  } else {
    conv_fprintf(fp,"#### ID:ERR_MSG-DECREASE\n#### Log-likelihood decreased!\n####\n");
  }
  conv_fprintf(fp,"クラス数,%d\n",num_class);
  conv_fprintf(fp,"初期パラメータ試行,%d\n",r);
  conv_fprintf(fp,"更新回数,%d\n",m);
  if (mode_smooth) {
    conv_fprintf(fp,"対数事後確率（更新前）,%.15Lf\n",pre_lambda);
    conv_fprintf(fp,"対数事後確率（更新後）,%.15Lf\n",post_lambda);
  } else {
    conv_fprintf(fp,"対数尤度（更新前）,%.15Lf\n",pre_lambda);
    conv_fprintf(fp,"対数尤度（更新後）,%.15Lf\n",post_lambda);
  }
}

void print_time(FILE *fp)
{
  conv_fprintf(fp,"#### ID:CPUTIME\n#### CPU時間(sec)\n####\n");
  conv_fprintf(fp,"読み込み,%.6Lf\n",(etime_read - stime_read));
  if (mode_em)
    conv_fprintf(fp,"EM推定,%.6Lf\n",(etime_em - stime_em));
  if (mode_memp)
    conv_fprintf(fp,"帰属分布,%.6Lf\n",(etime_memp - stime_memp));
  if (mode_dist) {
    conv_fprintf(fp,"KL距離,%.6Lf\n",(etime_kl_dist - stime_kl_dist));
  }
  fflush(fp);
  
  conv_fprintf(fp,"#### ID:LAPTIME\n#### 経過時間(sec)\n####\n");
  conv_fprintf(fp,"読み込み,%.6Lf\n",(etime_lap_read - stime_lap_read));
  if (mode_em)
    conv_fprintf(fp,"EM推定,%.6Lf\n",(etime_lap_em - stime_lap_em));
  if (mode_memp)
    conv_fprintf(fp,"帰属分布,%.6Lf\n",(etime_lap_memp - stime_lap_memp));
  if (mode_dist) {
    conv_fprintf(fp,"KL距離,%.6Lf\n",(etime_lap_kl_dist - stime_lap_kl_dist));
  }
  fflush(fp);
}

void print_jmerge(FILE *fp)
{
  int k1,k2;
  conv_fprintf(fp,"#### ID:JMERGE\n#### クラス1~, クラス2~, Jmerge\n####\n");
  for (k1=0; k1<num_class; k1++) {
    for (k2=0; k2<num_class; k2++) {
      conv_fprintf(fp,"%d,%d,%.9Lf\n",k1,k2,jmerge[k1][k2]);
    }
  }
}

void print_jsplit(FILE *fp)
{
  int k;
  conv_fprintf(fp,"#### ID:JSPLIT\n#### クラス1~, クラス2~, Jsplit\n####\n");
  for (k=0; k<num_class; k++) {
    conv_fprintf(fp,"%d,%.9Lf\n",k,jsplit[k]);
  }
}

void print_log_likelihood_sm(FILE *fp, int k)
{
  int t,m;
  
  if (mode_smooth) {
    conv_fprintf(fp,"#### ID:LOGLIKE-SMEM-K%d\n#### クラス数~, 分割併合回数~, 更新回数~, 対数尤度~, 対数事前確率~, 対数事後確率 \n####\n",k);
    for (t=0; t<(num_sm+1); t++) {
      for (m=0; m<num_iterate_sm_full[t]; m++) {
	conv_fprintf(fp,"%d,%d,%d,%.12Lf,%.12Lf,%.12Lf\n",
		     k,t,m,
		     log_likelihood_sm[t][m],
		     log_prior_sm[t][m],
		     lambda_sm[t][m]);
      }
    }
  } else {
    conv_fprintf(fp,"#### ID:LOGLIKE-SMEM-K%d\n#### クラス数~, 分割併合回数~, 更新回数~, 対数尤度 \n####\n",k);
    for (t=0; t<(num_sm+1); t++) {
      for (m=0; m<num_iterate_sm_full[t]; m++) {
	conv_fprintf(fp,"%d,%d,%d,%.12Lf\n",
		     k,t,m,log_likelihood_sm[t][m]);
      }
    }
  }
  
}

void print_num_result_sm(FILE *fp, int k)
{
  int t;
  
  conv_fprintf(fp,"#### ID:SMEM-STAT-SM-K%d\n#### クラス数~, 分割併合回数~, 失敗候補数~, 無視候補数 \n####\n",k);
  for (t=0; t<num_sm; t++) {
    conv_fprintf(fp,"%d,%d,%d,%d\n",k,t,num_result_rejected[t],num_result_skipped[t]);
  }
}

void print_num_iterate_sm(FILE *fp, int k)
{
  int t;
  
  conv_fprintf(fp,"#### ID:SMEM-STAT-EM-K%d\n#### クラス数~, 分割併合回数~, 部分EM更新回数~, 全EM更新回数 \n####\n",k);
  for (t=0; t<(num_sm+1); t++) {
    if (t==0) {
      conv_fprintf(fp,"%d,%d,N/A,%d\n",k,t,num_iterate_sm_full[t]);
    } else {
      conv_fprintf(fp,"%d,%d,%d,%d\n",
		   k,t,num_iterate_sm_part[t],
		   (num_iterate_sm_full[t]-num_iterate_sm_part[t]));
    }
  }
  
}

void print_settings(FILE *fp)
{
  int v;
  
  conv_fprintf(fp,"#### ID:SETTING\n#### 設定 \n####\n");
  conv_fprintf(fp,"頻度データファイル(入力),%s\n",fname_freq);
  conv_fprintf(fp,"パラメータファイル(出力),%s\n",fname_param);
  conv_fprintf(fp,"帰属分布ファイル(出力),%s\n",fname_memp);
  conv_fprintf(fp,"距離ファイル(出力),%s\n",fname_dist);
  conv_fprintf(fp,"ログファイル(出力),%s\n",fname_log);
  if (mode_cluster) {
    conv_fprintf(fp,"クラスタファイル(出力),%s\n",fname_cluster);
  }
  if (mode_rank) {
    conv_fprintf(fp,"属性順位ファイル(出力),%s\n",fname_rank);
  }
  if (mode_rank_sm) {
    conv_fprintf(fp,"属性順位SMEMログファイル(出力),%s\n",fname_rank_sm);
  }
  if (!mode_search_num_class) {
    conv_fprintf(fp,"クラス数,%d\n",num_class);
  } else {
    conv_fprintf(fp,"クラス数(最小),%d\n",kmin);
    conv_fprintf(fp,"クラス数(最大),%d\n",kmax);
    conv_fprintf(fp,"クラス数(間隔),%d\n",kstep);
  }
  conv_fprintf(fp,"共起属性数,%d\n",tuple_size);
  conv_fprintf(fp,"平滑化モード,%d\n",mode_smooth);
  conv_fprintf(fp,"平滑化モード（全指定）,%d\n",mode_smooth_strict);
  conv_fprintf(fp,"帰属分布計算モード,%d\n",mode_memp);
  conv_fprintf(fp,"属性間距離計算モード,%d\n",mode_dist);
  conv_fprintf(fp,"最適クラス数探索モード,%d\n",mode_search_num_class);
  conv_fprintf(fp,"CSV入力モード,%d\n",mode_csv_in);
  conv_fprintf(fp,"CSV出力モード,%d\n",mode_csv_out);
#ifdef USE_ICONV
  conv_fprintf(fp,"入力文字コード,%s\n",charcode_in);
  conv_fprintf(fp,"出力文字コード,%s\n",charcode_out);
#endif
  conv_fprintf(fp,"EMメモリ節約モード,%d\n",mode_nomem);
  conv_fprintf(fp,"順位付けモード,%d\n",mode_rank);
  conv_fprintf(fp,"順位付けモード(SMEM),%d\n",mode_rank_sm);
  conv_fprintf(fp,"順位付け表示属性数,%d\n",num_rank);
  conv_fprintf(fp,"順位付け表示属性数(SMEM),%d\n",num_rank_sm);
  conv_fprintf(fp,"順位付け基準,%s\n",rel_m_str);
  if (mode_auto_smooth_c == MODE_SMOOTH_AUTO) {
    conv_fprintf(fp,"クラスパラメータ用平滑化定数,AutoClass\n");
  } else if (mode_auto_smooth_c == MODE_SMOOTH_AUTO) {
    conv_fprintf(fp,"クラスパラメータ用平滑化定数,BDeu\n");
  } else {
    conv_fprintf(fp,"クラスパラメータ用平滑化定数,%.6Lf\n",smooth_c);
  }
  for (v=0; v<tuple_size; v++) {
    if (mode_auto_smooth_w[v] == MODE_SMOOTH_AUTO) {
      conv_fprintf(fp,"属性パラメータ%d用平滑化定数,AutoClass\n",v);
    } else if (mode_bdeu) {
      conv_fprintf(fp,"属性パラメータ%d用平滑化定数,BDeu\n",v);
    } else {
      conv_fprintf(fp,"属性パラメータ%d用平滑化定数,%.6Lf\n",v,smooth_w[v]);
    }
  }
  conv_fprintf(fp,"EMの最大繰返し数,%d\n",max_iterate);
  conv_fprintf(fp,"収束判定閾値,%.3Le\n",epsilon);
  conv_fprintf(fp,"パラメータ初期化方法（クラス）,%s\n",init_type_str_class);
  conv_fprintf(fp,"パラメータ初期化方法（属性）,%s\n",init_type_str_word);
  if (mode_search_num_class) {
    conv_fprintf(fp,"情報量規準,%s\n",ic_type_str);
  }
  if (num_init > 1) {
    conv_fprintf(fp,"再出発回数,%d\n",num_init);
    conv_fprintf(fp,"再出発EMの最大繰返し数,%d\n",max_init_iterate);
  }
  if (num_sm > 0) {
    conv_fprintf(fp,"分割併合操作回数,%d\n",num_sm);
    conv_fprintf(fp,"分割併合候補数の上限,%d\n",max_candidate);
    conv_fprintf(fp,"Jmerge正規化モード,%d\n",mode_norm);
    conv_fprintf(fp,"分割併合するクラス確率の閾値（0:閾値なし）,%.6Lf\n",min_param_c);
    conv_fprintf(fp,"分割併合後のクラス確率の閾値（0:閾値なし）,%.6Lf\n",min_new_param_c);
    conv_fprintf(fp,"分割併合の改善幅の下限,%.6Lf\n",min_smem_improve);
  }
  conv_fprintf(fp,"乱数の種,%d\n",rand_seed);
#ifndef NO_LDOUBLE
  conv_fprintf(fp,"Byteサイズ(double),%d\n",sizeof(double));
  conv_fprintf(fp,"Byteサイズ(long-double),%d\n",sizeof(FLOAT_TYPE));
#endif
  
  fflush(fp);
}

void print_att_list(FILE *fp)
{
  int v,i,j;
  
  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"#### ID:ATTVAL\n#### 属性値ID~, 属性値x%d\n####\n",v,v);
    for (i=0; i<num_word[v]; i++) {
      j = print_order[v][i];
      conv_fprintf(fp,"%d,%s\n",j,word_list[v][j]);
    }
  }

  fflush(fp);
}

void print_att_freq_list(FILE *fp)
{
  int v,i,j;

  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"#### ID:ATTVAL-FREQ\n#### 属性値ID~, 属性値x%d~, 頻度N(x%d)\n####\n",v,v);
    for (i=0; i<num_word[v]; i++) {
      j = print_order[v][i];
      conv_fprintf(fp,"%d,%s,%d\n",j,word_list[v][j],word_freq_list[v][j]);
    }
  }

  fflush(fp);
}

void print_case_freq_list(FILE *fp)
{
  int v,e;

  conv_fprintf(fp,"#### ID:CASE-FREQ\n#### ");
  if (mode_eval) {
    conv_fprintf(fp,"正解クラス~, ");
  }
  if (tuple_size > 2) {
    conv_fprintf(fp,"属性値x1~, x2~, ...~, x%d~, ",tuple_size);
    conv_fprintf(fp,"出現回数N(x1~,x2~,...~,x%d)\n####\n",tuple_size);
  } else {
    conv_fprintf(fp,"属性値x1~, x2~, ");
    conv_fprintf(fp,"出現回数N(x1~,x2)\n####\n");
  }

  for (e=0; e<num_entry; e++) {
    if (mode_eval) conv_fprintf(fp,"%s,",ans_class_list[freq_list[e].ans_id]);
    for (v=0; v<tuple_size; v++)
      conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
    conv_fprintf(fp,"%d\n",
	    freq_list[e].freq);
  }

  fflush(fp);
}

void print_ans_class_list(FILE *fp)
{
  int k;
  conv_fprintf(fp,"#### ID:ANS-CLASS\n#### 正解クラス（評価用）番号~, "
	       "名称 \n####\n");
  for (k=0; k<num_ans_class; k++) {
    conv_fprintf(fp,"%d,%s\n",k,ans_class_list[k]);
  }
}

void print_smooth(FILE *fp, int k)
{
  int v;

  conv_fprintf(fp,"#### ID:AUTO-SMOOTH-CLASS-K%d\n#### クラス数~, クラス平滑化定数 \n####\n",k);
  conv_fprintf(fp,"%d,%.12Lf\n",k,smooth_c);

  conv_fprintf(fp,"#### ID:AUTO-SMOOTH-ATT-K%d\n#### クラス数~, 属性ID~, 属性平滑化定数 \n####\n",k);
  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"%d,%d,%.12Lf\n",k,v,smooth_w[v]);
  }
}

void print_size(FILE *fp)
{
  int k,v;
  int total_num_word;

  conv_fprintf(fp,"#### ID:VALUESIZE\n#### 属性ID~, 属性値数 \n####\n");
  for (v=0; v<tuple_size; v++)
    conv_fprintf(fp,"%d,%d\n",v,num_word[v]);

  conv_fprintf(fp,"#### ID:DATASIZE\n#### 総事例数~, エントリー数（入力）~, エントリー数（重複数え上げ後） \n####\n");
  conv_fprintf(fp,"%d,%d,%d\n",total_freq,num_entry_pre,num_entry);

  conv_fprintf(fp,"#### ID:PARAMSIZE\n#### クラス数~, クラスパラメータ数~, ");
  if (tuple_size > 2) {
    conv_fprintf(fp,"属性1パラメータ数~, 属性2パラメータ数~, ...~, 属性%dパラメータ数~, ",(tuple_size+1));
  } else {
    conv_fprintf(fp,"属性1パラメータ数~, 属性2パラメータ数~, ",(tuple_size+1));
  }
  conv_fprintf(fp,"総パラメータ数 \n####\n");
  if (mode_search_num_class) {
    for (k = kmin; k <= kmax; k += kstep) {
      conv_fprintf(fp,"%d,%d",k,(k-1));
      for (v=0; v<tuple_size; v++)
	conv_fprintf(fp,"%d,",k*(num_word[v] - 1));
      total_num_word = 0;
      for (v=0; v<tuple_size; v++)
	total_num_word += num_word[v];
      conv_fprintf(fp,"%d\n",(k*(total_num_word - 1) - 1));
    }
  } else {
    k = num_class;
    conv_fprintf(fp,"%d,%d,",k,(k-1));
    for (v=0; v<tuple_size; v++)
      conv_fprintf(fp,"%d,",k*(num_word[v] - 1));
    total_num_word = 0;
    for (v=0; v<tuple_size; v++)
      total_num_word += num_word[v];
    conv_fprintf(fp,"%d\n",(k*(total_num_word - 1) - 1));
  }

  fflush(fp);
}

void print_param(FILE *fp)
{
  int k,v,i,j;

  conv_fprintf(fp,"#### ID:PARAM-CLASS\n#### クラスk~, クラスパラメータP(k)\n####\n");
  for (k=0; k<num_class; k++) {
    conv_fprintf(fp,"%d,%.12Lf\n",k,param_c[k]);
  }

  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"#### ID:PARAM-ATT%d\n#### クラスk~, 属性x%d~, 属性パラメータP(x%d|k)\n####\n",v,v,v);

    for (k=0; k<num_class; k++) {
      for (i=0; i<num_word[v]; i++) {
        j = print_order[v][i];
	conv_fprintf(fp,"%d,%s,%.12Lf\n",k,word_list[v][j],param_w[k][v][j]);
      }
    }
  }

  fflush(fp);
}

void print_complete_prob(FILE *fp)
{
  int k,v,e;

  conv_fprintf(fp,"#### ID:COMP\n#### クラスc~, ");
  for (v=0; v<tuple_size; v++)
    conv_fprintf(fp,"属性x%d~, ",v);
  if (tuple_size > 2) {
    conv_fprintf(fp,"完全データ確率P(k,x1~,x2~,...~,x%d)\n####\n",tuple_size);
  } else {
    conv_fprintf(fp,"完全データ確率P(k,x1~,x2)\n####\n");
  }

  for (k=0; k<num_class; k++) {
    for (e=0; e<num_entry; e++) {
      conv_fprintf(fp,"%d,",k);
      for (v=0; v<tuple_size; v++)
	conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
      conv_fprintf(fp,"%.12Lf\n",prob_comp[k][e]);
    }
  }
  fflush(fp);
}

void print_incomplete_prob(FILE *fp)
{
  int v,e;

  conv_fprintf(fp,"#### ID:INCOMP\n#### ");
  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"属性x%d~, ",v);
  }
  if (tuple_size > 2) {
    conv_fprintf(fp,"不完全データ確率P(x1~,x2~,...~,x%d)\n####\n",tuple_size);
  } else {
    conv_fprintf(fp,"不完全データ確率P(x1~,x2)\n####\n");
  }

  for (e=0; e<num_entry; e++) {
    for (v=0; v<tuple_size; v++)
      conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
    conv_fprintf(fp,"%.12Lf\n",prob_incomp[e]);
  }

  fflush(fp);
}

void print_conditional_prob(FILE *fp)
{
  int k,e,v;

  conv_fprintf(fp,"#### ID:COND\n#### ");
  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"属性x%d~, ",v);
  }
  if (tuple_size > 2) {
    conv_fprintf(fp,"クラスc~, 条件付き確率P(k|x1~,x2~,...~,x%d)\n####\n",tuple_size);
  } else {
    conv_fprintf(fp,"クラスc~, 条件付き確率P(k|x1~,x2)\n####\n");
  }

  for (e=0; e<num_entry; e++) {
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size; v++)
	conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
      conv_fprintf(fp,"%d,%.12Lf\n",k,prob_cond[k][e]);
    }
  }

  fflush(fp);
}

void print_expectation(FILE *fp)
{
  int k,v,i,j;

  conv_fprintf(fp,"#### ID:EXPECT-CLASS\n#### クラスk~, クラス期待値E[k]\n####\n");
  for (k=0; k<num_class; k++) {
    conv_fprintf(fp,"%d,%.12Lf\n",k,expect_c[k]);
  }

  for (v=0; v<tuple_size; v++) {

    conv_fprintf(fp,"#### ID:EXPECT-ATT%d\n#### クラスk~, 属性x%d~, 属性期待値E[x%d|k]\n####\n",v,v,v);
    for (k=0; k<num_class; k++) {
      for (i=0; i<num_word[v]; i++) {
        j = print_order[v][i];
	conv_fprintf(fp,"%d,%s,%.12Lf\n",k,word_list[v][j],expect_w[k][v][j]);
      }
    }
  }

  fflush(fp);
}

void print_member_prob(FILE *fp)
{
  int k,v,i,j,e;

  conv_fprintf(fp,"#### ID:MEMBER-CASE\n#### 事例ID~, ");
  if (mode_eval) conv_fprintf(fp,"正解クラス~, ");
  if (tuple_size > 2) {
    conv_fprintf(fp,"属性x1~, ...~, 属性x%d~, ",tuple_size);
  }
  else {
    conv_fprintf(fp,"属性x1~, 属性x2~, ");
  }
  conv_fprintf(fp,"クラスk~, 帰属確率P(k|x)\n####\n");
  for (e=0; e<num_entry; e++) {
    for (k=0; k<num_class; k++) {
      conv_fprintf(fp,"%d,",e);
      if (mode_eval) conv_fprintf(fp,"%s,",ans_class_list[freq_list[e].ans_id]);
      for (v=0; v<tuple_size; v++)
	conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
      conv_fprintf(fp,"%d,",k);
      conv_fprintf(fp,"%.12Lf\n",prob_case_mem[k][e]);
    }
  }

  for (v=0; v<tuple_size; v++) {

    conv_fprintf(fp,"#### ID:MEMBER-ATT%d\n#### 属性x%d~, クラスk~, 帰属確率分布P(k|x%d)\n####\n",v,v,v);

    for (i=0; i<num_word[v]; i++) {
      j = print_order[v][i];
      for (k=0; k<num_class; k++)
	conv_fprintf(fp,"%s,%d,%.12Lf\n",word_list[v][j],k,prob_att_mem[k][v][j]);
    }
  }

  fflush(fp);
}

void print_cluster(FILE *fp)
{
  int k,t,i,v,e;

  switch (cl_type) {

  case CL_MP:
    conv_fprintf(fp,"#### ID:CLUSTER-CASE-BY-CASE\n#### 事例ID~, クラスターk~, ");
    if (mode_eval) conv_fprintf(fp,"正解クラス~, ");
    if (tuple_size > 2) {
      conv_fprintf(fp,"属性x1~, ...~, 属性x%d~, ",tuple_size);
    }
    else {
      conv_fprintf(fp,"属性x1~, 属性x2~, ");
    }
    conv_fprintf(fp,"帰属確率P(k|x)\n####\n");
    for (e=0; e<num_entry; e++) {
      k = case_cluster_pred[e];
      conv_fprintf(fp,"%d,",e);
      if (mode_eval) conv_fprintf(fp,"%s,",ans_class_list[freq_list[e].ans_id]);
      conv_fprintf(fp,"%d,",k);
      for (v=0; v<tuple_size; v++)
	conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
      conv_fprintf(fp,"%.12Lf\n",prob_case_mem[k][e]);
    }
    conv_fprintf(fp,"#### ID:CLUSTER-CASE-BY-CLUSTER\n#### クラスターk~, 事例ID~, ");
    if (mode_eval) conv_fprintf(fp,"正解クラス~, ");
    if (tuple_size > 2) {
      conv_fprintf(fp,"属性x1~, ...~, 属性x%d~, ",tuple_size);
    }
    else {
      conv_fprintf(fp,"属性x1~, 属性x2~, ");
    }
    conv_fprintf(fp,"帰属確率P(k|x)\n####\n");
    for (k=0; k<num_class; k++) {
      for (t=0; t<case_cluster_freq[k]; t++) {
        e = case_cluster[k][t];
	conv_fprintf(fp,"%d,%d,",k,e);
	if (mode_eval) conv_fprintf(fp,"%s,",ans_class_list[freq_list[e].ans_id]);
        for (v=0; v<tuple_size; v++)
          conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
        conv_fprintf(fp,"%.12Lf\n",prob_case_mem[k][e]);
      }
    }
    break;
    
  case CL_WODD:
    conv_fprintf(fp,"#### ID:CLUSTER-CASE-BY-CASE\n#### 事例ID~, ");
    if (mode_eval) conv_fprintf(fp,"正解クラス~, ");
    conv_fprintf(fp,"クラスターk~, ");
    if (tuple_size > 2) {
      conv_fprintf(fp,"属性x1~, ...~, 属性x%d~, ",tuple_size);
    }
    else {
      conv_fprintf(fp,"属性x1~, 属性x2~, ");
    }
    conv_fprintf(fp,"帰属確率P(k|x)~, 重み付き尤度比P(k|x)log{P(k|x)/P(k|not x)}\n####\n");
    for (e=0; e<num_entry; e++) {
      k = case_cluster_pred[e];
      conv_fprintf(fp,"%d,",e);
      if (mode_eval) conv_fprintf(fp,"%s,",ans_class_list[freq_list[e].ans_id]);
      conv_fprintf(fp,"%d,",k);
      for (v=0; v<tuple_size; v++)
	conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
      conv_fprintf(fp,"%.12Lf,%.12Lf\n",
		   prob_case_mem[k][e],wodd_case[k][e]);
    }
    conv_fprintf(fp,"#### ID:CLUSTER-CASE-BY-CLUSTER\n#### クラスターk~, 事例ID~, ");
    if (mode_eval) conv_fprintf(fp,"正解クラス~, ");
    if (tuple_size > 2) {
      conv_fprintf(fp,"属性x1~, ...~, 属性x%d~, ",tuple_size);
    }
    else {
      conv_fprintf(fp,"属性x1~, 属性x2~, ");
    }
    conv_fprintf(fp,"帰属確率P(k|x)~, 重み付き尤度比P(k|x)log{P(k|x)/P(k|not x)}\n####\n");
    for (k=0; k<num_class; k++) {
      for (t=0; t<case_cluster_freq[k]; t++) {
        e = case_cluster[k][t];
	conv_fprintf(fp,"%d,%d,",k,e);
	if (mode_eval) conv_fprintf(fp,"%s,",ans_class_list[freq_list[e].ans_id]);
        for (v=0; v<tuple_size; v++)
          conv_fprintf(fp,"%s,",word_list[v][freq_list[e].id[v]]);
        conv_fprintf(fp,"%.12Lf,%.12Lf\n",
		     prob_case_mem[k][e],wodd_case[k][e]);
      }
    }
    break;
    
  default:
    die("Internal error: Unsupported clustering method\n");
  }
  
  for (v=0; v<tuple_size; v++) {
    switch (cl_type) {
      
    case CL_MP:
      conv_fprintf(fp,"#### ID:CLUSTER-ATT%d-BY-ATT\n#### クラスターk~, 属性x%d~, 帰属確率P(k|x%d)\n####\n",v,v,v,v);
      for (i=0; i<num_word[v]; i++) {
	k = att_cluster_pred[v][i];
	conv_fprintf(fp,"%d,%s,%.12Lf\n",
		     k,word_list[v][i],prob_att_mem[k][v][i]);
      }
      conv_fprintf(fp,"#### ID:CLUSTER-ATT%d-BY-CLUSTER\n#### クラスターk~, 属性x%d~, 帰属確率P(k|x%d)\n####\n",v,v,v,v);
      for (k=0; k<num_class; k++) {
	for (t=0; t<att_cluster_freq[v][k]; t++) {
	  i = att_cluster[v][k][t];
	  conv_fprintf(fp,"%d,%s,%.12Lf\n",
		       k,word_list[v][i],prob_att_mem[k][v][i]);
	}
      }
      break;
      
    case CL_WODD:
      conv_fprintf(fp,"#### ID:CLUSTER-ATT%d-BY-ATT\n#### クラスターk~, 属性x%d~, 帰属確率P(k|x%d)~, 重み付き尤度比P(k|x%d)log{P(k|x%d)/P(k|not x%d)}\n####\n",v,v,v,v,v,v);
      for (i=0; i<num_word[v]; i++) {
	k = att_cluster_pred[v][i];
	conv_fprintf(fp,"%d,%s,%.12Lf,%.12Lf\n",
		     k,word_list[v][i],prob_att_mem[k][v][i],wodd_att[k][v][i]);
      }
      conv_fprintf(fp,"#### ID:CLUSTER-ATT%d-BY-CLUSTER\n#### クラスターk~, 属性x%d~, 帰属確率P(k|x%d)~, 重み付き尤度比P(k|x%d)log{P(k|x%d)/P(k|not x%d)}\n####\n",v,v,v,v,v,v);
      for (k=0; k<num_class; k++) {
	for (t=0; t<att_cluster_freq[v][k]; t++) {
	  i = att_cluster[v][k][t];
	  conv_fprintf(fp,"%d,%s,%.12Lf,%.12Lf\n",
		       k,word_list[v][i],prob_att_mem[k][v][i],wodd_att[k][v][i]);
	}
      }
      break;
      
    default:
      die("Internal error: Unsupported clustering method\n");
    }
  }
}

void print_cluster_freq(FILE *fp)
{
  int k,v;
  
  conv_fprintf(fp,"#### ID:CLUSTER-FREQ-CASE\n#### 事例に対するクラスター~, サイズ \n####\n");
  for (k=0; k<num_class; k++) {
    conv_fprintf(fp,"%d,%d\n",k,case_cluster_freq[k]);
  }
  
  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"#### ID:CLUSTER-FREQ-ATT%d\n#### 属性%dに対するクラスター~, サイズ \n####\n",v,v);
    for (k=0; k<num_class; k++) {
      conv_fprintf(fp,"%d,%d\n",k,att_cluster_freq[v][k]);
    }
  }

  fflush(fp);
}

void print_rank_ml(FILE *fp)
{
  int k,k0,i,j,v,e;
  int num_show_case_rank;
  int *num_show_att_rank;

  num_show_att_rank = (int *)malloc(sizeof(int)*tuple_size);
  if (num_show_att_rank == NULL)
    die("Couldn't allocate memory for num_show_att_rank\n");
  
  num_show_case_rank = (num_rank > 0 && num_rank < num_entry) ? num_rank : num_entry;
  for (v=0; v<tuple_size; v++)
    num_show_att_rank[v] = (num_rank > 0 && num_rank < num_word[v]) ? num_rank : num_word[v];
  
  conv_fprintf(fp,"#### ID:RANK-MP-CASE\n#### クラスk~, 事例ID~, ");
  if (mode_eval) conv_fprintf(fp,"正解クラス~, ");
  if (tuple_size > 2) {
    conv_fprintf(fp,"属性x1~, ...~, 属性x%d~, ",tuple_size);
  }
  else {
    conv_fprintf(fp,"属性x1~, 属性x2~, ");
  }
  conv_fprintf(fp,"クラス確率P(k)~, 帰属確率P(k|x)\n#### （クラス確率，帰属確率で順序付け） \n####\n");
  
  for (k=0; k<num_class; k++) {
    k0 = rank_class[k];
    for (e=0; e<num_show_case_rank; e++) {
      j = rank_case_mem[k0][e];
      conv_fprintf(fp,"%d,%d,",k0,j);
      if (mode_eval) conv_fprintf(fp,"%s,",ans_class_list[freq_list[j].ans_id]);
      for (v=0; v<tuple_size; v++)
	conv_fprintf(fp,"%s,",word_list[v][freq_list[j].id[v]]);
      conv_fprintf(fp,"%.12Lf,%.12Lf\n",
		   param_c[k0],prob_case_mem[k0][j]);
    }
  }
  
  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"#### ID:RANK-MP-ATT%d\n#### クラスk~, 属性x%d~, クラス確率P(k)~, 帰属確率分布P(k|x%d)\n#### （クラス確率，帰属確率で順序付け） \n####\n",v,v,v);
    for (k=0; k<num_class; k++) {
      k0 = rank_class[k];
      for (i=0; i<num_show_att_rank[v]; i++) {
	j = rank_att_mem[k0][v][i];
	conv_fprintf(fp,"%d,%s,%.12Lf,%.12Lf\n",
		     k0,word_list[v][j],param_c[k0],prob_att_mem[k0][v][j]);
      }
    }
  }
  
  free(num_show_att_rank);
  
  fflush(fp);
}

void print_rank_wodd(FILE *fp)
{
  int k,k0,i,j,v,e;
  int num_show_case_rank;
  int *num_show_att_rank;
  
  num_show_att_rank = (int *)malloc(sizeof(int)*tuple_size);
  if (num_show_att_rank == NULL)
    die("Couldn't allocate memory for num_show_att_rank\n");
  
  num_show_case_rank = (num_rank > 0 && num_rank < num_entry) ? num_rank : num_entry;
  for (v=0; v<tuple_size; v++)
    num_show_att_rank[v] = (num_rank > 0 && num_rank < num_word[v]) ? num_rank : num_word[v];
  
  conv_fprintf(fp,"#### ID:RANK-WODD-CASE\n#### クラスk~, 事例ID~, ");
  if (mode_eval) conv_fprintf(fp,"正解クラス~, ");
  if (tuple_size > 2) {
    conv_fprintf(fp,"属性x1~, ...~, 属性x%d~, ",tuple_size);
  }
  else {
    conv_fprintf(fp,"属性x1~, 属性x2~, ");
  }
  conv_fprintf(fp,"クラス確率P(k)~, 帰属確率P(k|x)~, "
	       "重み付き対数尤度比 P(x|k)log{P(x|k)/P(x|not k)}\n"
	       "#### （クラス確率と重み付き対数尤度比で順序付け） \n####\n");
  for (k=0; k<num_class; k++) {
    k0 = rank_class[k];
    for (e=0; e<num_show_case_rank; e++) {
      j = rank_case_mem[k0][e];
      conv_fprintf(fp,"%d,%d,",k0,j);
      if (mode_eval) conv_fprintf(fp,"%s,",ans_class_list[freq_list[j].ans_id]);
      for (v=0; v<tuple_size; v++)
	conv_fprintf(fp,"%s,",word_list[v][freq_list[j].id[v]]);
      conv_fprintf(fp,"%.12Lf,%.12Lf,%.12Lf\n",
		   param_c[k0],prob_case_mem[k0][j],wodd_case[k0][j]);
    }
  }
  
  for (v=0; v<tuple_size; v++) {
    conv_fprintf(fp,"#### ID:RANK-WODD-ATT%d\n#### クラスk~, 属性x%d~, "
		 "クラス確率P(k)~, 帰属確率P(k|x%d)~, "
		 "重み付き尤度比P(x%d|k)log{P(x%d|k)/P(x%d|not k)}\n"
		 "#### （クラス確率と重み付き対数尤度比で順序付け） \n####\n",
		 v,v,v,v,v,v);
    for (k=0; k<num_class; k++) {
      k0 = rank_class[k];
      for (i=0; i<num_show_att_rank[v]; i++) {
	j = rank_att_mem[k0][v][i];
	conv_fprintf(fp,"%d,%s,%.12Lf,%.12Lf,%.12Lf\n",
		     k0,word_list[v][j],param_c[k0],prob_att_mem[k0][v][j],wodd_att[k0][v][j]);
      }
    }
  }
  
  free(num_show_att_rank);
  
  fflush(fp);
}

void print_rank(FILE *fp)
{
  switch (rel_m) {
  case REL_MP:
    print_rank_ml(fp);
    break;
  case REL_WODD:
    print_rank_wodd(fp);
    break;
  default:
    fprintf(stderr,"Internal error: Unexpected relevance measure between class and word\n");
    exit(1);
  }
}

void print_rank_sm(FILE *fp,int t,int *cand)
{
  int k0,k1,k2,k3,i,j,l,v;
  int *num_show_rank;

  num_show_rank = (int *)malloc(sizeof(int)*tuple_size);
  if (num_show_rank == NULL)
    die("Couldn't allocate memory for num_show_rank\n");

  for (v=0; v<tuple_size; v++)
    num_show_rank[v] =
      (num_rank_sm > 0 && num_rank_sm < num_word[v]) ? num_rank_sm : num_word[v];

  k1 = cand[0];
  k2 = cand[1];
  k3 = cand[2];

  for (v=0; v<tuple_size; v++) {

    conv_fprintf(fp,"#### ID:RANK-SMEM-BEFORE-ATT%d\n",v);
    conv_fprintf(fp,"#### クラスc~, 属性x%d~, クラス確率P(k)~, 帰属確率分布P(k|x%d)\n#### （帰属確率で順序付け） \n",v,v);
    conv_fprintf(fp,"#### 分割併合操作 %d: [%d~,%d] -> [%d]~, [%d] -> [%d~,%d]\n####\n",
	    t,k1,k2,k1,k3,k2,k3);
    
    for (l=0; l<3; l++) {
      k0 = cand[l];
      for (i=0; i<num_show_rank[v]; i++) {
	j = rank_att_mem_sm[k0][v][i];
	conv_fprintf(fp,"%d,%s,%.12Lf,%.12Lf\n",
		k0,word_list[v][j],param_c_sm[k0],prob_att_mem_sm[k0][v][j]);
      }
    }

    conv_fprintf(fp,"#### ID:RANK-SMEM-AFTER-ATT%d\n",v);
    conv_fprintf(fp,"#### クラスc~, 属性x%d~, クラス確率P(k)~, 帰属確率分布P(k|x%d)\n#### （帰属確率で順序付け） \n",v,v);
    conv_fprintf(fp,"#### 分割併合操作 %d: [%d~,%d] -> [%d]~, [%d] -> [%d~,%d]\n####\n",
	    t,k1,k2,k1,k3,k2,k3);

    for (l=0; l<3; l++) {
      k0 = cand[l];
      for (i=0; i<num_show_rank[v]; i++) {
	j = rank_att_mem_part[l][v][i];
	conv_fprintf(fp,"%d,%s,%.12Lf,%.12Lf\n",
		k0,word_list[v][j],param_c[k0],prob_att_mem_part[l][v][j]);
      }
    }
  }

  free(num_show_rank); num_show_rank = NULL;

  fflush(fp);
}

void print_case_kl_dist_header(FILE *fp)
{
  if (!mode_eval)
    conv_fprintf(fp,"#### ID:DISSIM-KL-CASE\n#### 事例ID(x)~, 事例ID(x')~, "
		 "KL距離D(x||x')\n####\n");
  else
    conv_fprintf(fp,"#### ID:DISSIM-KL-CASE\n#### 事例ID(x)~, 正解クラス(x)~, "
		 "事例ID(x')~, 正解クラス(x')~, KL距離D(x||x')\n####\n");
}

void print_att_kl_dist_header(FILE *fp, int v)
{
  conv_fprintf(fp,"#### ID:DISSIM-KL-ATT\n#### 属性 x%d~, 属性 x%d'~, KL距離 D(x%d||x%d')\n####\n",(v+1),(v+1),(v+1),(v+1));
}

void print_case_kl_dist(FILE *fp, int e0, int e1, FLOAT_TYPE d)
{
  if (!mode_eval)
    conv_fprintf(fp,"%d,%d,%.12Lf\n",e0,e1,d);
  else
    conv_fprintf(fp,"%d,%s,%d,%s,%.12Lf\n",
	    e0,ans_class_list[freq_list[e0].ans_id],
	    e1,ans_class_list[freq_list[e1].ans_id],
	    d);
}

void print_att_kl_dist(FILE *fp, int v, int i, int j, FLOAT_TYPE d)
{
  conv_fprintf(fp,"%s,%s,%.12Lf\n",word_list[v][i],word_list[v][j],d);
}

#endif

