#include "nbc_def.h"
#include "nbc_ext.h"

void rank_main(void)
{
  FILE *fp_rank;

  allocate_rank();
  compute_rank();
  
  fp_rank = fopen(fname_rank,"wb");
  if (fp_rank == NULL)
    die("Couldn't open %s\n",fname_rank);

  print_rank(fp_rank);
  fclose(fp_rank);

  release_rank();
}

int compare_prob_case_mem(const int *x, const int *y)
{
  int ix = *x;
  int iy = *y;

  FLOAT_TYPE px = tmp_rel_m[ix];
  FLOAT_TYPE py = tmp_rel_m[iy];

  if (px < py)  return 1;
  if (px > py)  return -1;

  return 0;
}

int compare_prob_att_mem(const int *x, const int *y)
{
  int ix = *x;
  int iy = *y;

  FLOAT_TYPE px = tmp_rel_m[ix];
  FLOAT_TYPE py = tmp_rel_m[iy];

  if (px < py)  return 1;
  if (px > py)  return -1;

  return 0;
}

int compare_param_c(const int *x, const int *y)
{
  int ix = *x;
  int iy = *y;

  FLOAT_TYPE px = tmp_param_c[ix];
  FLOAT_TYPE py = tmp_param_c[iy];

  if (px < py) return 1;
  if (px > py) return -1;

  return 0;
}

void compute_rank(void)
{
  compute_rank_case();
  compute_rank_att();
}

void compute_rank_case(void)
{
  int k,v,e;
  FLOAT_TYPE p_neg;
  FLOAT_TYPE u0;
  FLOAT_TYPE u1;
  FLOAT_TYPE p_ic,p_c,p_x;

  for (k=0; k<num_class; k++) {
    rank_class[k] = k;
  }
  for (k=0; k<num_class; k++)
    for (e=0; e<num_entry; e++)
      rank_case_mem[k][e] = e;

  tmp_param_c = param_c;
  qsort(rank_class,num_class,sizeof(int),
	(int(*)(const void *,const void *))compare_param_c);

  switch (rel_m) {
  case REL_MP:
    for (k=0; k<num_class; k++) {
      tmp_rel_m = prob_case_mem[k];
      qsort(rank_case_mem[k],num_entry,sizeof(int),
	    (int(*)(const void *, const void *))compare_prob_case_mem);
    }
    break;
  case REL_WODD:
    for (e=0; e<num_entry; e++) {
      p_ic = 0.0;

      for (v=0; v<tuple_size; v++) {
	freq_w_id[v] = freq_list[e].id[v];
      }

      for (k=0; k<num_class; k++) {
	p_c = param_c[k];
	for (v=0; v<tuple_size; v++)
	  p_c *= param_w[k][v][freq_w_id[v]];
	p_ic += p_c;
      }

      for (k=0; k<num_class; k++) {
	p_x = 1.0;
	for (v=0; v<tuple_size; v++)
	  p_x *= param_w[k][v][freq_w_id[v]];

	p_c = param_c[k] * p_x;

	u0 = p_ic - p_c;
	u1 = 1.0 - param_c[k];
	p_neg = u0/u1;
	
	wodd_case[k][e] = (p_x < TINY_FLOAT) ? 0.0 : p_x * MATH_LOG(p_x / p_neg);
      }
    }

    for (k=0; k<num_class; k++) {
      tmp_rel_m = wodd_case[k];
      qsort(rank_case_mem[k],num_entry,sizeof(int),
	    (int(*)(const void *, const void *))compare_prob_case_mem);
    }
    break;
  default:
    fprintf(stderr,"Internal error: Unexpected relevance measure between class and word\n");
    exit(1);
  }
}

void compute_rank_att(void)
{
  int k,v,i;
  FLOAT_TYPE p_neg;
  FLOAT_TYPE u0;
  FLOAT_TYPE u1;

  for (k=0; k<num_class; k++) {
    rank_class[k] = k;
  }
  for (k=0; k<num_class; k++)
    for (v=0; v<tuple_size; v++)
      for (i=0; i<num_word[v]; i++)
	rank_att_mem[k][v][i] = i;

  tmp_param_c = param_c;
  qsort(rank_class,num_class,sizeof(int),
	(int(*)(const void *,const void *))compare_param_c);

  switch (rel_m) {
  case REL_MP:
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size; v++) {
	tmp_rel_m = prob_att_mem[k][v];
	qsort(rank_att_mem[k][v],num_word[v],sizeof(int),
	      (int(*)(const void *, const void *))compare_prob_att_mem);
      }
    }
    break;
  case REL_WODD:
    for (k=0; k<num_class; k++) {
      for (v=0; v<tuple_size; v++) {
	tmp_rel_m = wodd_att[k][v];
	for (i=0; i<num_word[v]; i++) {
	  u0 = prob_w[v][i] - param_c[k]*param_w[k][v][i] + TINY_FLOAT;
	  u1 = 1.0 - param_c[k] + TINY_FLOAT;
	  p_neg = u0/u1;
	  if (param_w[k][v][i] < TINY_FLOAT)
	    tmp_rel_m[i] = 0;  /* We assume that 0 log 0 = 0 */
	  else
	    tmp_rel_m[i] = param_w[k][v][i] * MATH_LOG(param_w[k][v][i] / p_neg);
	}
	qsort(rank_att_mem[k][v],num_word[v],sizeof(int),
	      (int(*)(const void *, const void *))compare_prob_att_mem);
      }
    }
    break;
  default:
    fprintf(stderr,"Internal error: Unexpected relevance measure between class and word\n");
    exit(1);
  }
}

void compute_rank_sm(void)
{
  int k,v,i;

  for (k=0; k<num_class; k++)
    for (v=0; v<tuple_size; v++) 
      for (i=0; i<num_word[v]; i++)
	rank_att_mem_sm[k][v][i] = i;

  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      tmp_rel_m = prob_att_mem_sm[k][v];
      qsort(rank_att_mem_sm[k][v],num_word[v],
	    sizeof(int),(int(*)(const void *, const void *))compare_prob_att_mem);
    }
  }
}

void compute_rank_part(int *cand)
{
  int l,v,i;

  for (l=0; l<3; l++)
    for (v=0; v<tuple_size; v++)
      for (i=0; i<num_word[v]; i++)
	rank_att_mem_part[l][v][i] = i;

  for (l=0; l<3; l++) {
    for (v=0; v<tuple_size; v++) {
      tmp_rel_m = prob_att_mem_part[l][v];
      qsort(rank_att_mem_part[l][v],num_word[v],
	    sizeof(int),(int(*)(const void *,const void *))compare_prob_att_mem);
    }
  }
}
