#include "nbc_def.h"
#include "nbc_ext.h"

/* 
 * Count the number of non-comment lines
 * - this is not safe since we should not scan the file relying on '#'.
 */
int get_number_of_entries(void)
{
  FILE *fp_freq;
  int c, pre_c;
  int ignore;
  int num_entry;

  fp_freq = fopen(fname_freq,"rb");
  if (fp_freq == NULL) {
    die("Couldn't open %s\n",fname_freq);
  }
  
  num_entry = 0;
  ignore = 0;
  pre_c = '\n';
  while ((c = getc(fp_freq)) != EOF) {
    if (pre_c == '\n' && c == '#') ignore = 1;
    if (c == '\n') {
      if (!ignore) num_entry++;
      ignore = 0;
    }
    pre_c = c;
  }

  fclose(fp_freq);

  return num_entry;
}

/*
 * [CSV form]
 *   - does not understand the double quatation character
 *   - allows the space/tab characters
 *   - ignores the lines starting with '#'
 *
 * [tab-separated from]
 *   - does not understand the double quatation character
 *   - allows the space characters
 *   - regards the consecutive tab characteres as one tab character
 *   - ignores the lines starting with '#'
 */
void read_freq_body(void)
{
  FILE *fp_freq;
  char *buf, **word;
  char str_freq[MAX_FREQ_LEN + 1];
  char attr_vec[MAX_LINE_LEN + 1];
  char *ans_class = NULL;
  int freq;
  char *b, *q;
  int line_c;
  int c,v;
  char *endptr;
  char *buf_conv = NULL;
  char **word_conv = NULL;
#ifdef USE_ICONV
  char *bb;
  char **ww = NULL;
  size_t buf_len_in, buf_len_out;
  size_t *w_len_in;
  size_t *w_len_out;
#ifdef ICONV_CONST
  const char **word_tmp = NULL;
  const char *buf_tmp = NULL;
#else
  char **word_tmp = NULL;
  char *buf_tmp = NULL;
#endif
#endif

  fp_freq = fopen(fname_freq,"rb");
  if (fp_freq == NULL)
    die("Couldn't open %s\n", fname_freq);

  buf = (char *)malloc(sizeof(char)*(MAX_LINE_LEN + 1));
  if (buf == NULL)
    die("Couldn't allocate memory for buf\n");

  word = (char **)malloc(sizeof(char *)*tuple_size);
  if (word == NULL)
    die("Couldn't allocate memory for word\n");

  for (v=0; v<tuple_size; v++) {
    word[v] = (char *)malloc(sizeof(char)*(MAX_WORD_LEN + 1));
    if (word[v] == NULL)
      die("Couldn't allocate memory for word[v]\n");
  }

  if (mode_conv_in) {
    buf_conv = (char *)malloc(sizeof(char)*(MAX_LINE_LEN + 1));
    if (buf_conv == NULL)
      die("Couldn't allocate memory for buf_conv\n");
  }

  if (mode_eval) {
    ans_class = (char *)malloc(sizeof(char)*(MAX_WORD_LEN + 1));
    if (ans_class == NULL)
      die("Couldn't allocate memory for ans_class\n");
  }

#ifdef USE_ICONV
  w_len_in = (size_t *)malloc(sizeof(size_t)*tuple_size);
  w_len_out = (size_t *)malloc(sizeof(size_t)*tuple_size);
  ww = (char **)malloc(sizeof(char *)*tuple_size);
  if (w_len_in == NULL || w_len_out == NULL || ww == NULL)
    die("Couldn't allocate memory for w_len_{in,out}, ww\n");

#ifdef ICONV_CONST
  word_tmp = (const char **)malloc(sizeof(const char *)*tuple_size);
#else
  word_tmp = (char **)malloc(sizeof(char *)*tuple_size);
#endif

#endif

  word_conv = (char **)malloc(sizeof(char *)*tuple_size);
  if (word_conv == NULL)
    die("Couldn't allocate memory for word_conv\n");
  
  for (v=0; v<tuple_size; v++) {
    word_conv[v] = (char *)malloc(sizeof(char)*(MAX_WORD_LEN + 1));
    if (word_conv[v] == NULL)
      die("Couldn't allocate memory for word_conv[v]\n");
  }

  line_c = 0;
  while (fgets(buf,MAX_LINE_LEN,fp_freq) != NULL) {
    line_c++;

#ifdef USE_ICONV
    /* converted into the internal code if necessary */
    if (mode_conv_in) {
      buf_tmp = buf;
      bb = buf_conv;
      buf_len_in  = strlen(buf)+1;
      buf_len_out = MAX_LINE_LEN+1;
      if (iconv(cd_in, &buf_tmp, &buf_len_in, &bb, &buf_len_out) == (size_t)-1)
	die("[line %d] Couldn't convert to the internal charcode (errno=%d)\n",
	    line_c,errno);
    } else {
      buf_conv = buf;
    }
#else
    buf_conv = buf;
#endif

    /**** tab-separated form ****/
    if (!mode_csv_in) {

      b = buf_conv;

      if (*b == '#') continue;

      while (*b == ' ') b++;
      
      if (mode_eval) {
	c = 0;
	q = ans_class;
	while (*b != ' ' && *b != '\t' && *b != '\r' &&
	       *b != '\n' && *b != '\0') {
	  if (c > MAX_WORD_LEN) die_line(line_c, __LINE__, buf_conv);
	  *q = *b;
	  q++; b++; c++;
	}
	*q = '\0';
	
	while (*b == ' ') b++;
	if (*b != '\t') die_line(line_c, __LINE__, buf_conv);
	b++; v++;
	while (*b == '\t') b++;
	while (*b == ' ') b++;
      }
      else {
	ans_class = NULL;
      }

      freq = -1;
      v = 0;
      while (v < tuple_size) {
	c = 0;
	q = word[v];
	while (*b != ' ' && *b != '\t' && *b != '\r' &&
	       *b != '\n' && *b != '\0') {
	  if (c > MAX_WORD_LEN) die_line(line_c, __LINE__, buf_conv);
	  *q = *b;
	  q++; b++; c++;
	}
	*q = '\0';
	
	while (*b == ' ') b++;
	if (v == (tuple_size - 1) && (*b == '\r' || *b == '\n')) {
	  freq = 1;
	  break;
	}
	else if (*b != '\t') die_line(line_c, __LINE__, buf_conv);
	b++; v++;
	while (*b == '\t') b++;
	while (*b == ' ') b++;
      }

      if (freq < 0) {
	c = 0;
	q = str_freq;
	while (*b != ' ' && *b != '\t' && *b != '\r' && *b != '\n') {
	  if (!isdigit(*b)) die_line(line_c, __LINE__, buf_conv);
	  if (c > MAX_FREQ_LEN) die_line(line_c, __LINE__, buf_conv);
	  *q = *b;
	  q++; b++;
	}
	*q = '\0';
	
	while (*b == ' ') b++;
	if (*b != '\r' && *b != '\n') die_line(line_c, __LINE__, buf_conv);
	
	q = str_freq;
	while (*q == '0') q++; /* convert "00083" --> "83", for example */
	
	freq = strtol(q, &endptr, 10);
	if (*endptr != '\0') {
	  print_usage(stderr);
	  exit(1);
	}
      }
    }
    /**** CSV form ****/
    else {

      b = buf_conv;
      
      if (*b == '#') continue;

      while (*b == ' ' || *b == '\t') b++;
      
      if (mode_eval) {
	c = 0;
	q = ans_class;
	while (*b != ' ' && *b != '\t' && *b != ',' &&
	       *b != '\r' && *b != '\n' ) {
	  if (c > MAX_WORD_LEN) die_line(line_c, __LINE__, buf_conv);
	  *q = *b;
	  q++; b++; c++;
	}
	*q = '\0';
      
	while (*b == ' ' || *b == '\t') b++;
	if (*b != ',') die_line(line_c, __LINE__, buf_conv);
	b++; v++;
	while (*b == ' ' || *b == '\t') b++;
      }
      else {
	ans_class = NULL;
      }

      freq = -1;
      v = 0;
      while (v < tuple_size) {
	c = 0;
	q = word[v];
	while (*b != ' ' && *b != '\t' && *b != ',' &&
	       *b != '\r' && *b != '\n' ) {
	  if (c > MAX_WORD_LEN) die_line(line_c, __LINE__, buf_conv);
	  *q = *b;
	  q++; b++; c++;
	}
	*q = '\0';
      
	while (*b == ' ' || *b == '\t') b++;

	if (v == (tuple_size - 1) && (*b == '\r' || *b == '\n')) {
	  freq = 1;
	  break;
	}
	else if (*b != ',') die_line(line_c, __LINE__, buf_conv);
	b++; v++;
	while (*b == ' ' || *b == '\t') b++;
      }

      if (freq < 0) {
	c = 0;
	q = str_freq;
	while (*b != ' ' && *b != '\t' && *b != ',' && *b != '\r' && *b != '\n') {
	  if (!isdigit(*b)) die_line(line_c, __LINE__, buf_conv);
	  if (c > MAX_FREQ_LEN) die_line(line_c, __LINE__, buf_conv);
	  *q = *b;
	  q++; b++;
	}
	*q = '\0';
	
	while (*b == ' ' || *b == '\t') b++;
	if (*b != '\r' && *b != '\n') die_line(line_c, __LINE__, buf_conv);
	
	q = str_freq;
	while (*q == '0') q++;  /* "00083" --> "83", for example */
	
	freq = strtol(q, &endptr, 10);
	if (*endptr != '\0') {
	  print_usage(stderr);
	  exit(1);
	}
      }
    }

#ifdef USE_ICONV
    /* converted into the output code if necessary */
    if (mode_conv_out) {
      for (v=0; v<tuple_size; v++) {
	word_tmp[v] = word[v];
	ww[v] = word_conv[v];
	w_len_in[v] = strlen(word[v])+1;
	w_len_out[v] = MAX_WORD_LEN+1;
	if (iconv(cd_out,&word_tmp[v],&w_len_in[v],&ww[v],&w_len_out[v]) == (size_t)-1 ||
	    iconv(cd_out,&word_tmp[v],&w_len_in[v],&ww[v],&w_len_out[v]) == (size_t)-1)
	  die("[line %d] Couldn't convert to %s (errno=%d)\n",
	      line_c,charcode_out,errno);
      }
    } else {
      for (v=0; v<tuple_size; v++) {
	strncpy(word_conv[v],word[v],MAX_WORD_LEN);
      }
    }
#else
    for (v=0; v<tuple_size; v++) {
      strncpy(word_conv[v],word[v],MAX_WORD_LEN);
    }
#endif

    /* Construct the string representation of the attribute vector: */
    attr_vec[0] = '\0';
    for (v=0; v<tuple_size; v++) {
      if (v > 0) strcat(attr_vec,",");
      strcat(attr_vec,word_conv[v]);
    }

    register_freq(line_c,word_conv,freq,attr_vec,ans_class,buf_conv);
  }

  free(buf);   buf = NULL;

  for (v=0; v<tuple_size; v++) {
    free(word[v]); word[v] = NULL;
  }
  free(word); word = NULL;

  if (mode_conv_in) {
    free(buf_conv); buf_conv = NULL;
  }

  for (v=0; v<tuple_size; v++) {
    free(word_conv[v]); word_conv[v] = NULL;
  }
  free(word_conv); word_conv = NULL;

  if (mode_eval) {
    free(ans_class); ans_class = NULL;
  }

#ifdef USE_ICONV
  free(w_len_in);  w_len_in = NULL;
  free(w_len_out); w_len_out = NULL;
  free(ww);        ww = NULL;
#  ifndef ICONV_CONST
  free(word_tmp);  word_tmp = NULL;
#  endif
#endif

  fclose(fp_freq);
}

void read_freq(void)
{
  int e,f,v,i;

  num_entry_pre = get_number_of_entries();

  freq_list = (struct FREQ *)malloc(sizeof(struct FREQ)*num_entry_pre);
  if (freq_list == NULL)
    die("Couldn't allocate memory for freq_list\n");

  for (e=0; e<num_entry_pre; e++) {
    freq_list[e].id = (int *)malloc(sizeof(int)*tuple_size);
    if (freq_list[e].id == NULL)
      die("Couldn't allocate memory for freq_list[e].id\n");
    for (v=0; v<tuple_size; v++) {
      freq_list[e].id[v] = -1;
    }
    freq_list[e].freq = 0;
    freq_list[e].line_no = -1;
    freq_list[e].ans_id = -1;
  }

  freq_w_id = (int *)malloc(sizeof(int)*tuple_size);
  if (freq_w_id == NULL)
    die("Couldn't allocate memory for freq_w_id\n");

  /* The frequency file has lower priority than the parameter file: */
  if (mode_as_is) {
    ht = (HASHTABLE *)malloc(sizeof(HASHTABLE)*tuple_size);
    if (ht == NULL)
      die("Couldn't allocate memory for ht\n");
    
    for (v=0; v<tuple_size; v++)
      init_hashtable(&ht[v]);
  }

  init_hashtable(&ht_vec);

  if (mode_eval) {
    init_hashtable(&ht_class);
  }

  num_word = (int *)malloc(sizeof(int)*tuple_size);
  if (num_word == NULL)
    die("Couldn't allocate memory for num_word\n");

  read_freq_body();

  /* Get the number of actually distinct attribute vectors: */
  num_entry = ht_vec.serial_id;

  /* Get the total number of cases: */
  f = 0;
  for (e=0; e<num_entry; e++) {
    f += freq_list[e].freq;
  }
  total_freq = f;

  word_list = (char ***)malloc(sizeof(char **)*tuple_size);
  if (word_list == NULL)
    die("Couldn't allocate memory for word_list\n");

  for (v=0; v<tuple_size; v++) {
    num_word[v] = ht[v].serial_id;
    word_list[v] = make_entry_list(&ht[v]); 
  }

  /* Get the number of classes in the evaluation data: */
  if (mode_eval) {
    num_ans_class = ht_class.serial_id;
    ans_class_list = make_entry_list(&ht_class);
  }    

  print_order = (int **)malloc(sizeof(int *)*tuple_size);
  if (print_order == NULL)
    die("Couldn't allocate memory for print_order\n");

  for (v=0; v<tuple_size; v++) {
    print_order[v] = (int *)malloc(sizeof(int)*num_word[v]);
    for (i=0; i<num_word[v]; i++) {
      print_order[v][i] = i;
    }
    if (mode_lex_order) {
      tmp_word_list = word_list[v];
      qsort(print_order[v],num_word[v],sizeof(int),
            (int(*)(const void *,const void *))compare_word);
    }
  }

  make_word_freq_list();

  if (mode_eval)
    print_ans_class_list(fp_log);

  if (mode_verbose) {
    print_att_freq_list(fp_log);
    print_case_freq_list(fp_log);
  }
}

int compare_word(const int *x, const int *y)
{
  int ix = *x;
  int iy = *y;

  char *word_x = tmp_word_list[ix];
  char *word_y = tmp_word_list[iy];

  return strncmp(word_x,word_y,MAX_WORD_LEN);
}

void register_freq(int line_c, char **word, int freq,
		   char *attr_vec, char *ans_class, char *buf)
{
  int v,i,e,a;
  int check_only;

  e = register_hashtable(&ht_vec, attr_vec, 0);

  check_only = (!mode_as_is) ? 1 : 0;
  for (v=0; v<tuple_size; v++) {
    i = register_hashtable(&ht[v], word[v], check_only);
    if (i < 0) die("Unknown attribute value -- %s\n",word[v]);
    freq_list[e].id[v] = i;
  }

  if (freq_list[e].line_no < 0) freq_list[e].line_no = line_c;

  if (mode_eval) {
    a = register_hashtable(&ht_class, ans_class, 0);

    if (freq_list[e].ans_id >= 0 && freq_list[e].ans_id != a)
      die("Illegal line %d (different answer classes for the same object): %s",
	  line_c,conv_msg(buf));

    freq_list[e].ans_id = a;
  }

  freq_list[e].freq += freq;
}

void make_word_freq_list(void)
{
  int i,e,v;

  word_freq_list = (int **)malloc(sizeof(int *)*tuple_size);
  if (word_freq_list == NULL)
    die("Couldn't allocate memory for word_freq_list\n");

  for (v=0; v<tuple_size; v++) {
    word_freq_list[v] = (int *)malloc(sizeof(int)*num_word[v]);
    if (word_freq_list[v] == NULL)
      die("Couldn't allocate memory for word_freq_list[v]\n");
    
    for (i=0; i<num_word[v]; i++)
      word_freq_list[v][i] = 0;
    
    for (e=0; e<num_entry; e++) {
      word_freq_list[v][freq_list[e].id[v]] += freq_list[e].freq;
    }
  }
}

