#include "nbc_def.h"
#include "nbc_ext.h"

void copy_read_param(void)
{
  int k,v,i;

  for (k=0; k<num_class; k++) {
    if (read_param_c[k] < 0 || read_param_c[k] > 1)
      die("Illegal attribute parameter value %.12f (class=%d)",
	  (double)read_param_c[k],k);
    param_c[k] = read_param_c[k];
  }

  for (k=0; k<num_class; k++)
    for (v=0; v<tuple_size; v++)
      for (i=0; i<num_word[v]; i++) {
	if (read_param_w[k][v][i] < 0 || read_param_w[k][v][i] > 1)
	  die("Illegal attribute parameter value %.12f (class=%d, attribute=%d)",
	      (double)read_param_w[k][v][i],k,v);
	param_w[k][v][i] = read_param_w[k][v][i];
      }
}

/*
 * [CSV form]
 *   - does not understand the double quatation character
 *   - allows the space/tab characters
 *   - ignores the lines starting with '#'
 *
 * [tab-separated from]
 *   - does not understand the double quatation character
 *   - allows the space characters
 *   - regards the consecutive tab characteres as one tab character
 *   - ignores the lines starting with '#'
 *
 * [Syntax of the parameter specification]
 *
 *   ## comment1 (could be multiple lines)
 *   0,P(c0)
 *   1,P(c1)
 *    :
 *   ## comment2-0
 *   c0,w0_0,P(w0_0|c0)
 *    :
 *   c0,w0_N,P(w0_N|c0)
 *   c1,w0_0,P(w0_0|c1)
 *    :
 *   ## comment2-1
 *   c0,w1_0,P(w1_0|c0)
 *    :
 *   ## comment2-2
 *   c0,w2_0,P(w2_0|c0)
 *    :
 */
int read_param_body(FILE *fp)
{
  char *buf, **word;
  char str_class[MAX_INT_LEN + 1];
  char str_param[MAX_FLOAT_LEN + 1];
  FLOAT_TYPE param;
  char *b, *q;
  int line_c,c,v;
  char *buf_conv = NULL;
  char **word_conv = NULL;
#ifdef USE_ICONV
  char *bb;
  char **ww = NULL;
  size_t buf_len_in, buf_len_out;
  size_t *w_len_in = NULL;
  size_t *w_len_out = NULL;
#  ifdef ICONV_CONST
  const char **word_tmp;
  const char *buf_tmp;
#  else
  char **word_tmp = NULL;
  char *buf_tmp = NULL;
#  endif
#endif
  int num_comment_field;
  int is_comment;
  char *endptr;
  int tmp_class_id = -1;
  int tmp_word_id;
  int max_num_class;
  int new_max_num_class;
  int *max_num_word = NULL;
  int *new_max_num_word = NULL;
  int prev_tmp_class_id;
  int num_sep;

  buf = (char *)malloc(sizeof(char)*(MAX_LINE_LEN + 1));
  if (buf == NULL)
    die("Couldn't allocate memory for buf\n");

  word = (char **)malloc(sizeof(char *)*tuple_size);
  if (word == NULL) die("Couldn't allocate memory for word\n");

  for (v=0; v<tuple_size; v++) {
    word[v] = (char *)malloc(sizeof(char)*(MAX_WORD_LEN + 1));
    if (word[v] == NULL)
      die("Couldn't allocate memory for word[v]\n");
  }

  if (mode_conv_in) {
    buf_conv = (char *)malloc(sizeof(char)*(MAX_LINE_LEN + 1));
    if (buf_conv == NULL)
      die("Couldn't allocate memory for buf_conv\n");
  }

  if (mode_conv_out) {
    word_conv = (char **)malloc(sizeof(char *)*tuple_size);
    if (word_conv == NULL)
      die("Couldn't allocate memory for word_conv\n");

    for (v=0; v<tuple_size; v++) {
      word_conv[v] = (char *)malloc(sizeof(char)*(MAX_WORD_LEN + 1));
      if (word_conv[v] == NULL)
	die("Couldn't allocate memory for word_conv[v]\n");
    }

#ifdef USE_ICONV
#  ifdef ICONV_CONST
    word_tmp = (const char **)malloc(sizeof(const char *)*tuple_size);
#  else
    word_tmp = (char **)malloc(sizeof(char *)*tuple_size);
#  endif
    if (word_tmp == NULL)
      die("Couldn't allocate memory for word_tmp\n");

    for (v=0; v<tuple_size; v++) {
      word_tmp[v] = (char *)malloc(sizeof(char)*(MAX_WORD_LEN + 1));
      if (word_tmp[v] == NULL)
	die("Couldn't allocate memory for word_tmp[v]\n");
    }

    w_len_in = (size_t *)malloc(sizeof(size_t)*tuple_size);
    w_len_out = (size_t *)malloc(sizeof(size_t)*tuple_size);
    ww = (char **)malloc(sizeof(char *)*tuple_size);
    if (w_len_in == NULL || w_len_out == NULL || ww == NULL)
      die("Couldn't allocate memory for w_len_{in,out}, ww\n");
#endif
  }

  /* allocate the temporary areas */
  max_num_class = INIT_TMP_CLASS;
  max_num_word = (int *)malloc(sizeof(int)*tuple_size);
  if (max_num_word == NULL)
    die("Couldn't allocate max_num_word\n");
  for (v=0; v<tuple_size; v++) {
    max_num_word[v] = INIT_TMP_WORD;
  }
  allocate_read_param(max_num_class, max_num_word);

  line_c = 0;
  num_comment_field = 0;
  is_comment = 0;
  num_class = 0;
  
  num_word = (int *)malloc(sizeof(int)*tuple_size);
  if (num_word == NULL)
    die("Couldn't allocate memory for num_word\n");
  for (v=0; v<tuple_size; v++)
    num_word[v] = 0;

  while (fgets(buf,MAX_LINE_LEN,fp) != NULL) {
    line_c++;

#ifdef USE_ICONV
    if (mode_conv_in) {
      buf_tmp = buf;
      bb = buf_conv;
      buf_len_in  = strlen(buf)+1;
      buf_len_out = MAX_LINE_LEN+1;
      if (iconv(cd_in, &buf_tmp, &buf_len_in, &bb, &buf_len_out) == (size_t)-1)
	die("[line %d] Couldn't convert to the internal charcode (errno=%d)\n",
	    line_c,errno);
    }
    else buf_conv = buf;
#else
    buf_conv = buf;
#endif

    b = buf_conv;

    /* we count up the number of comment areas */
    if (*b == '#') {
      is_comment = 1;
      continue;
    }
    else {
      if (is_comment) num_comment_field++;
      is_comment = 0;
    }

    /*
     * parse the area of P(c),P(w1|c),P(w2|c),P(w3|c),... which are separated by
     * the comment lines
     */

    if (num_comment_field > tuple_size + 1) {
      die("Too many attribute parameter fields in %s\n",fname_param);
    }
    else if (num_comment_field == 1) { /** Reading P(c) ... **/
      /**** tab-separated form ****/
      if (!mode_csv_in) {

        while (*b == ' ') b++;
        
        c = 0;
        q = str_class;
	num_sep = 0;
        while (*b != ' ' && *b != '\t') {
          if (c > MAX_INT_LEN) die_line(line_c, __LINE__, buf_conv);
          *q = *b;
          q++; b++; c++;
        }
        *q = '\0';

        q = str_class;
        tmp_class_id = strtol(q, &endptr, 10);
        if (*endptr != '\0') die_line(line_c, __LINE__, buf_conv);

        /* we assume that the class numbers occurs in the incremental order (0,1,2,...) */
        if (num_class != tmp_class_id)
          die_line(line_c, __LINE__, buf_conv);
        else
          num_class++;

        while (*b == ' ') b++;
        if (*b != '\t') die_line(line_c, __LINE__, buf_conv);

        b++;
        while (*b == '\t') b++;
        while (*b == ' ') b++;

        c = 0;
        q = str_param;
        while (*b != ' ' && *b != '\t' && *b != '\r' && *b != '\n') {
          if (!isdigit(*b) && *b != '.' && *b != '-')  /* Allows negative values here */
	    die_line(line_c, __LINE__, buf_conv);
          if (c > MAX_FLOAT_LEN) die_line(line_c, __LINE__, buf_conv);
          *q = *b;
          q++; b++; c++;
        }
        *q = '\0';

        q = str_param;
        param = strtod(q, &endptr);
        if (*endptr != '\0') die_line(line_c, __LINE__, buf_conv);
        
        /* expand the temporary areas if necessary */
        if (tmp_class_id >= max_num_class) {
	  new_max_num_class = max_num_class * 2;
          expand_read_param_c(max_num_class, new_max_num_class);
          expand_read_param_w(max_num_class, new_max_num_class,
			      max_num_word, max_num_word);
	  max_num_class = new_max_num_class;
        }

        read_param_c[tmp_class_id] = param;

        while (*b == ' ') b++;
        if (*b != '\r' && *b != '\n') die_line(line_c, __LINE__, buf_conv);

      }
      /**** CSV form ****/
      else {

        while (*b == ' ' || *b == '\t') b++;

        c = 0;
        q = str_class;
        while (*b != ' ' && *b != '\t' && *b != ',') {
          if (c > MAX_INT_LEN) die_line(line_c, __LINE__, buf_conv);
          *q = *b;
          q++; b++; c++;
        }
        *q = '\0';

        q = str_class;
        tmp_class_id = strtol(q, &endptr, 10);
        if (*endptr != '\0') die_line(line_c, __LINE__, buf_conv);

        if (num_class != tmp_class_id)
          die_line(line_c, __LINE__, buf_conv);
        else
          num_class++;

        while (*b == ' ' || *b == '\t') b++;
        if (*b != ',') die_line(line_c, __LINE__, buf_conv);

        b++;
        while (*b == ' ' || *b == '\t') b++;
        
        c = 0;
        q = str_param;
        while (*b != ' ' && *b != '\t' && *b != ',' && *b != '\r' && *b != '\n') {
          if (!isdigit(*b) && *b != '.' && *b != '-')   /* Allows negative values here */
            die_line(line_c, __LINE__, buf_conv);
          if (c > MAX_FLOAT_LEN) die_line(line_c, __LINE__, buf_conv);
          *q = *b;
          q++; b++; c++;
        }
        *q = '\0';

        q = str_param;
        param = strtod(q, &endptr);
        if (*endptr != '\0') die_line(line_c, __LINE__, buf_conv);
        
        if (tmp_class_id >= max_num_class) {
	  new_max_num_class = max_num_class * 2;
          expand_read_param_c(max_num_class, new_max_num_class);
          expand_read_param_w(max_num_class, new_max_num_class,
			      max_num_word, max_num_word);
	  max_num_class = new_max_num_class;
        }
        
        read_param_c[tmp_class_id] = param;

        while (*b == ' ' || *b == '\t') b++;
        if (*b != '\r' && *b != '\n') die_line(line_c, __LINE__, buf_conv);

      }
    }
    else {
      v = num_comment_field - 2;  /* Get the attribute index */

      /**** tab-separated form ****/
      if (!mode_csv_in) {

        while (*b == ' ') b++;
        
        c = 0;
        q = str_class;
        while (*b != ' ' && *b != '\t') {
          if (c > MAX_INT_LEN) die_line(line_c, __LINE__, buf_conv);
          *q = *b;
          q++; b++; c++;
        }
        *q = '\0';

        q = str_class;
        prev_tmp_class_id = tmp_class_id;
        tmp_class_id = strtol(q, &endptr, 10);
        if (*endptr != '\0') die_line(line_c, __LINE__, buf_conv);

        /* from the syntax of the parameter file, the number of class has been
         * already determined here.
         */
        if (tmp_class_id >= num_class)
          die("Different number of classes exist in entries for P(c) and P(w1|c)\n");

        if (tmp_class_id != prev_tmp_class_id)
          num_word[v] = 0;  /* reset the number of attribute values */

        while (*b == ' ') b++;
        if (*b != '\t') die_line(line_c, __LINE__, buf_conv);

        b++;
        while (*b == '\t') b++;
        while (*b == ' ') b++;

        c = 0;
        q = word[v];
        while (*b != ' ' && *b != '\t') {
          if (c > MAX_WORD_LEN) die_line(line_c, __LINE__, buf_conv);
          *q = *b;
          q++; b++; c++;
        }
        *q = '\0';
        
#ifdef USE_ICONV
        if (mode_conv_out) {
          word_tmp[v] = word[v];
          ww[v] = word_conv[v];
          w_len_in[v] = strlen(word[v])+1;
          w_len_out[v] = MAX_WORD_LEN+1;
          if (iconv(cd_out,&word_tmp[v],&w_len_in[v],&ww[v],&w_len_out[v]) == (size_t)-1)
            die("[line %d] Couldn't convert to %s (errno=%d)\n",
		line_c,charcode_out,errno);
        }
        else word_conv = word;
#else
	word_conv = word;
#endif

        tmp_word_id = register_hashtable(&ht[v], word_conv[v], 0);
        if (num_word[v] != tmp_word_id)
          die("[line %d] Found an inconsistent entry (%s)",line_c,word_conv[v]);
        else 
          num_word[v]++;

        while (*b == ' ') b++;
        if (*b != '\t') die_line(line_c, __LINE__, buf_conv);

        b++;
        while (*b == '\t') b++;
        while (*b == ' ') b++;

        c = 0;
        q = str_param;
        while (*b != ' ' && *b != '\t' && *b != '\r' && *b != '\n') {
          if (!isdigit(*b) && *b != '.' && *b != '-')   /* Allows negative values here */
            die_line(line_c, __LINE__, buf_conv);
          if (c > MAX_FLOAT_LEN) die_line(line_c, __LINE__, buf_conv);
          *q = *b;
          q++; b++; c++;
        }
        *q = '\0';
        
        q = str_param;
        param = strtod(q, &endptr);
        if (*endptr != '\0') die_line(line_c, __LINE__, buf_conv);
        
        if (tmp_class_id >= max_num_class || tmp_word_id >= max_num_word[v]) {
	  new_max_num_class =
	    (tmp_class_id >= max_num_class) ? max_num_class * 2 : max_num_class;
	  new_max_num_word[v] =
	    (tmp_word_id >= max_num_word[v]) ? max_num_word[v] * 2 : max_num_word[v];
	  expand_read_param_w(max_num_class, new_max_num_class,
			      max_num_word, new_max_num_word);
	  max_num_class = new_max_num_class;
	  max_num_word = new_max_num_word;
	}

        read_param_w[tmp_class_id][v][tmp_word_id] = param;

        while (*b == ' ') b++;
        if (*b != '\r' && *b != '\n') die_line(line_c, __LINE__, buf_conv);

      }
      /**** CSV form ****/
      else {

        while (*b == ' ' || *b == '\t') b++;

        c = 0;
        q = str_class;
        while (*b != ' ' && *b != '\t' && *b != ',') {
          if (c > MAX_INT_LEN) die_line(line_c, __LINE__, buf_conv);
          *q = *b;
          q++; b++; c++;
        }
        *q = '\0';
        
        q = str_class;
        prev_tmp_class_id = tmp_class_id;
        tmp_class_id = strtol(q, &endptr, 10);
        if (*endptr != '\0') die_line(line_c, __LINE__, buf_conv);

        if (tmp_class_id >= num_class)
          die("Different number of classes exist in entries for P(c) and P(w|c)\n");

        if (tmp_class_id != prev_tmp_class_id)
          num_word[v] = 0;

        while (*b == ' ' || *b == '\t') b++;
        if (*b != ',') die_line(line_c, __LINE__, buf_conv);

        b++;
        while (*b == ' ' || *b == '\t') b++;

        c = 0;
        q = word[v];
        while (*b != ' ' && *b != '\t' && *b != ',') {
          if (c > MAX_WORD_LEN) die_line(line_c, __LINE__, buf_conv);
          *q = *b;
          q++; b++; c++;
        }
        *q = '\0';
        
#ifdef USE_ICONV
        if (mode_conv_out) {
          word_tmp[v] = word[v];
          ww[v] = word_conv[v];
          w_len_in[v] = strlen(word[v])+1;
          w_len_out[v] = MAX_WORD_LEN+1;
          if (iconv(cd_out,&word_tmp[v],&w_len_in[v],&ww[v],&w_len_out[v]) == (size_t)-1)
            die("[line %d] Couldn't convert to %s (errno=%d)\n",
		line_c,charcode_out,errno);
        }
        else word_conv = word;
#else
	word_conv = word;
#endif

        tmp_word_id = register_hashtable(&ht[v], word_conv[v], 0);
        if (num_word[v] != tmp_word_id)
          die("[line %d] Found an entry which contains a duplicated word (%s)",
              line_c,word_conv[v]);
        else 
          num_word[v]++;

        while (*b == ' ' || *b == '\t') b++;
        if (*b != ',') die_line(line_c, __LINE__, buf_conv);

        b++;
        while (*b == ' ' || *b == '\t') b++;

        c = 0;
        q = str_param;
        while (*b != ' ' && *b != '\t' && *b != '\r' && *b != '\n') {
          if (!isdigit(*b) && *b != '.' && *b != '-')   /* Allows negative values here */
            die_line(line_c, __LINE__, buf_conv);
          if (c > MAX_FLOAT_LEN) die_line(line_c, __LINE__, buf_conv);
          *q = *b;
          q++; b++; c++;
        }
        *q = '\0';
        
        q = str_param;
        param = strtod(q, &endptr);
        if (*endptr != '\0') die_line(line_c, __LINE__, buf_conv);
        
        if (tmp_class_id >= max_num_class || tmp_word_id >= max_num_word[v]) {
	  new_max_num_class =
	    (tmp_class_id >= max_num_class) ? max_num_class * 2 : max_num_class;
	  new_max_num_word[v] =
	    (tmp_word_id >= max_num_word[v]) ? max_num_word[v] * 2 : max_num_word[v];
	  expand_read_param_w(max_num_class, new_max_num_class,
			      max_num_word, new_max_num_word);
	  max_num_class = new_max_num_class;
	  max_num_word = new_max_num_word;
	}
        
        read_param_w[tmp_class_id][v][tmp_word_id] = param;

        while (*b == ' ' || *b == '\t') b++;
        if (*b != '\r' && *b != '\n') die_line(line_c, __LINE__, buf_conv);

      }
    }

  }

  if (num_comment_field < tuple_size + 1)
    die("Too few attribute parameter fields in %s\n",fname_param);

  for (v=0; v<tuple_size; v++) {
    num_word[v] = ht[v].serial_id;
  }

  return max_num_class;
}

void read_param(void)
{
  int v,i;
  int max_num_class;
  FILE *fp_param;

  ht = (HASHTABLE *)malloc(sizeof(HASHTABLE)*tuple_size);
  if (ht == NULL)
    die("Couldn't allocate memory for ht\n");

  for (v=0; v<tuple_size; v++)
    init_hashtable(&ht[v]);

  num_word = (int *)malloc(sizeof(int)*tuple_size);
  if (num_word == NULL)
    die("Couldn't allocate memory for num_word\n");

  fp_param = fopen(fname_param,"rb");
  if (fp_param == NULL)
    die("Couldn't open %s\n", fname_param);

  max_num_class = read_param_body(fp_param); /* allocate_read_param() is called */

  allocate_param();
  copy_read_param();
  release_read_param(max_num_class);

  word_list = (char ***)malloc(sizeof(char **)*tuple_size);
  if (word_list == NULL)
    die("Couldn't allocate memory for word_list\n");

  for (v=0; v<tuple_size; v++)
    word_list[v] = make_entry_list(&ht[v]); 

  print_order = (int **)malloc(sizeof(int *)*tuple_size);
  if (print_order == NULL)
    die("Couldn't allocate memory for print_order\n");

  for (v=0; v<tuple_size; v++) {
    print_order[v] = (int *)malloc(sizeof(int)*num_word[v]);
    for (i=0; i<num_word[v]; i++) {
      print_order[v][i] = i;
    }
    if (mode_lex_order) {
      tmp_word_list = word_list[v];
      qsort(print_order[v],num_word[v],sizeof(int),
            (int(*)(const void *,const void *))compare_word);
    }
  }

  if (mode_verbose) print_att_list(fp_log);
}
