#include "nbc_def.h"
#include "nbc_ext.h"

/*
 * SMEM run(s) with some paticular number of classes
 * -- u is the index for searching the optimal number of classes
 */
void smem_iterate(int u)
{
  int t,h;
  FLOAT_TYPE last_lambda;
  FLOAT_TYPE curr_jsplit, curr_jmerge;
  int sm_ok;
  int n,n0;
  int m1,m2;
  int k1,k2,k3,last_k1=-1,last_k2=-1;
  int num_tried_cand, num_tried_jsplit_cand=-1;
  int num_iterate,num_iterate_best=-1;
  int r,r_best=-1;
  unsigned int curr_seed=0;
  int *cand=NULL;

  num_candidate = num_class*(num_class - 1)*(num_class - 2)/2;
  n = (max_iterate > 0) ? max_iterate : MAX_MAX_ITERATE;
  if (num_init > 1) {
    n0 = (max_init_iterate > 0) ? max_init_iterate : n;
  } else {
    n0 = (max_iterate > 0) ? max_iterate : MAX_MAX_ITERATE;
  }
  
  allocate_jmerge();
  allocate_jsplit();
  allocate_candidate();
  allocate_param_sm();
  allocate_log_likelihood_sm();
  allocate_num_iterate_sm();
  allocate_num_result_sm();
  allocate_param_part();
  allocate_prob_part();
  if (mode_rank_sm) {
    allocate_prob_att_mem_sm();
    allocate_rank_sm();
  }

  for (t=0; t<(num_sm+1); t++) {
    num_iterate_sm_full[t] = 0;
    num_iterate_sm_part[t] = 0;
  }

  for (t=0; t<num_sm; t++) {
    num_result_rejected[t] = 0;
    num_result_skipped[t] = 0;
  }

  for (r=0; r<num_init; r++) { /* random restarts */
    if (!mode_set_seed) {
      curr_seed = (unsigned int)rand();
      srand(curr_seed);
    }

    fprintf(fp_msg,"    [%d] #iterations ",r);

    initialize_param();

    num_iterate = smem_iterate_full(0,0,n0);

    num_iterate_sm_full[0] = num_iterate;

    log_seed[u][r] = (!mode_set_seed) ? curr_seed : 0;
    log_num_iterate[u][r] = num_iterate;
    log_final_lambda[u][r] = final_lambda;

    if (r == 0 || final_lambda > final_lambda_best) {
      r_best = r;
      final_lambda_best = final_lambda;
      final_log_likelihood_best = final_log_likelihood;
      final_log_prior_best = final_log_prior;
      num_iterate_best = num_iterate;
      save_log_likelihood_best_sm(0,num_iterate_best);
      save_param_best();
    }
  }
  log_r_best[u] = r_best;

  final_lambda = final_lambda_best;
  final_log_likelihood = final_log_likelihood_best;
  final_log_prior = final_log_prior_best;
  restore_param_best();
  restore_log_likelihood_best_sm(0,num_iterate_best);
  
  /* used for the judgement of the acceptance of a split-merge operation */
  last_lambda = final_lambda;


  /**** Split-Merge EM algorithm ****/

  for (t=0; t<num_sm; t++) {

    /* save the membership probs and ranking before a SM operation */
    if (mode_rank_sm) {
      compute_member_prob_sm();
      compute_rank_sm();
    }
    
    save_param_sm();

    compute_jmerge();
    compute_jsplit();
    sort_candidate();
    
    sm_ok = 0;
    num_tried_cand = 0;
    for (h=0; (max_candidate==0 || num_tried_cand<max_candidate) && h<num_candidate; h++) {

      cand = sm_candidate[h];
      k1 = cand[0];
      k2 = cand[1];
      k3 = cand[2];
      curr_jmerge = jmerge[k1][k2];
      curr_jsplit = jsplit[k3];

      fprintf(fp_msg,conv_msg("  [Op=%d]\n"
			      "    Candidate %d (%d not skipped): Jmerge(%d,%d)=%.12Lf / Jsplit(%d)=%.12Lf\n"),
	      (t+1),h,num_tried_cand,k1,k2,curr_jmerge,k3,curr_jsplit);

      if (h==0 || last_k1 != k1 || last_k2 != k2) {
	num_tried_jsplit_cand = 0;
      }	else {
	num_tried_jsplit_cand++;
      }

      if (max_jsplit>0 && num_tried_jsplit_cand >= max_jsplit) {
	fprintf(fp_msg,"    -- Skipped (the number of examined split candidates reached the limit)\n");
	num_result_skipped[t]++;
	continue;
      }
      if ((min_param_c>0 && param_c[k1] < min_param_c && param_c[k2] < min_param_c)
	  || (min_param_c>0 && param_c[k3] < min_param_c)) {
	fprintf(fp_msg,"    -- Skipped (too small class probability)\n");
	num_result_skipped[t]++;
	continue;
      }
      num_tried_cand++;
      last_k1 = k1;
      last_k2 = k2;

      fprintf(fp_msg,"    #iterations ");

      m1 = smem_iterate_part((t+1),h,0,n);
      m2 = smem_iterate_full((t+1),m1,n);

      /* Conditions for accepting the SM operation
       *   - log-likelihood (or log of a posteriori prob) is improved
       *   - the improvement is large enough
       *   - all class probabilities of newly obtained classes are larger than
       *     pre-defined threshold (this condition is optional)
       */
      if (final_lambda > last_lambda) {
	if ((final_lambda - last_lambda) > min_smem_improve) {
	  if (!(min_new_param_c > 0) ||
	   (param_c[k1] > min_new_param_c &&
	    param_c[k2] > min_new_param_c &&
	    param_c[k3] > min_new_param_c)) {
	    
	    fprintf(fp_msg,conv_msg("    -- Accepted [%d,%d]->[%d] & [%d]->[%d,%d] "
				    "(L current: %.9Lf / last: %.9Lf)\n"),
		    k1,k2,k1,k3,k2,k3,final_lambda,last_lambda);
	    sm_ok = 1;
	  }
	  else {
	    fprintf(fp_msg,conv_msg("    -- Rejected [%d,%d]->[%d] & [%d]->[%d,%d] "
				    "(L current: %.9Lf / last: %.9Lf)\n"),
		    k1,k2,k1,k3,k2,k3,final_lambda,last_lambda);
	    fprintf(fp_msg,conv_msg("       <Too small split/merged class probabilities: "
				    "%d - %.9Lf, %d - %.9Lf, %d - %.9Lf>\n"),
		    k1,param_c[k1],k2,param_c[k2],k3,param_c[k3]);
	    sm_ok = 0;
	  }
	}
	else {
	  fprintf(fp_msg,conv_msg("    -- Rejected [%d,%d]->[%d] & [%d]->[%d,%d] "
				  "(L current: %.9Lf / last: %.9Lf)\n"),
		  k1,k2,k1,k3,k2,k3,final_lambda,last_lambda);
	  fprintf(fp_msg,conv_msg("       <Too small improvement: %.9Lf>\n"),
		  (final_lambda - last_lambda));
	  sm_ok = 0;
	}

      } else {
	fprintf(fp_msg,conv_msg("    -- Rejected [%d,%d]->[%d] & [%d]->[%d,%d] "
				"(L current: %.9Lf / last: %.9Lf)\n"),
		k1,k2,k1,k3,k2,k3,final_lambda,last_lambda);
	fprintf(fp_msg,conv_msg("       <No improvement>\n"));
	sm_ok = 0;
      }

      if (sm_ok) {
	last_lambda = final_lambda;
	num_iterate_sm_part[t+1] = m1;
	num_iterate_sm_full[t+1] = m2;

	if (mode_rank_sm) {
	  compute_member_prob_part(cand);
	  compute_rank_part(cand);
	  print_rank_sm(fp_sm,t,cand);
	}

	break;
      }
      else {
	num_result_rejected[t]++;
	restore_param_sm();  /* backtrack */
      }

    }

    if (sm_ok == 0) {
      if (h == num_candidate) {
	fprintf(fp_msg,"  [Op=%d]\n  <<All candidates are rejected>>\n",(t+1));
      } else {
	fprintf(fp_msg,"  [Op=%d]\n  <<The number of rejected split candidates reached the limit>>\n",(t+1));
      }
      break;
    }
  }
  
  print_num_iterate_sm(fp_log,num_class);
  print_log_likelihood_sm(fp_log,num_class);
  print_num_result_sm(fp_log,num_class);

  release_jmerge();
  release_jsplit();
  release_candidate();
  release_param_sm();
  release_log_likelihood_sm();
  release_num_iterate_sm();
  release_num_result_sm();
  release_param_part();
  release_prob_part();
  if (mode_rank_sm) {
    release_prob_att_mem_sm();
    release_rank_sm();
  }
}

int smem_iterate_full(int t, int m, int n)
{
  if (m == 0) {
    compute_complete_prob();
    compute_incomplete_prob();
    
    log_likelihood_sm[t][m] = compute_log_likelihood();
    if (mode_smooth) {
      log_prior_sm[t][m] = compute_log_prior();
      lambda_sm[t][m] = log_likelihood_sm[t][m] + log_prior_sm[t][m];
    } else {
      lambda_sm[t][m] = log_likelihood_sm[t][m];
    }

    print_dot_message(m);
  }

  while (1) {
    compute_conditional_prob();
    compute_expect();
    update_param();

    m++;

    compute_complete_prob();
    compute_incomplete_prob();

    log_likelihood_sm[t][m] = compute_log_likelihood();
    if (mode_smooth) {
      log_prior_sm[t][m] = compute_log_prior();
      lambda_sm[t][m] = log_likelihood_sm[t][m] + log_prior_sm[t][m];
    } else {
      lambda_sm[t][m] = log_likelihood_sm[t][m];
    }

    print_dot_message(m);

    if (lambda_sm[t][m] < lambda_sm[t][m-1] &&
	lambda_sm[t][m-1] - lambda_sm[t][m] > ACCEPT_SMALL_NUMBER) {
      fprintf(fp_msg,conv_msg("\n(full) [%d] LL or LPost decreased! (curr: %.6Lf / last: %.6Lf)\n"),
	      m,lambda_sm[t][m],lambda_sm[t][m-1]);
      /* print_param(fp_msg); */
      break;
    }

    if (m >= n) {
      fprintf(fp_msg,conv_msg(" (Stopped: %d iterations)\tL=%.6Lf\n"),m,lambda_sm[t][m]);
      break;
    } else {
      if (lambda_sm[t][m] - lambda_sm[t][m-1] < epsilon) {
	fprintf(fp_msg,conv_msg(" (Converged: %d iterations)\tL=%.6Lf\n"),m,lambda_sm[t][m]);
	break;
      }
    }
  }

  final_log_likelihood = log_likelihood_sm[t][m];
  final_lambda = lambda_sm[t][m];
  if (mode_smooth) final_log_prior = log_prior_sm[t][m];

  return m;
}

int smem_iterate_part(int t, int h, int m, int n)
{
  int *cand;

  cand = sm_candidate[h];

  initialize_param_part(cand);
  save_param_part(cand);

  compute_complete_prob();
  compute_incomplete_prob();
  save_incomplete_prob_part(cand);

  log_likelihood_sm[t][m] = compute_log_likelihood();
  if (mode_smooth) {
    log_prior_sm[t][m] = compute_log_prior();
    lambda_sm[t][m] = log_likelihood_sm[t][m] + log_prior_sm[t][m];
  } else {
    lambda_sm[t][m] = log_likelihood_sm[t][m];
  }
  
  print_dot_message(m);

  while (1) {
    compute_conditional_prob_part(cand);
    compute_expect_part(cand);
    update_param_part(cand);

    m++;

    compute_complete_prob_part(cand);
    compute_incomplete_prob_part(cand);

    log_likelihood_sm[t][m] = compute_log_likelihood();
    if (mode_smooth) {
      log_prior_sm[t][m] = compute_log_prior();
      lambda_sm[t][m] = log_likelihood_sm[t][m] + log_prior_sm[t][m];
    } else {
      lambda_sm[t][m] = log_likelihood_sm[t][m];
    }

    print_dot_message(m);

    if (lambda_sm[t][m] < lambda_sm[t][m-1] &&
	lambda_sm[t][m-1] - lambda_sm[t][m] > ACCEPT_SMALL_NUMBER) {
      fprintf(fp_msg,conv_msg("\n(part) [%d] LL or LPost decreased! (curr: %.6Lf / last: %.6Lf)\n"),
	      m,lambda_sm[t][m],lambda_sm[t][m-1]);
      break;
    }

    if (m >= n) {
      fprintf(fp_msg,"(S:%d)",m);
      break;
    } else {
      if (lambda_sm[t][m] - lambda_sm[t][m-1] < epsilon) {
	fprintf(fp_msg,"(C:%d)",m);
	break;
      }
    }
  }

  /* do not update final_log_likelihood, etc. */

  return m;
}

void compute_jmerge()
{
  int k,k1,k2,e;
  FLOAT_TYPE *norm;

  if (mode_norm) {

    norm = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
    if (norm == NULL) die("Couldn't allocate memory for norm\n");

#pragma omp parallel for private(e)
    for (k=0; k<num_class; k++) {
      norm[k] = 0.0;
      for (e=0; e<num_entry; e++) {
	norm[k] += freq_list[e].freq * prob_cond[k][e] * prob_cond[k][e];
      }
    }
    
#pragma omp parallel for private(k2,e)
    for (k1=0; k1<num_class; k1++) {
      for (k2=(k1+1); k2<num_class; k2++) {
	jmerge[k1][k2] = 0.0;
	for (e=0; e<num_entry; e++) {
	  jmerge[k1][k2] += freq_list[e].freq * prob_cond[k1][e] * prob_cond[k2][e];
	}
	jmerge[k1][k2] /= sqrt(norm[k1]*norm[k2]);
      }
    }
    
    if (mode_norm) {
      free(norm); norm = NULL;
    }

  } else {

#pragma omp parallel for private(k2,e)
    for (k1=0; k1<num_class; k1++) {
      for (k2=(k1+1); k2<num_class; k2++) {
	jmerge[k1][k2] = 0.0;
	for (e=0; e<num_entry; e++) {
	  jmerge[k1][k2] += freq_list[e].freq * prob_cond[k1][e] * prob_cond[k2][e];
	}
      }
    }
    
  }

}

void sort_candidate()
{
  int k1,k2,k3;
  int h;

  h = 0;
  for (k1=0; k1<num_class; k1++) {
    for (k2=(k1+1); k2<num_class; k2++) {
      for (k3=0; k3<num_class; k3++) {
	if (k3 == k1 || k3 == k2) continue;
	sm_candidate[h][0] = k1;
	sm_candidate[h][1] = k2;
	sm_candidate[h][2] = k3;
	h++;
      }
    }
  }

  tmp_jmerge = jmerge;
  tmp_jsplit = jsplit;

  qsort(sm_candidate,num_candidate,sizeof(int *),
	(int(*)(const void *, const void *))compare_candidate);
}

int compare_candidate(const int **x, const int **y)
{
  int k1_x = (*x)[0];
  int k2_x = (*x)[1];
  int k3_x = (*x)[2];
  int k1_y = (*y)[0];
  int k2_y = (*y)[1];
  int k3_y = (*y)[2];

  FLOAT_TYPE jm_x = tmp_jmerge[k1_x][k2_x];
  FLOAT_TYPE jm_y = tmp_jmerge[k1_y][k2_y];
  FLOAT_TYPE js_x = tmp_jsplit[k3_x];
  FLOAT_TYPE js_y = tmp_jsplit[k3_y];

  if (k1_x == k1_y && k2_x == k2_y) {
    if (js_x > js_y) return -1;
    if (js_x < js_y) return 1;
    return 0;
  }
  else {
    if (jm_x > jm_y) return -1;
    if (jm_x < jm_y) return 1;
    return 0;
  }
}

void compute_incomplete_prob_part(int *cand)
{
  int l,k,e;

#pragma omp parallel for private(l,k)
  for (e=0; e<num_entry; e++) {
    prob_incomp[e] = prob_incomp_part[e];
    for (l=0; l<3; l++) {
      k = cand[l];
      prob_incomp[e] += prob_comp[k][e];
    }
  }
}

void compute_conditional_prob_part(int *cand)
{
  int l,k;
  int e;
  FLOAT_TYPE p;

  for (l=0; l<3; l++) {
    k = cand[l];
#pragma omp parallel for private(p)
    for (e=0; e<num_entry; e++) {
      p = prob_incomp[e];
      prob_cond[k][e] = (p > 0) ? prob_comp[k][e]/p : 0.0;
    }
  }
}

void save_incomplete_prob_part(int *cand)
{
  int k,e;
  int k1,k2,k3;

  k1 = cand[0];
  k2 = cand[1];
  k3 = cand[2];

#pragma omp parallel for
  for (e=0; e<num_entry; e++) {
    prob_incomp_part[e] = 0.0;
  }

  /* Cannot immediately add a `pragma omp' statement here */ 
  for (k=0; k<num_class; k++) {
    if (k == k1 || k == k2 || k == k3) continue;
    for (e=0; e<num_entry; e++) {
      prob_incomp_part[e] += prob_comp[k][e];
    }
  }
}

void save_conditional_prob_part(int *cand)
{
  int l,k,e;

  for (l=0; l<3; l++) {
    k = cand[l];
#pragma omp parallel for
    for (e=0; e<num_entry; e++) {
      prob_cond_part[l][e] = prob_cond[k][e];
    }
  }
}

void save_log_likelihood_best_sm(int t,int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++) {
    log_likelihood_best[m] = log_likelihood_sm[t][m];
    lambda_best[m] = lambda_sm[t][m];
  }

  if (mode_smooth)
    for (m=0; m<num_iterate; m++) {
      log_prior_best[m] = log_prior_sm[t][m];
    }
}

void restore_log_likelihood_best_sm(int t,int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++) {
    log_likelihood_sm[t][m] = log_likelihood_best[m];
    lambda_sm[t][m] = lambda_best[m];
  }

  if (mode_smooth)
    for (m=0; m<num_iterate; m++) {
      log_prior_sm[t][m] = log_prior_best[m];
    }
}

void initialize_param_part(int *cand)
{
  int i,k1,k2,k3,v;
  FLOAT_TYPE r1,r2;
  FLOAT_TYPE noise1,noise2;
  FLOAT_TYPE sum1,sum2,sum3;
  FLOAT_TYPE p1,p2,p3;

  k1 = cand[0];
  k2 = cand[1];
  k3 = cand[2];

  param_c[k1] = param_c[k1] + param_c[k2];
  param_c[k2] = 0.5 * param_c[k3];
  param_c[k3] = 0.5 * param_c[k3];

  r1 = expect_c[k1]/(expect_c[k1] + expect_c[k2]);
  r2 = expect_c[k2]/(expect_c[k1] + expect_c[k2]);

  /* To keep the repeatability, we do not use OpenMP here */
  for (v=0; v<tuple_size; v++) {
    sum1 = 0.0;
    sum2 = 0.0;
    sum3 = 0.0;
    for (i=0; i<num_word[v]; i++) {
      p1 = param_w[k1][v][i];
      p2 = param_w[k2][v][i];
      p3 = param_w[k3][v][i];
      noise1 = NOISE_RATIO*(rand() % RAND_MAX)/((FLOAT_TYPE)RAND_MAX*num_word[v]);
      noise2 = NOISE_RATIO*(rand() % RAND_MAX)/((FLOAT_TYPE)RAND_MAX*num_word[v]);
      param_w[k1][v][i] = r1 * p1 + r2 * p2;
      param_w[k2][v][i] = p3 + noise1;
      param_w[k3][v][i] = p3 + noise2;
      sum1 += param_w[k1][v][i];
      sum2 += param_w[k2][v][i];
      sum3 += param_w[k3][v][i];
    }
    /* normalization */
    for (i=0; i<num_word[v]; i++) {
      param_w[k1][v][i] /= sum1;
      param_w[k2][v][i] /= sum2;
      param_w[k3][v][i] /= sum3;
    }
  }

}

void compute_jsplit()
{
  int e,k,v;
  FLOAT_TYPE p_emp;
  FLOAT_TYPE pw;
  FLOAT_TYPE tmp_jsplit;

#pragma omp parallel for private(e,tmp_jsplit,p_emp,v,pw)
  for (k=0; k<num_class; k++) {
    jsplit[k] = 0.0;
    if (!(expect_c[k] > 0)) continue;

    for (e=0; e<num_entry; e++) {

      tmp_jsplit = 0.0;
      p_emp = freq_list[e].freq*prob_cond[k][e]/expect_c[k];

      if (p_emp > 0) tmp_jsplit += log(p_emp);

      for (v=0; v<tuple_size; v++) {
	pw = param_w[k][v][freq_list[e].id[v]];
	if (p_emp > 0 && pw > 0) tmp_jsplit -= log(pw);
      }

      jsplit[k] += p_emp * tmp_jsplit;
    }
  }
}

void compute_complete_prob_part(int *cand)
{
  int l,k,v,e;
  
  for (l=0; l<3; l++) {
    k = cand[l];
#pragma omp parallel for private(v)
    for (e=0; e<num_entry; e++) {
      prob_comp[k][e] = param_c[k];
      for (v=0; v<tuple_size; v++) {
	prob_comp[k][e] *= param_w[k][v][freq_list[e].id[v]];
      }
    }
  }
}

void compute_expect_part(int *cand)
{
  int l,k,i,e,v;
  FLOAT_TYPE tmp_e;
  FREQ_PTR fp;
  FLOAT_TYPE e_c_k;

  for (l=0; l<3; l++) {
    k = cand[l];
    e_c_k = 0.0;
#pragma omp parallel for reduction(+: e_c_k)
    for (e=0; e<num_entry; e++) {
      e_c_k += freq_list[e].freq * prob_cond[k][e];
    }
    expect_c[k] = e_c_k;
  }

  for (l=0; l<3; l++) {
    k = cand[l];

#pragma omp parallel for private(i)
    for (v=0; v<tuple_size; v++)
      for (i=0; i<num_word[v]; i++)
	expect_w[k][v][i] = 0.0;

    /* Cannot immediately add a `pragma omp' statement here */
    for (e=0; e<num_entry; e++) {
      fp = &freq_list[e];
      tmp_e = fp->freq * prob_cond[k][e];
      for (v=0; v<tuple_size; v++) {
	expect_w[k][v][fp->id[v]] += tmp_e;
      }
    }
  }

}

void check_expect_part(int *cand)
{
  int k,l,i,v;
  FLOAT_TYPE sum_c, sum_w;

  sum_c = 0.0;
  for (l=0; l<3; l++) {
    k = cand[l];
    sum_c += expect_c[k];
  }
  fprintf(fp_msg,conv_msg("check> (part) sum_E_c = %.9Lf\n"),sum_c);

  for (l=0; l<3; l++) {
    k = cand[l];
#pragma omp parallel for private(sum_w,i)
    for (v=0; v<tuple_size; v++) {
      sum_w = 0.0;
      for (i=0; i<num_word[v]; i++) {
	sum_w += expect_w[k][v][i];
      }
      fprintf(fp_msg,conv_msg("check> (part) sum_E_w[%d][%d] = %.9Lf\n"),k,v,sum_w);
    }
  }
}

void compute_member_prob_sm(void)
{
  int k,v,i;
  FLOAT_TYPE p;

#pragma omp parallel for private(p,v,i)
  for (k=0; k<num_class; k++) {
    p = param_c[k];
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++)
	prob_w_sm[v][i] = 0.0;
      for (i=0; i<num_word[v]; i++) {
	prob_w_sm[v][i] += (prob_att_mem_sm[k][v][i] = p * param_w[k][v][i]);
      }
    }
  }

#pragma omp parallel for private(v,i)
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
	prob_att_mem_sm[k][v][i] /= prob_w_sm[v][i];
      }
    }
  }
}

void compute_member_prob_part(int *cand)
{
  int k,i,l,v;
  FLOAT_TYPE p,q;
  
  /* compute the normalizing factor */
#pragma omp parallel for private(v,i,p,q)
  for (k=0; k<num_class; k++) {
    p = param_c[k];

    for (v=0; v<tuple_size; v++)
      for (i=0; i<num_word[v]; i++)
	prob_w_sm[v][i] = 0.0;

    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
	q = p * param_w[k][v][i];
	if (k == cand[0]) prob_att_mem_part[0][v][i] = q;
	if (k == cand[1]) prob_att_mem_part[1][v][i] = q;
	if (k == cand[2]) prob_att_mem_part[2][v][i] = q;
	prob_w_sm[v][i] += q;
      }
    }
  }

  /* normalization */
  for (l=0; l<3; l++) {
#pragma omp parallel for private(i)
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
	prob_att_mem_part[l][v][i] /= prob_w_sm[v][i];
      }
    }
  }
}

/* implemented for the future use */
void save_param_part(int *cand)
{
  int l,k;

  for (l=0; l<3; l++) {
    k = cand[l];
    param_c_part[l] = param_c[k];
  }
}

void update_param_part(int *cand)
{
  FLOAT_TYPE denom_w;
  int k,l,i,v;
  FLOAT_TYPE param_c_sum0, param_c_sum;

  /* compute the sum of parameters of a SM candidate */
  param_c_sum0 = 0.0;
  for (l=0; l<3; l++) {
    param_c_sum0 += param_c_part[l];
  }

  if (mode_read_hparam) {
    param_c_sum = 0.0;
    for (l=0; l<3; l++) {
      k = cand[l];
      param_c[k] = (expect_c[k] + hparam_c[k] - 1.0);
      param_c_sum += param_c[k];
    }
    for (l=0; l<3; l++) {
      k = cand[l];
      param_c[k] = param_c_sum0 * param_c[k] / param_c_sum;
    }
  }
  else if (smooth_c > 0) {
    param_c_sum = 0.0;
    for (l=0; l<3; l++) {
      k = cand[l];
      param_c[k] = (expect_c[k] + smooth_c);
      param_c_sum += param_c[k];
    }
    for (l=0; l<3; l++) {
      k = cand[l];
      param_c[k] = param_c_sum0 * param_c[k] / param_c_sum;
    }
  }
  else {
    param_c_sum = 0.0;
    for (l=0; l<3; l++) {
      k = cand[l];
      param_c_sum += expect_c[k];
    }
    for (l=0; l<3; l++) {
      k = cand[l];
      param_c[k] = param_c_sum0 * expect_c[k] / param_c_sum;
    }
  }

  if (mode_read_hparam) {
    for (l=0; l<3; l++) {
      k = cand[l];
#pragma omp parallel for private(denom_w,i)
      for (v=0; v<tuple_size; v++) {
	denom_w = 0.0;
	for (i=0; i<num_word[v]; i++)
	  denom_w += (param_w[k][v][i] = expect_w[k][v][i] + hparam_w[k][v][i] - 1.0);
	for (i=0; i<num_word[v]; i++)
	  param_w[k][v][i] /= denom_w;
      }
    }
  }
  else {
#pragma omp parallel for private(k,l,denom_w,i)
    for (v=0; v<tuple_size; v++) {
      if (smooth_w[v] > 0) {
	for (l=0; l<3; l++) {
	  k = cand[l];
	  denom_w = expect_c[k] + smooth_w[v] * num_word[v];
	  for (i=0; i<num_word[v]; i++) {
	    param_w[k][v][i] = (expect_w[k][v][i] + smooth_w[v])/denom_w;
	  }
	}
      } else {
	for (l=0; l<3; l++) {
	  k = cand[l];
	  denom_w = expect_c[k];
	  for (i=0; i<num_word[v]; i++) {
	    param_w[k][v][i] = expect_w[k][v][i]/denom_w;
	  }
	}
      }
    }
  }

}

void check_param_part(int *cand)
{
  int k,l,i,v;
  FLOAT_TYPE sum_c, sum_w;

  sum_c = 0.0;
  for (l=0; l<3; l++) {
    k = cand[l];
    sum_c += param_c[k];
  }
  fprintf(fp_msg,conv_msg("check> (part) sum_c = %.9Lf\n"),sum_c);

  for (l=0; l<3; l++) {
    k = cand[l];
#pragma omp parallel for private(sum_w,i)
    for (v=0; v<tuple_size; v++) {
      sum_w = 0.0;
      for (i=0; i<num_word[v]; i++) {
	sum_w += param_w[k][v][i];
      }
      fprintf(fp_msg,conv_msg("check> (part) sum_w[%d][%d] = %.9Lf\n"),k,v,sum_w);
    }
  }
}

void save_param_sm(void)
{
  int k,v,i;

#pragma omp parallel for private(v,i)
  for (k=0; k<num_class; k++) {
    param_c_sm[k] = param_c[k];
    for (v=0; v<tuple_size; v++)
      for (i=0; i<num_word[v]; i++)
	param_w_sm[k][v][i] = param_w[k][v][i];
  }
}

void restore_param_sm(void)
{
  int k,v,i;

#pragma omp parallel for private(v,i)
  for (k=0; k<num_class; k++) {
    param_c[k] = param_c_sm[k];
    for (v=0; v<tuple_size; v++)
      for (i=0; i<num_word[v]; i++)
	param_w[k][v][i] = param_w_sm[k][v][i];
  }
}
