#include "nbc_def.h"
#include "nbc_ext.h"
#include "vnbc_def.h"
#include "vnbc_ext.h"

void vbem_main(void)
{
  int u;
  FLOAT_TYPE ic_opt=0;
  FILE *fp_hparam;
  FILE *fp_memp;

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  stime_read = current_sec();
  stime_lap_read = elapsed_sec();
  read_freq();
  etime_read = current_sec();
  etime_lap_read = elapsed_sec();
#else
  read_freq();
#endif

  print_size(fp_log);

  if (mode_search_num_class) allocate_ic();
  allocate_seed();
  allocate_seed_vb();
  
  if (mode_rank_sm) {
    fp_sm = fopen(fname_rank_sm,"wb");
    if (fp_sm == NULL)
      die("Couldn't open %s\n",fname_rank_sm);
  }

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  stime_em = current_sec();
  stime_lap_em = elapsed_sec();
#endif

  if (mode_search_num_class) {
    u = 0;

    for (num_class = kmin; num_class <= kmax; num_class += kstep) {
      fprintf(fp_msg,"|Classes| = %d:\n",num_class);

      set_auto_smooth();

      allocate_hparam();
      allocate_hparam_upd();
      allocate_hparam_best();
      allocate_pi();
      if (!mode_nomem) allocate_prob();
      allocate_expect();
      allocate_free_energy();
      allocate_free_energy_best();

      if (num_sm > 0)
	vbsmem_iterate(u);
      else 
	vbem_iterate(u);

      if (mode_nomem) {
        ic[u] = compute_free_energy_nomem();
      }
      else {
        compute_pi();
        compute_complete_prob();
        compute_incomplete_prob();
        ic[u] = compute_free_energy();
      }

      release_hparam_best();
      release_pi();
      if (!mode_nomem) release_prob();
      release_expect();
      release_free_energy();
      release_free_energy_best();

      if (num_class == kmin || ic[u] > ic_opt) {
	fprintf(fp_msg,conv_msg("  [%d] Free Energy = %.6Lf (temporarily optimal)\n"),
                num_class,ic[u]);
	if (num_class > kmin) release_hparam_opt();
	num_class_opt = num_class;
	ic_opt = ic[u];

	save_hparam_opt();

      } else {
	fprintf(fp_msg,conv_msg("  [%d] Free Energy = %.6Lf\n"),num_class,ic[u]);
      }

      release_hparam();

      u++;
    }
    fprintf(fp_msg,"Optimal |Classes| = %d\n",num_class_opt);

    restore_hparam_opt();
    release_hparam_opt();
    num_class = num_class_opt;

    print_ic(fp_log);
    print_seed(fp_log);
  }
  else {
    fprintf(fp_msg,"|Classes| = %d:\n",num_class);

    set_auto_smooth();
      
    allocate_hparam();
    allocate_hparam_upd();
    allocate_free_energy();
    allocate_pi();
    if (!mode_nomem) allocate_prob();
    allocate_expect();
    allocate_hparam_best();
    allocate_free_energy_best();

    if (num_sm > 0)
      vbsmem_iterate(0);
    else
      vbem_iterate(0);

    release_hparam_best();
    release_pi();
    if (!mode_nomem) release_prob();
    release_expect();
    release_free_energy();
    release_free_energy_best();

    print_seed(fp_log);
  }

  if (mode_search_num_class) release_ic();
  release_seed();

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  etime_em = current_sec();
  etime_lap_em = elapsed_sec();
#endif

  fp_hparam = fopen(fname_smooth,"wb");
  if (fp_hparam == NULL)
    die("Couldn't open %s\n", fname_smooth);
  print_hparam_upd(fp_hparam);
  fclose(fp_hparam);

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  if (mode_memp) stime_memp = current_sec();
  if (mode_memp) stime_lap_memp = elapsed_sec();
#endif

  if (mode_memp || mode_cluster || mode_rank || mode_dist) {
    allocate_param();
    allocate_prob_mem();
    update_param_vbem();
    compute_member_prob();
  }

  if (mode_memp) {
    fp_memp = fopen(fname_memp,"wb");
    if (fp_memp == NULL)
      die("Couldn't open %s\n", fname_memp);
    print_member_prob(fp_memp);
    fclose(fp_memp);
  }

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  if (mode_memp) etime_memp = current_sec();
  if (mode_memp) etime_lap_memp = elapsed_sec();
#endif

  if (mode_dist) dist_main();
  if (mode_rank) rank_main();
  if (mode_cluster) cluster_main();

  release_hparam();
  if (mode_memp || mode_cluster || mode_rank || mode_dist) {
    release_prob_mem();
    release_param();
  }

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  print_time_vb(fp_log);
#endif
}

void no_vbem_main(void)
{
  FILE *fp_memp;

  fprintf(fp_msg,"Reading hyperparameters from %s...", fname_smooth);
  fflush(fp_msg);
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  stime_read = current_sec();
  stime_lap_read = elapsed_sec();
#endif
  read_hparam();  /* allocate_hparam() is called */
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  etime_read = current_sec();
  etime_lap_read = elapsed_sec();
#endif
  fprintf(fp_msg,"done\n");

  if (mode_cluster) {
    fprintf(fp_msg,"Reading cases from %s...", fname_freq);
    fflush(fp_msg);
    read_freq();
    fprintf(fp_msg,"done\n");
  }

  set_auto_smooth();

  print_size(fp_log);

  if (mode_memp) {
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
    stime_memp = current_sec();
    stime_lap_memp = elapsed_sec();
#endif
  }

  /** computing membership probs. **/
  if (mode_memp || mode_cluster || mode_rank || mode_dist) {
    allocate_param();
    allocate_prob_mem();
    update_param_no_vbem();
    fprintf(fp_msg,"Computing the membership distributions...");
    fflush(fp_msg);
    compute_member_prob();
    fprintf(fp_msg,"done.\n");
  }
  if (mode_memp) {
    fp_memp = fopen(fname_memp,"wb"); /* binary mode */
    if (fp_memp == NULL)
      die("Couldn't open %s\n", fname_memp);
    print_member_prob(fp_memp);
    fclose(fp_memp);
  }
  if (mode_memp) {
#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
    etime_memp = current_sec();
    etime_lap_memp = elapsed_sec();
#endif
  }

  /** computing distances **/
  if (mode_dist) {
    fprintf(fp_msg,"Computing the (dis)similarities between attribute values...");
    fflush(fp_msg);
    dist_main();
    fprintf(fp_msg,"done.\n");
  }

  /** relevance analysis **/
  if (mode_rank) {
    fprintf(fp_msg,"Ranking relevant words...");
    fflush(fp_msg);
    rank_main();
    fprintf(fp_msg,"done.\n");
  }

  /** clustering **/
  if (mode_cluster) {
    fprintf(fp_msg,"Clustering cases and attribute values...");
    fflush(fp_msg);
    cluster_main();
    fprintf(fp_msg,"done.\n");
  }

  /* cleaning up arrays */
  release_hparam();
  if (mode_memp || mode_cluster || mode_rank || mode_dist) {
    release_param();
    release_prob_mem();
  }

#if defined(HAVE_GETRUSAGE) && defined(HAVE_GETTIMEOFDAY)
  print_time_vb(fp_log);
#endif
}

void vbem_iterate(int u)
{
  int r,r_best=0,n;
  int num_iterate,num_iterate_best=-1;
  unsigned int curr_seed=0;

  for (r=0; r<num_init; r++) { /* loop of restarting */
    if (!mode_set_seed) {
      curr_seed = (unsigned int)rand();
      srand(curr_seed);
    }
  
    initialize_hparam();

    n = (num_init > 1) ? max_init_iterate : max_iterate;

    if (mode_nomem) {
      num_iterate = vbem_iterate_body_nomem(r,0,n);
    }
    else {
      num_iterate = vbem_iterate_body(r,0,n);
    }

    log_seed[u][r] = (!mode_set_seed) ? curr_seed : 0;
    log_num_iterate[u][r] = num_iterate;
    log_final_free_energy[u][r] = final_free_energy;

    if (r == 0 || final_free_energy > final_free_energy_best) {
      r_best = r;
      final_free_energy_best = final_free_energy;
      num_iterate_best = num_iterate;
      save_free_energy_best(num_iterate_best);
      save_hparam_best();
    }
  }

  log_r_best[u] = r_best;

  final_free_energy = final_free_energy_best;
  restore_free_energy_best(num_iterate_best);
  restore_hparam_best();

  num_iterate = num_iterate_best;

  if (num_init > 1) {
    fprintf(fp_msg,"  <<Resumed best hyperparameter set #%d>>\n",r_best);
    if (mode_nomem) {
      num_iterate = vbem_iterate_body_nomem(r_best,num_iterate,max_iterate);
    }
    else {
      num_iterate = vbem_iterate_body(r_best,num_iterate,max_iterate);
    }
  }

  print_free_energy(fp_log,num_class,num_iterate);
}

int vbem_iterate_body(int r, int m, int n)
{
  compute_pi();
  compute_complete_prob();
  compute_incomplete_prob();

  free_energy[m] = compute_free_energy();

  if (num_init > 1) {
    fprintf(fp_msg,"  [%d] #iterations ",r);
  } else {
    fprintf(fp_msg,"  #iterations ");
  }
  print_dot_message(m);
    
  while (1) {
    compute_conditional_prob();
    compute_expect();
    update_hparam();
    
    m++;
    
    compute_pi();
    compute_complete_prob();
    compute_incomplete_prob();
    
    free_energy[m] = compute_free_energy();
    
    print_dot_message(m);

    if (is_decreased(r,m,n)) break;
    if (is_converged(r,m,n)) break;
  }
  
  final_free_energy = free_energy[m];

  return m;
}

int vbem_iterate_body_nomem(int r, int m, int n)
{
  if (num_init > 1) {
    fprintf(fp_msg,"  [%d] #iterations ",r);
  } else {
    fprintf(fp_msg,"  #iterations ");
  }
  print_dot_message(m);

  compute_pi();
  free_energy[m] = compute_expect_nomem();

  while (1) {
    update_hparam();
    m++;
    print_dot_message(m);

    compute_pi();
    free_energy[m] = compute_expect_nomem();

    if (is_decreased(r,m,n)) break;
    if (is_converged(r,m,n)) break;
  }

  final_free_energy = free_energy[m];

  return m;
}

void compute_pi(void)
{
  int k,v,i;
  FLOAT_TYPE sum;
  FLOAT_TYPE psi;

  sum = 0.0;
  for (k=0; k<num_class; k++) {
    sum += hparam_c_upd[k];
  }
  psi = DIGAMMA(sum);

  for (k=0; k<num_class; k++) {
    pi_c[k] = exp(DIGAMMA(hparam_c_upd[k]) - psi);
  }

  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      sum = 0.0;
      for (i=0; i<num_word[v]; i++) {
        sum += hparam_w_upd[k][v][i];
      }
      psi = DIGAMMA(sum);

      for (i=0; i<num_word[v]; i++) {
        pi_w[k][v][i] = exp(DIGAMMA(hparam_w_upd[k][v][i]) - psi);
      }
    }
  }

  if (mode_verbose) print_pi(fp_log);
}

void compute_complete_prob(void)
{
  int k,e,v;

#pragma omp parallel for private(e,v)
  for (k=0; k<num_class; k++) {
    for (e=0; e<num_entry; e++) {
      prob_comp[k][e] = pi_c[k];
      for (v=0; v<tuple_size; v++) {
	prob_comp[k][e] *= pi_w[k][v][freq_list[e].id[v]];
      }
    }
  }

  if (mode_verbose) print_complete_prob(fp_log);
}

void compute_incomplete_prob(void)
{
  int k,e;

#pragma omp parallel for
  for (e=0; e<num_entry; e++)
    prob_incomp[e] = 0.0;

  /* Cannot immediately add a `pragma omp' statement here */
  for (e=0; e<num_entry; e++)
    for (k=0; k<num_class; k++)
      prob_incomp[e] += prob_comp[k][e];

  if (mode_verbose) print_incomplete_prob(fp_log);
}

void compute_conditional_prob(void)
{
  int k,e;
  FLOAT_TYPE p;

#pragma omp parallel for private(e,p)
  for (k=0; k<num_class; k++) {
    for (e=0; e<num_entry; e++) {
      p = prob_incomp[e];
      prob_cond[k][e] = (p > 0) ? prob_comp[k][e]/p : 0.0;
    }
  }

  if (mode_verbose) print_conditional_prob(fp_log);
}

void compute_expect(void)
{
  int k,i,v,e;
  FLOAT_TYPE tmp_e;

#pragma omp parallel for private(e)
  for (k=0; k<num_class; k++) {
    expect_c[k] = 0.0;
    for (e=0; e<num_entry; e++)
      expect_c[k] += freq_list[e].freq * prob_cond[k][e];
  }

#pragma omp parallel for private(i,e,v,tmp_e)
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++)
      for (i=0; i<num_word[v]; i++)
        expect_w[k][v][i] = 0.0;

    for (e=0; e<num_entry; e++) {
      tmp_e = freq_list[e].freq * prob_cond[k][e];
      for (v=0; v<tuple_size; v++)
        expect_w[k][v][freq_list[e].id[v]] += tmp_e;
    }
  }

  if (mode_verbose) print_expectation(fp_log);
}

FLOAT_TYPE compute_free_energy(void)
{
  FLOAT_TYPE fe;
  FLOAT_TYPE sum_c, sum_w;
  int k,v,i,e,f;

  fe = 0.0;
#pragma omp parallel for private(f,k,v,i,sum_c,sum_w) reduction(+: fe)
  for (e=0; e<num_entry; e++) {
    f = freq_list[e].freq;
    if (f > 0) fe += f * MATH_LOG(prob_incomp[e]);
  }
  
  sum_c = 0.0;
  for (k=0; k<num_class; k++) {
    sum_c += hparam_c[k];
  }
  fe += LGAMMA(sum_c);

  sum_c = 0.0;
  for (k=0; k<num_class; k++) {
    sum_c += hparam_c_upd[k];
  }
  fe -= LGAMMA(sum_c);

  for (k=0; k<num_class; k++) {
    fe += LGAMMA(hparam_c_upd[k]) - LGAMMA(hparam_c[k]);
  }

#pragma omp parallel for private(v,i,sum_w) reduction(+: fe)
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      sum_w = 0.0;
      for (i=0; i<num_word[v]; i++) {
	sum_w += hparam_w[k][v][i];
      }
      fe += LGAMMA(sum_w);

      sum_w = 0.0;
      for (i=0; i<num_word[v]; i++) {
	sum_w += hparam_w_upd[k][v][i];
      }
      fe -= LGAMMA(sum_w);

      for (i=0; i<num_word[v]; i++) {
	fe += LGAMMA(hparam_w_upd[k][v][i]) - LGAMMA(hparam_w[k][v][i]);
      }
    }
  }

  for (k=0; k<num_class; k++) {
    fe += (hparam_c[k] - hparam_c_upd[k]) * log(pi_c[k]);
  }

#pragma omp parallel for private(v,i) reduction(+: fe)
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
	fe += (hparam_w[k][v][i] - hparam_w_upd[k][v][i]) * log(pi_w[k][v][i]);
      }
    }
  }

  return fe;
}

FLOAT_TYPE compute_free_energy_nomem(void)
{
  FLOAT_TYPE fe;
  FLOAT_TYPE p_ic,p_c;
  FLOAT_TYPE sum_c, sum_w;
  int k,v,i,e,f;

  fe = 0.0;

#pragma omp parallel for private(f,k,p_ic,p_c,v) reduction(+: fe)
  for (e=0; e<num_entry; e++) {
    f = freq_list[e].freq;
    if (f <= 0) continue;

    p_ic = 0.0;
    for (k=0; k<num_class; k++) {
      p_c = pi_c[k];
      for (v=0; v<tuple_size; v++) {
	p_c *= pi_w[k][v][freq_list[e].id[v]];
      }
      p_ic += p_c;
    }

    fe += f * MATH_LOG(p_ic);
  }
  
  sum_c = 0.0;
  for (k=0; k<num_class; k++) {
    sum_c += hparam_c[k];
  }
  fe += LGAMMA(sum_c);

  sum_c = 0.0;
  for (k=0; k<num_class; k++) {
    sum_c += hparam_c_upd[k];
  }
  fe -= LGAMMA(sum_c);

  for (k=0; k<num_class; k++) {
    fe += LGAMMA(hparam_c_upd[k]) - LGAMMA(hparam_c[k]);
  }

#pragma omp parallel for private(sum_w,v,i) reduction(+: fe)
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      sum_w = 0.0;
      for (i=0; i<num_word[v]; i++) {
	sum_w += hparam_w[k][v][i];
      }
      fe += LGAMMA(sum_w);

      sum_w = 0.0;
      for (i=0; i<num_word[v]; i++) {
	sum_w += hparam_w_upd[k][v][i];
      }
      fe -= LGAMMA(sum_w);

      for (i=0; i<num_word[v]; i++) {
	fe += LGAMMA(hparam_w_upd[k][v][i]) - LGAMMA(hparam_w[k][v][i]);
      }
    }
  }

  for (k=0; k<num_class; k++) {
    fe += (hparam_c[k] - hparam_c_upd[k]) * log(pi_c[k]);
  }

#pragma omp parallel for private(v,i) reduction(+: fe)
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
	fe += (hparam_w[k][v][i] - hparam_w_upd[k][v][i]) * log(pi_w[k][v][i]);
      }
    }
  }

  return fe;
}

FLOAT_TYPE compute_expect_nomem(void)
{
  int k,e,f,i,v;
  FLOAT_TYPE p_ic,p_c,p_e;

#pragma omp parallel for private(v,i)
  for (k=0; k<num_class; k++) {
    expect_c[k] = 0.0;
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
	expect_w[k][v][i] = 0.0;
      }
    }
  }

  /* Cannot immediately add a `pragma omp' statement here */
  for (e=0; e<num_entry; e++) {
    p_ic = 0.0;
    f = freq_list[e].freq;
    
    for (v=0; v<tuple_size; v++) {
      freq_w_id[v] = freq_list[e].id[v];
    }

    for (k=0; k<num_class; k++) {
      p_c = pi_c[k];
      for (v=0; v<tuple_size; v++)
	p_c *= pi_w[k][v][freq_w_id[v]];
      p_ic += p_c;
    }    

    for (k=0; k<num_class; k++) {
      p_e = f * pi_c[k];
      for (v=0; v<tuple_size; v++)
	p_e *= pi_w[k][v][freq_w_id[v]];
      p_e /= p_ic;
      expect_c[k] += p_e;
      for (v=0; v<tuple_size; v++)
	expect_w[k][v][freq_w_id[v]] += p_e;
    }
  }
  
  if (mode_verbose) print_expectation(fp_log);

  return compute_free_energy_nomem();
}

/* We do not parallelize while using the random number generator. */
void initialize_hparam(void)
{
  int k,v,i;
  FLOAT_TYPE r;
  FLOAT_TYPE noise_c,noise_w;

  for (k=0; k<num_class; k++) {
    noise_c = NOISE_RATIO*((rand() % RAND_MAX)/(FLOAT_TYPE)RAND_MAX);
    r = (smooth_c + 1.0)*(1.0 + noise_c);
    hparam_c[k] = r;
  }

  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      for (i=(num_word[v]-1); i>=0; i--) {
        noise_w = NOISE_RATIO*((rand() % RAND_MAX)/(FLOAT_TYPE)RAND_MAX - 0.5);
        r = (smooth_w[v] + 1.0)*(1.0 + noise_w);
        hparam_w[k][v][i] = r;
      }
    }
  }

#pragma omp parallel for
  for (k=0; k<num_class; k++)
    hparam_c_upd[k] = hparam_c[k];

#pragma omp parallel for private(k,i)
  for (v=0; v<tuple_size; v++) {
    for (k=0; k<num_class; k++) {
      for (i=0; i<num_word[v]; i++)
        hparam_w_upd[k][v][i] = hparam_w[k][v][i];
    }
  }

  if (mode_verbose) print_hparam_upd(fp_log);
}

void update_param_vbem(void)
{
  int k,v,i;
  FLOAT_TYPE sum;

  sum = 0.0;
#pragma omp parallel for reduction(+: sum)
  for (k=0; k<num_class; k++) {
    param_c[k] = hparam_c_upd[k];
    sum += hparam_c_upd[k];
  }
#pragma omp parallel for
  for (k=0; k<num_class; k++) {
    param_c[k] /= sum;
  }

#pragma omp parallel for private(v,sum,i)
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      sum = 0.0;
      for (i=0; i<num_word[v]; i++) {
        param_w[k][v][i] = hparam_w_upd[k][v][i];
	sum += hparam_w_upd[k][v][i];
      }
      for (i=0; i<num_word[v]; i++) {
        param_w[k][v][i] /= sum;
      }
    }
  }

  if (mode_verbose) print_param(fp_log);
}

void update_param_no_vbem(void)
{
  int k,v,i;
  FLOAT_TYPE sum;

  sum = 0.0;
#pragma omp parallel for reduction(+: sum)
  for (k=0; k<num_class; k++) {
    param_c[k] = hparam_c[k];
    sum += hparam_c[k];
  }
#pragma omp parallel for
  for (k=0; k<num_class; k++) {
    param_c[k] /= sum;
  }

#pragma omp parallel for private(v,sum,i)
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      sum = 0.0;
      for (i=0; i<num_word[v]; i++) {
        param_w[k][v][i] = hparam_w[k][v][i];
	sum += hparam_w[k][v][i];
      }
      for (i=0; i<num_word[v]; i++) {
        param_w[k][v][i] /= sum;
      }
    }
  }

  if (mode_verbose) print_param(fp_log);
}

void update_hparam(void)
{
  int k,v,i;

#pragma omp parallel for
  for (k=0; k<num_class; k++)
    hparam_c_upd[k] = hparam_c[k] + expect_c[k];

#pragma omp parallel for private(k,i)
  for (v=0; v<tuple_size; v++) {
    for (k=0; k<num_class; k++) {
      for (i=0; i<num_word[v]; i++)
        hparam_w_upd[k][v][i] = hparam_w[k][v][i] + expect_w[k][v][i];
    }
  }

  if (mode_verbose) print_hparam_upd(fp_log);
}

int is_decreased(int r, int m, int n)
{
    if (free_energy[m] < free_energy[m-1] &&
         (free_energy[m-1] - free_energy[m]) > ACCEPT_SMALL_NUMBER) {
    print_decrease(fp_log,r,m,free_energy[m-1],free_energy[m]);
    fprintf(fp_msg,"\nFree energy decreased! -- see the log file\n");
    return 1;
  }
  return 0;
}

int is_converged(int r, int m, int n)
{
  if (m >= MAX_MAX_ITERATE-1) {
    fprintf(fp_msg,conv_msg(" (Stopped: %d iterations)\tF=%.6Lf\n"),m,free_energy[m]);
    return 1;
  } else if (n > 0 && m >= n) {
    fprintf(fp_msg,conv_msg(" (Stopped: %d iterations)\tF=%.6Lf\n"),m,free_energy[m]);
    return 1;
  } else if (max_iterate > 0 && m >= max_iterate) {
    fprintf(fp_msg,conv_msg(" (Stopped: %d iterations)\tF=%.6Lf\n"),m,free_energy[m]);
    return 1;
  } else {
    if (free_energy[m] - free_energy[m-1] < epsilon) {
      fprintf(fp_msg,conv_msg(" (Converged: %d iterations)\tF=%.6Lf\n"),m,free_energy[m]);
      return 1;
    }
  }
  return 0;
}

void compute_member_prob(void)
{
  int k,v,i,e;
  FLOAT_TYPE p_c, p_ic;

#pragma omp parallel for private(i)
  for (v=0; v<tuple_size; v++)
    for (i=0; i<num_word[v]; i++)
      prob_w[v][i] = 0.0;

  /* Cannot immediately add a `pragma omp' statement here */
  for (k=0; k<num_class; k++)
    for (v=0; v<tuple_size; v++)
      for (i=0; i<num_word[v]; i++)
        prob_w[v][i] += (prob_att_mem[k][v][i] = param_c[k] * param_w[k][v][i]);

  /* Normalization */
#pragma omp parallel for private(v,i)
  for (k=0; k<num_class; k++)
    for (v=0; v<tuple_size; v++)
      for (i=0; i<num_word[v]; i++)
        prob_att_mem[k][v][i] /= prob_w[v][i];

#pragma omp parallel for private(v,p_ic,k,p_c)
  for (e=0; e<num_entry; e++) {
    for (v=0; v<tuple_size; v++)
      freq_w_id[v] = freq_list[e].id[v];

    p_ic = 0.0;
    for (k=0; k<num_class; k++) {
      p_c = param_c[k];
      for (v=0; v<tuple_size; v++)
	p_c *= param_w[k][v][freq_w_id[v]];
      prob_case_mem[k][e] = p_c;
      p_ic += p_c;
    }
    
    for (k=0; k<num_class; k++) {
      prob_case_mem[k][e] /= p_ic;
    }
  }

  if (mode_verbose) print_member_prob(fp_log);
}

void save_hparam_opt(void)
{
  int k,v,i;

  hparam_c_opt = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class_opt);
  hparam_c_upd_opt = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class_opt);
  hparam_w_opt = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class_opt);
  hparam_w_upd_opt = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class_opt);
  if (hparam_c_opt == NULL || hparam_w_opt == NULL ||
      hparam_c_upd_opt == NULL || hparam_w_upd_opt == NULL)
    die("Couldn't allocate memory for hparam_{c,w}_opt or hparam_{c,w}_opt_upd\n");

#pragma omp parallel for private(v)
  for (k=0; k<num_class_opt; k++) {
    hparam_w_opt[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    hparam_w_upd_opt[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (hparam_w_opt[k] == NULL || hparam_w_upd_opt[k] == NULL)
      die("Couldn't allocate memory for hparam_w_opt{,_upd}[k]\n");
    for (v=0; v<tuple_size; v++) {
      hparam_w_opt[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      hparam_w_upd_opt[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      if (hparam_w_opt[k][v] == NULL || hparam_w_upd_opt[k][v] == NULL)
        die("Couldn't allocate memory for hparam_w_opt{,_upd}[k][v]\n");
    }
  }

#pragma omp parallel for private(v,i)
  for (k=0; k<num_class_opt; k++) {
    hparam_c_opt[k] = hparam_c[k];
    hparam_c_upd_opt[k] = hparam_c_upd[k];
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
	hparam_w_opt[k][v][i] = hparam_w[k][v][i];
	hparam_w_upd_opt[k][v][i] = hparam_w_upd[k][v][i];
      }
    }
  }
}

void restore_hparam_opt(void)
{
  int k,v,i;

  hparam_c = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class_opt);
  hparam_c_upd = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class_opt);
  hparam_w = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class_opt);
  hparam_w_upd = (FLOAT_TYPE ***)malloc(sizeof(FLOAT_TYPE **)*num_class_opt);
  if (hparam_c == NULL || hparam_w == NULL ||
      hparam_c_upd == NULL || hparam_w_upd == NULL)
    die("Couldn't allocate memory for hparam_{c,w}{,_upd}\n");

#pragma omp parallel for private(v)
  for (k=0; k<num_class_opt; k++) {
    hparam_w[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    hparam_w_upd[k] = (FLOAT_TYPE **)malloc(sizeof(FLOAT_TYPE *)*tuple_size);
    if (hparam_w[k] == NULL || hparam_w_upd[k] == NULL)
      die("Couldn't allocate memory for hparam_w{,_upd}[k]\n");
    for (v=0; v<tuple_size; v++) {
      hparam_w[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      hparam_w_upd[k][v] = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_word[v]);
      if (hparam_w[k][v] == NULL || hparam_w_upd[k][v] == NULL)
        die("Couldn't allocate memory for hparam_w{,_upd}[k][v]\n");
    }
  }

#pragma omp parallel for private(v,i)
  for (k=0; k<num_class_opt; k++) {
    hparam_c[k] = hparam_c_opt[k];
    hparam_c_upd[k] = hparam_c_upd_opt[k];
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
        hparam_w[k][v][i] = hparam_w_opt[k][v][i];
        hparam_w_upd[k][v][i] = hparam_w_upd_opt[k][v][i];
      }
    }
  }
}

void save_free_energy_best(int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++) {
    free_energy_best[m] = free_energy[m];
  }
}

void restore_free_energy_best(int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++) {
    free_energy[m] = free_energy_best[m];
  }
}

void save_hparam_best(void)
{
  int k,v,j;

#pragma omp parallel for private(v,j)
  for (k=0; k<num_class; k++) {
    hparam_c_best[k] = hparam_c[k];
    hparam_c_upd_best[k] = hparam_c_upd[k];
    for (v=0; v<tuple_size; v++) {
      for (j=0; j<num_word[v]; j++) {
	hparam_w_best[k][v][j] = hparam_w[k][v][j];
	hparam_w_upd_best[k][v][j] = hparam_w_upd[k][v][j];
      }
    }
  }
}

void restore_hparam_best(void)
{
  int k,v,j;

#pragma omp parallel for private(v,j)
  for (k=0; k<num_class; k++) {
    hparam_c[k] = hparam_c_best[k];
    hparam_c_upd[k] = hparam_c_upd_best[k];
    for (v=0; v<tuple_size; v++) {
      for (j=0; j<num_word[v]; j++) {
	hparam_w[k][v][j] = hparam_w_best[k][v][j];
	hparam_w_upd[k][v][j] = hparam_w_upd_best[k][v][j];
      }
    }
  }
}
