#include "nbc_def.h"
#include "nbc_ext.h"
#include "vnbc_def.h"
#include "vnbc_ext.h"

void vbsmem_iterate(int u)
{
  int t,h;
  FLOAT_TYPE last_free_energy;
  FLOAT_TYPE curr_jsplit, curr_jmerge;
  int sm_ok;
  int n,n0;
  int m1,m2;
  int k1,k2,k3,last_k1=-1,last_k2=-1;
  int num_tried_cand, num_tried_jsplit_cand=-1;
  int num_iterate,num_iterate_best=-1;
  int r,r_best=-1;
  unsigned int curr_seed=0;
  int *cand;

  num_candidate = num_class*(num_class - 1)*(num_class - 2)/2;
  n = (max_iterate > 0) ? max_iterate : MAX_MAX_ITERATE;
  if (num_init > 1) {
    n0 = (max_init_iterate > 0) ? max_init_iterate : n;
  } else {
    n0 = (max_iterate > 0) ? max_iterate : MAX_MAX_ITERATE;
  }
  
  allocate_jmerge();
  allocate_jsplit();
  allocate_candidate();
  allocate_hparam_sm();
  allocate_free_energy_sm();
  allocate_num_iterate_sm();
  allocate_num_result_sm();
  allocate_hparam_part();
  allocate_prob_part();
  allocate_param();
  allocate_prob_mem();
  if (mode_rank_sm) {
    allocate_prob_att_mem_sm();
    allocate_rank_sm();
    allocate_param_sm();
  }

  for (t=0; t<(num_sm+1); t++) {
    num_iterate_sm_full[t] = 0;
    num_iterate_sm_part[t] = 0;
  }

  for (t=0; t<num_sm; t++) {
    num_result_rejected[t] = 0;
    num_result_skipped[t] = 0;
  }

  for (r=0; r<num_init; r++) {
    if (!mode_set_seed) {
      curr_seed = (unsigned int)rand();
      srand(curr_seed);
    }

    fprintf(fp_msg,"    [%d] #iterations ",r);

    initialize_hparam();

    num_iterate = vbsmem_iterate_full(0,0,n0);

    num_iterate_sm_full[0] = num_iterate;

    log_seed[u][r] = (!mode_set_seed) ? curr_seed : 0;
    log_num_iterate[u][r] = num_iterate;
    log_final_free_energy[u][r] = final_free_energy;

    if (r == 0 || final_free_energy > final_free_energy_best) {
      r_best = r;
      final_free_energy_best = final_free_energy;
      num_iterate_best = num_iterate;
      save_free_energy_best_sm(0,num_iterate_best);
      save_hparam_best();
    }
  }
  log_r_best[u] = r_best;

  final_free_energy = final_free_energy_best;
  restore_hparam_best();
  restore_free_energy_best_sm(0,num_iterate_best);
  
  last_free_energy = final_free_energy;

  for (t=0; t<num_sm; t++) {

    update_param_vbem();

    if (mode_rank_sm) {
      compute_member_prob_sm();
      compute_rank_sm();
    }
    
    save_hparam_sm();
    if (mode_rank_sm) save_param_sm();

    compute_jmerge();
    compute_jsplit();
    sort_candidate();
    
    sm_ok = 0;
    num_tried_cand = 0;
    for (h=0; (max_candidate==0 || num_tried_cand<max_candidate) && h<num_candidate; h++) {

      cand = sm_candidate[h];
      k1 = cand[0];
      k2 = cand[1];
      k3 = cand[2];
      curr_jmerge = jmerge[k1][k2];
      curr_jsplit = jsplit[k3];

      fprintf(fp_msg,conv_msg("  [Op=%d]\n"
			      "    Candidate %d (%d not skipped): Jmerge(%d,%d)=%.12Lf / Jsplit(%d)=%.12Lf\n"),
	      (t+1),h,num_tried_cand,k1,k2,curr_jmerge,k3,curr_jsplit);

      if (h==0 || last_k1 != k1 || last_k2 != k2) {
	num_tried_jsplit_cand = 0;
      }	else {
	num_tried_jsplit_cand++;
      }

      if (max_jsplit>0 && num_tried_jsplit_cand >= max_jsplit) {
	fprintf(fp_msg,"    -- Skipped (the number of examined split candidates reached the limit)\n");
	num_result_skipped[t]++;
	continue;
      }
      if ((min_param_c>0 && param_c[k1] < min_param_c && param_c[k2] < min_param_c)
	  || (min_param_c>0 && param_c[k3] < min_param_c)) {
	fprintf(fp_msg,"    -- Skipped (too small class probability)\n");
	num_result_skipped[t]++;
	continue;
      }

      num_tried_cand++;
      last_k1 = k1;
      last_k2 = k2;

      fprintf(fp_msg,"    #iterations ");

      m1 = vbsmem_iterate_part((t+1),h,0,n);
      m2 = vbsmem_iterate_full((t+1),m1,n);

      if (final_free_energy > last_free_energy) {
	if ((final_free_energy - last_free_energy) > min_smem_improve) {
	  if (!(min_new_param_c > 0) ||
	   (param_c[k1] > min_new_param_c &&
	    param_c[k2] > min_new_param_c &&
	    param_c[k3] > min_new_param_c)) {
	    
	    fprintf(fp_msg,conv_msg("    -- Accepted [%d,%d]->[%d] & [%d]->[%d,%d] "
				    "(F current: %.9Lf / last: %.9Lf)\n"),
		    k1,k2,k1,k3,k2,k3,final_free_energy,last_free_energy);
	    sm_ok = 1;
	  }
	  else {
	    fprintf(fp_msg,conv_msg("    -- Rejected [%d,%d]->[%d] & [%d]->[%d,%d] "
				    "(F current: %.9Lf / last: %.9Lf)\n"),
		    k1,k2,k1,k3,k2,k3,final_free_energy,last_free_energy);
	    fprintf(fp_msg,conv_msg("       <Too small split/merged class probabilities: "
				    "%d - %.9Lf, %d - %.9Lf, %d - %.9Lf>\n"),
		    k1,param_c[k1],k2,param_c[k2],k3,param_c[k3]);
	    sm_ok = 0;
	  }
	}
	else {
	  fprintf(fp_msg,conv_msg("    -- Rejected [%d,%d]->[%d] & [%d]->[%d,%d] "
				  "(F current: %.9Lf / last: %.9Lf)\n"),
		  k1,k2,k1,k3,k2,k3,final_free_energy,last_free_energy);
	  fprintf(fp_msg,conv_msg("       <Too small improvement: %.9Lf>\n"),
		  (final_free_energy - last_free_energy));
	  sm_ok = 0;
	}

      } else {
	fprintf(fp_msg,conv_msg("    -- Rejected [%d,%d]->[%d] & [%d]->[%d,%d] "
				"(F current: %.9Lf / last: %.9Lf)\n"),
		k1,k2,k1,k3,k2,k3,final_free_energy,last_free_energy);
	fprintf(fp_msg,conv_msg("       <No improvement>\n"));
	sm_ok = 0;
      }

      if (sm_ok) {
	last_free_energy = final_free_energy;
	num_iterate_sm_part[t+1] = m1;
	num_iterate_sm_full[t+1] = m2;

	if (mode_rank_sm) {
	  update_param_vbem();
	  compute_member_prob_part(cand);
	  compute_rank_part(cand);
	  print_rank_sm(fp_sm,t,cand);
	}

	break;
      }
      else {
	num_result_rejected[t]++;
	restore_hparam_sm();
	if (mode_rank_sm) restore_param_sm();
      }

    }

    if (sm_ok == 0) {
      if (h == num_candidate) {
	fprintf(fp_msg,"  [Op=%d]\n  <<All candidates are rejected>>\n",(t+1));
      } else {
	fprintf(fp_msg,"  [Op=%d]\n  <<The number of rejected split candidates reached the limit>>\n",(t+1));
      }
      break;
    }
  }
  
  print_num_iterate_sm(fp_log,num_class);
  print_free_energy_sm(fp_log,num_class);
  print_num_result_sm(fp_log,num_class);

  release_jmerge();
  release_jsplit();
  release_candidate();
  release_hparam_sm();
  release_free_energy_sm();
  release_num_iterate_sm();
  release_num_result_sm();
  release_hparam_part();
  release_prob_part();
  if (mode_rank_sm) {
    release_param();
    release_param_sm();
    release_prob_att_mem_sm();
    release_rank_sm();
  }
}

int vbsmem_iterate_full(int t, int m, int n)
{
  compute_pi();
  compute_complete_prob();
  compute_incomplete_prob();
    
  free_energy_sm[t][m] = compute_free_energy();
  print_dot_message(m);

  while (1) {
    compute_conditional_prob();
    compute_expect();
    update_hparam();

    m++;

    compute_pi();
    compute_complete_prob();
    compute_incomplete_prob();

    free_energy_sm[t][m] = compute_free_energy();

    print_dot_message(m);

    if (free_energy_sm[t][m] < free_energy_sm[t][m-1] &&
	free_energy_sm[t][m-1] - free_energy_sm[t][m] > ACCEPT_SMALL_NUMBER) {
      fprintf(fp_msg,conv_msg("\n(full) [%d] Free energy decreased! "
			      "(curr: %.12Lf / last: %.12Lf)\n"),
	      m,free_energy_sm[t][m],free_energy_sm[t][m-1]);
      /* print_param(fp_msg); */
      break;
    }

    if (m >= n) {
      fprintf(fp_msg,conv_msg(" (Stopped: %d iterations)\tF=%.6Lf\n"),
	      m,free_energy_sm[t][m]);
      break;
    }
    else {
      if (free_energy_sm[t][m] - free_energy_sm[t][m-1] < epsilon) {
	fprintf(fp_msg,conv_msg(" (Converged: %d iterations)\tF=%.6Lf\n"),
		m,free_energy_sm[t][m]);
	break;
      }
    }
  }

  final_free_energy = free_energy_sm[t][m];

  return m;
}

int vbsmem_iterate_part(int t, int h, int m, int n)
{
  int *cand;

  cand = sm_candidate[h];

  initialize_hparam_part(cand);
  save_hparam_part(cand);

  compute_pi();
  compute_complete_prob();
  compute_incomplete_prob();
  save_incomplete_prob_part(cand);

  free_energy_sm[t][m] = compute_free_energy();
  
  print_dot_message(m);

  while (1) {
    compute_conditional_prob_part(cand);
    compute_expect_part(cand);
    update_hparam_part(cand);

    m++;

    compute_pi_part(cand);
    compute_complete_prob_part(cand);
    compute_incomplete_prob_part(cand);

    free_energy_sm[t][m] = compute_free_energy();

    print_dot_message(m);

    if (free_energy_sm[t][m] < free_energy_sm[t][m-1]
	&& free_energy_sm[t][m-1] - free_energy_sm[t][m] > ACCEPT_SMALL_NUMBER_PART) {
      fprintf(fp_msg,conv_msg("\n(part) [%d] Free energy decreased! "
			      "(curr: %.12Lf / last: %.12Lf / diff: %.12Lf)\n"),
	      m,free_energy_sm[t][m],free_energy_sm[t][m-1],
	      (free_energy_sm[t][m-1] - free_energy_sm[t][m]));
      break;
    }

    if (m >= n) {
      fprintf(fp_msg,"(S:%d)",m);
      break;
    } else {
      if (free_energy_sm[t][m] - free_energy_sm[t][m-1] < epsilon) {
	fprintf(fp_msg,"(C:%d)",m);
	break;
      }
    }
  }

  /* No need to update final_free_energy, etc.  */

  return m;
}

void compute_jmerge(void)
{
  int k,k1,k2,e;
  FLOAT_TYPE *norm;

  if (mode_norm) {

    norm = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*num_class);
    if (norm == NULL) die("Couldn't allocate memory for norm\n");

#pragma omp parallel for private(e)
    for (k=0; k<num_class; k++) {
      norm[k] = 0.0;
      for (e=0; e<num_entry; e++) {
	norm[k] += freq_list[e].freq * prob_cond[k][e] * prob_cond[k][e];
      }
    }
    
#pragma omp parallel for private(k2,e)
    for (k1=0; k1<num_class; k1++) {
      for (k2=(k1+1); k2<num_class; k2++) {
	jmerge[k1][k2] = 0.0;
	for (e=0; e<num_entry; e++) {
	  jmerge[k1][k2] += freq_list[e].freq * prob_cond[k1][e] * prob_cond[k2][e];
	}
	jmerge[k1][k2] /= sqrt(norm[k1]*norm[k2]);
      }
    }
    
    free(norm); norm = NULL;

  } else {

#pragma omp parallel for private(k2,e)
    for (k1=0; k1<num_class; k1++) {
      for (k2=(k1+1); k2<num_class; k2++) {
	jmerge[k1][k2] = 0.0;
	for (e=0; e<num_entry; e++) {
	  jmerge[k1][k2] += freq_list[e].freq * prob_cond[k1][e] * prob_cond[k2][e];
	}
      }
    }
    
  }

}

void compute_jsplit(void)
{
  int e,k,v;
  FLOAT_TYPE p_emp;
  FLOAT_TYPE pw;
  FLOAT_TYPE tmp_jsplit;

#pragma omp parallel for private(e,tmp_jsplit,p_emp,v,pw)
  for (k=0; k<num_class; k++) {
    jsplit[k] = 0.0;
    if (!(expect_c[k] > 0)) continue;

    for (e=0; e<num_entry; e++) {

      tmp_jsplit = 0.0;
      p_emp = freq_list[e].freq * prob_cond[k][e] / expect_c[k];

      if (p_emp > 0) tmp_jsplit += log(p_emp);

      for (v=0; v<tuple_size; v++) {
	pw = param_w[k][v][freq_list[e].id[v]];
	if (p_emp > 0 && pw > 0) tmp_jsplit -= log(pw);
      }

      jsplit[k] += p_emp * tmp_jsplit;
    }
  }
}

void initialize_hparam_part(int *cand)
{
  int i,k1,k2,k3,v;
  FLOAT_TYPE r1,r2;
  FLOAT_TYPE noise;
  FLOAT_TYPE p1,p2,p3;
  FLOAT_TYPE q1,q2,q3;

  k1 = cand[0];
  k2 = cand[1];
  k3 = cand[2];

  q1 = hparam_c_upd[k1];
  q2 = hparam_c_upd[k2];
  q3 = hparam_c_upd[k3];

  hparam_c_upd[k1] = hparam_c[k1] + (q1 - hparam_c[k1]) + (q2 - hparam_c[k2]);
  hparam_c_upd[k2] = hparam_c[k2] + 0.5 * (q3 - hparam_c[k3]);
  hparam_c_upd[k3] = hparam_c[k3] + 0.5 * (q3 - hparam_c[k3]);

  r1 = expect_c[k1]/(expect_c[k1] + expect_c[k2]);
  r2 = expect_c[k2]/(expect_c[k1] + expect_c[k2]);

  /* To keep the repeatability, we do not use OpenMP here */
  for (v=0; v<tuple_size; v++) {
    for (i=0; i<num_word[v]; i++) {
      p1 = hparam_w_upd[k1][v][i] - hparam_w[k1][v][i];
      p2 = hparam_w_upd[k2][v][i] - hparam_w[k2][v][i];
      p3 = hparam_w_upd[k3][v][i] - hparam_w[k3][v][i];
      noise = NOISE_RATIO * (rand() % RAND_MAX)/(FLOAT_TYPE)RAND_MAX;
      hparam_w_upd[k1][v][i] = hparam_w[k1][v][i] + r1 * p1 + r2 * p2;
      hparam_w_upd[k2][v][i] = hparam_w[k2][v][i] + p3 * (1.0 - noise);
      hparam_w_upd[k3][v][i] = hparam_w[k3][v][i] + p3 * (1.0 + noise);
    }
  }

}

void sort_candidate(void)
{
  int k1,k2,k3;
  int h;

  h = 0;
  for (k1=0; k1<num_class; k1++) {
    for (k2=(k1+1); k2<num_class; k2++) {
      for (k3=0; k3<num_class; k3++) {
	if (k3 == k1 || k3 == k2) continue;
	sm_candidate[h][0] = k1;
	sm_candidate[h][1] = k2;
	sm_candidate[h][2] = k3;
	h++;
      }
    }
  }

  tmp_jmerge = jmerge;
  tmp_jsplit = jsplit;

  qsort(sm_candidate,num_candidate,sizeof(int *),
	(int(*)(const void *, const void *))compare_candidate);
}

int compare_candidate(const int **x, const int **y)
{
  int k1_x = (*x)[0];
  int k2_x = (*x)[1];
  int k3_x = (*x)[2];
  int k1_y = (*y)[0];
  int k2_y = (*y)[1];
  int k3_y = (*y)[2];

  FLOAT_TYPE jm_x = tmp_jmerge[k1_x][k2_x];
  FLOAT_TYPE jm_y = tmp_jmerge[k1_y][k2_y];
  FLOAT_TYPE js_x = tmp_jsplit[k3_x];
  FLOAT_TYPE js_y = tmp_jsplit[k3_y];

  if (k1_x == k1_y && k2_x == k2_y) {
    if (js_x > js_y) return -1;
    if (js_x < js_y) return 1;
    return 0;
  }
  else {
    if (jm_x > jm_y) return -1;
    if (jm_x < jm_y) return 1;
    return 0;
  }
}

void compute_pi_part(int *cand)
{
  int l,k,v,i;
  
  FLOAT_TYPE sum;
  FLOAT_TYPE psi;

  sum = 0.0;
  for (k=0; k<num_class; k++) {
    sum += hparam_c_upd[k];
  }
  psi = DIGAMMA(sum);

  for (l=0; l<3; l++) {
    k = cand[l];
    pi_c[k] = exp(DIGAMMA(hparam_c_upd[k]) - psi);
  }

  for (l=0; l<3; l++) {
    k = cand[l];
#pragma omp parallel for private(sum,i,psi)
    for (v=0; v<tuple_size; v++) {
      sum = 0.0;
      for (i=0; i<num_word[v]; i++) {
        sum += hparam_w_upd[k][v][i];
      }
      psi = DIGAMMA(sum);

      for (i=0; i<num_word[v]; i++) {
        pi_w[k][v][i] = exp(DIGAMMA(hparam_w_upd[k][v][i]) - psi);
      }
    }
  }
}

void compute_complete_prob_part(int *cand)
{
  int l,k,v,e;
  
  for (l=0; l<3; l++) {
    k = cand[l];
#pragma omp parallel for private(v)
    for (e=0; e<num_entry; e++) {
      prob_comp[k][e] = pi_c[k];
      for (v=0; v<tuple_size; v++) {
	prob_comp[k][e] *= pi_w[k][v][freq_list[e].id[v]];
      }
    }
  }
}

void compute_conditional_prob_part(int *cand)
{
  int l,k;
  int e;
  FLOAT_TYPE p;

  for (l=0; l<3; l++) {
    k = cand[l];
#pragma omp parallel for private(p)
    for (e=0; e<num_entry; e++) {
      p = prob_incomp[e];
      prob_cond[k][e] = (p > 0) ? prob_comp[k][e]/p : 0.0;
    }
  }
}

void compute_expect_part(int *cand)
{
  int l,k,i,e,v;
  FLOAT_TYPE tmp_e;
  FREQ_PTR fp;
  FLOAT_TYPE e_c_k;

  for (l=0; l<3; l++) {
    k = cand[l];
    e_c_k = 0.0;
#pragma omp parallel for reduction(+: e_c_k)
    for (e=0; e<num_entry; e++) {
      e_c_k += freq_list[e].freq * prob_cond[k][e];
    }
    expect_c[k] = e_c_k;
  }

  for (l=0; l<3; l++) {
    k = cand[l];

#pragma omp parallel for private(i)
    for (v=0; v<tuple_size; v++)
      for (i=0; i<num_word[v]; i++)
	expect_w[k][v][i] = 0.0;

    /* Cannot immediately add a `pragma omp' statement here */
    for (e=0; e<num_entry; e++) {
      fp = &freq_list[e];
      tmp_e = fp->freq * prob_cond[k][e];
      for (v=0; v<tuple_size; v++) {
	expect_w[k][v][fp->id[v]] += tmp_e;
      }
    }
  }

}

void update_hparam_part(int *cand)
{
  int k,l,i,v;
  FLOAT_TYPE expect_c_sum0, expect_c_sum;

  expect_c_sum0 = 0.0;
  for (l=0; l<3; l++) {
    k = cand[l];
    expect_c_sum0 += (hparam_c_upd_part[l] - hparam_c[k]);
  }

  expect_c_sum = 0.0;
  for (l=0; l<3; l++) {
    k = cand[l];
    expect_c_sum += expect_c[k];
  }
  for (l=0; l<3; l++) {
    k = cand[l];
    hparam_c_upd[k] = hparam_c[k] + expect_c_sum0 * expect_c[k] / expect_c_sum;
  }

  for (l=0; l<3; l++) {
    k = cand[l];
#pragma omp parallel for private(i)
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
	hparam_w_upd[k][v][i] = hparam_w[k][v][i] + expect_w[k][v][i];
      }
    }
  }

}

void compute_incomplete_prob_part(int *cand)
{
  int l,k,e;

#pragma omp parallel for private(l,k)
  for (e=0; e<num_entry; e++) {
    prob_incomp[e] = prob_incomp_part[e];
    for (l=0; l<3; l++) {
      k = cand[l];
      prob_incomp[e] += prob_comp[k][e];
    }
  }
}

void compute_member_prob_sm(void)
{
  int k,v,i;
  FLOAT_TYPE p;

  #pragma omp parallel for private(i)
  for (v=0; v<tuple_size; v++)
    for (i=0; i<num_word[v]; i++)
      prob_w_sm[v][i] = 0.0;

  /* Cannot immediately add a `pragma omp' statement here */
  for (k=0; k<num_class; k++) {
    p = param_c[k];
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
	prob_w_sm[v][i] +=
	  (prob_att_mem_sm[k][v][i] = p * param_w[k][v][i]);
      }
    }
  }

#pragma omp parallel for private(v,i)
  for (k=0; k<num_class; k++) {
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
	prob_att_mem_sm[k][v][i] /= prob_w_sm[v][i];
      }
    }
  }
}

void compute_member_prob_part(int *cand)
{
  int k,i,l,v;
  FLOAT_TYPE p,q;
  
  #pragma omp parallel for private(i)
  for (v=0; v<tuple_size; v++)
    for (i=0; i<num_word[v]; i++)
      prob_w_sm[v][i] = 0.0;

  /* Cannot immediately add a `pragma omp' statement here */
  for (k=0; k<num_class; k++) {
    p = param_c[k];
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
	q = p * param_w[k][v][i];
	if (k == cand[0]) prob_att_mem_part[0][v][i] = q;
	if (k == cand[1]) prob_att_mem_part[1][v][i] = q;
	if (k == cand[2]) prob_att_mem_part[2][v][i] = q;
	prob_w_sm[v][i] += q;
      }
    }
  }

  for (l=0; l<3; l++) {
#pragma omp parallel for private(i)
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
	prob_att_mem_part[l][v][i] /= prob_w_sm[v][i];
      }
    }
  }
}

void save_free_energy_best_sm(int t,int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++) {
    free_energy_best[m] = free_energy_sm[t][m];
  }
}

void restore_free_energy_best_sm(int t,int num_iterate)
{
  int m;

  for (m=0; m<num_iterate; m++) {
    free_energy_sm[t][m] = free_energy_best[m];
  }
}

void save_hparam_sm(void)
{
  int k,v,i;

#pragma omp parallel for private(v,i)
  for (k=0; k<num_class; k++) {
    hparam_c_sm[k] = hparam_c[k];
    hparam_c_upd_sm[k] = hparam_c_upd[k];
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
	hparam_w_sm[k][v][i] = hparam_w[k][v][i];
	hparam_w_upd_sm[k][v][i] = hparam_w_upd[k][v][i];
      }
    }
  }
}

void restore_hparam_sm(void)
{
  int k,v,i;
  
#pragma omp parallel for private(v,i)
  for (k=0; k<num_class; k++) {
    hparam_c[k] = hparam_c_sm[k];
    hparam_c_upd[k] = hparam_c_upd_sm[k];
    for (v=0; v<tuple_size; v++) {
      for (i=0; i<num_word[v]; i++) {
	hparam_w[k][v][i] = hparam_w_sm[k][v][i];
	hparam_w_upd[k][v][i] = hparam_w_upd_sm[k][v][i];
      }
    }
  }
}

void save_hparam_part(int *cand)
{
  int l,k;

  for (l=0; l<3; l++) {
    k = cand[l];
    hparam_c_upd_part[l] = hparam_c_upd[k];
  }
}

void save_incomplete_prob_part(int *cand)
{
  int k,e;
  int k1,k2,k3;

  k1 = cand[0];
  k2 = cand[1];
  k3 = cand[2];

  #pragma omp parallel for
  for (e=0; e<num_entry; e++) {
    prob_incomp_part[e] = 0.0;
  }

  /* Cannot immediately add a `pragma omp' statement here */ 
  for (k=0; k<num_class; k++) {
    if (k == k1 || k == k2 || k == k3) continue;
    for (e=0; e<num_entry; e++) {
      prob_incomp_part[e] += prob_comp[k][e];
    }
  }
}

void save_param_sm(void)
{
  int k,v,i;

#pragma omp parallel for private(v,i)
  for (k=0; k<num_class; k++) {
    param_c_sm[k] = param_c[k];
    for (v=0; v<tuple_size; v++)
      for (i=0; i<num_word[v]; i++)
	param_w_sm[k][v][i] = param_w[k][v][i];
  }
}

void restore_param_sm(void)
{
  int k,v,i;

#pragma omp parallel for private(v,i)
  for (k=0; k<num_class; k++) {
    param_c[k] = param_c_sm[k];
    for (v=0; v<tuple_size; v++)
      for (i=0; i<num_word[v]; i++)
	param_w[k][v][i] = param_w_sm[k][v][i];
  }
}
