#include "nbc_def.h"
#include "nbc_var.h"
#include "vnbc_def.h"
#include "vnbc_var.h"

#ifdef USE_ICONV
#  define OPT_STRING "f:k:e:m:hvc:w:r:ip:q:jdn:l:s:t:z:o:a:I:G:U:FN:C:MQ:Rx:YDSEV:"
#else
#  define OPT_STRING "f:k:e:m:hvc:w:r:ijdn:l:s:t:z:o:a:I:G:U:FN:C:MQ:Rx:YDSEV:"
#endif

int main(int argc, char *argv[])
{
  int v;

  initialize_modes();
  allocate_buffer();
#ifdef USE_ICONV  
  initialize_iconv_sys();
#endif

  parse_args(argc,argv);
  open_log();

#ifdef USE_ICONV
  initialize_iconv();
#endif
  initialize_rand();

  print_settings(fp_log);

  if (mode_em)
    vbem_main();
  else
    no_vbem_main();

  fclose(fp_log);
#ifdef USE_ICONV
  iconv_close(cd_in);
  iconv_close(cd_out);
#endif

  for (v=0; v<tuple_size; v++) {
    release_hashtable(&ht[v]);
  }

  release_buffer();

  free(smooth_w); smooth_w = NULL;
  free(mode_auto_smooth_w); mode_auto_smooth_w = NULL;

  return 0;
}

void initialize_modes()
{
  fname_base[0] = '\0';

  num_class = 0;
  tuple_size = 0;
  kmin = -1;
  kmax = -1;
  kstep = -1;
  epsilon = DEFAULT_EPSILON;
  max_iterate = DEFAULT_MAX_ITERATE;
  rand_seed = DEFAULT_RAND_SEED;
  smooth_c = 0.0;
  smooth_w = NULL;
  ic_type = VB;
  rel_m = REL_MP;
  init_type_class = NOISY_UNIFORM_C;
  init_type_word = NOISY_UNIFORM_W;
  mode_verbose = 0;
  mode_msg = 0;
  mode_csv_in = 0;
  mode_csv_out = 0;
#ifdef USE_ICONV
#ifdef CHARCODE
  strncpy(charcode_in,CHARCODE,CHARCODE_LEN);
  strncpy(charcode_out,CHARCODE,CHARCODE_LEN);
#else
  strncpy(charcode_in,CHARCODE_INTERNAL,CHARCODE_LEN);
  strncpy(charcode_out,CHARCODE_INTERNAL,CHARCODE_LEN);
#endif
#endif
  mode_em = 1;
  mode_smooth = 0;
  mode_smooth_strict = 0;
  mode_auto_smooth = 0;
  mode_auto_smooth_c = MODE_SMOOTH_NONE;
  mode_auto_smooth_w = NULL;
  mode_bdeu = 0;
  mode_memp = 0;
  mode_dist = 0;
  mode_as_is = 0;
  mode_eval = 0;
  mode_search_num_class = 0;
  mode_suffix = 0;
  num_init = 1;
  max_init_iterate = 0;
  mode_cluster = 0;
  mode_set_seed = 0;
  mode_lex_order = 1;
  mode_rank = 0;
  num_sm = 0;
  max_candidate = 0;
  max_jsplit = 0;
  mode_rank_sm = 0;
  num_rank_sm = 0;
  mode_norm = 0;
  min_param_c = 0;
  min_smem_improve = -1;
  min_new_param_c = 0;
  mode_nomem = 0;

#ifdef USE_ICONV
#  ifdef CHARCODE
  mode_sys_out  = (strcmp(CHARCODE,CHARCODE_INTERNAL) == 0) ? 0 : 1;
#  endif
#endif
}

void parse_args(int argc, char *argv[])
{
  int c,v;
  int num_smooth_w = 0;
  FLOAT_TYPE tmp_smooth_w;
  enum smooth_auto_type tmp_mode_auto_smooth_w;

  while (1) {
    c = getopt(argc,argv,OPT_STRING);
    if (c == -1) break;

    switch (c) {
    case 'a':
      max_candidate = get_int(optarg);
      break;
    case 'c':
      smooth_c = get_smooth_c(optarg);
      break;
    case 'C':
      min_new_param_c = get_float(optarg);
      break;
    case 'd':
      mode_dist = 1;
      break;
    case 'D':
      mode_memp = 1;
      break;
    case 'e':
      epsilon = get_float(optarg);
      break;
    case 'E':
      mode_eval = 1;
      break;
    case 'f':
      strncpy(fname_base,optarg,MAX_FILENAME);
      break;
    case 'F':
      mode_norm = 1;
      break;
    case 'G':
      max_jsplit = get_int(optarg);
      break;
    case 'h':
      print_usage(stdout);
      exit(0);
    case 'i':
      mode_csv_in = 1;
      break;
    case 'j':
      mode_csv_out = 1;
      break;
    case 'k':
      get_num_class(optarg);  /* we need a special care */
      break;
    case 'l':
      max_init_iterate = get_int(optarg);
      break;
    case 'm':
      max_iterate = get_int(optarg);
      break;
    case 'M':
      mode_nomem = 1;
      break;
    case 'n':
      num_init = get_int(optarg);
      break;
    case 'N':
      min_smem_improve = get_float(optarg);
      break;
    case 'o':
      num_sm = get_int(optarg);
      break;
#ifdef USE_ICONV
    case 'p':
      strncpy(charcode_in,optarg,CHARCODE_LEN);
      break;
    case 'q':
      strncpy(charcode_out,optarg,CHARCODE_LEN);
      break;
#endif
    case 'Q':
      if (strcmp(optarg,"mp") == 0) {
	rel_m = REL_MP;
      } else if (strcmp(optarg,"wodd") == 0) {
	rel_m = REL_WODD;
      } else {
	fprintf(stderr,"Unexpected relevance measure: %s\n",optarg);
	print_usage(stderr);
	exit(1);
      }
      break;
    case 'r':
      rand_seed = get_uint(optarg);
      mode_set_seed = 1;
      break;
    case 'R':
      mode_em = 0;
      break;
    case 's':
      mode_suffix = 1;
      strncpy(fname_suffix,optarg,MAX_FILENAME);
      break;
    case 'S':
      mode_msg = 1;
      break;
    case 't':
      mode_cluster = 1;
      if (strcmp(optarg,"mp") == 0) {
	cl_type = CL_MP;
      } else if (strcmp(optarg,"wodd") == 0) {
	cl_type = CL_WODD;
      } else {
	print_usage(stderr);
	exit(1);
      }
      break;
    case 'U':
      min_param_c = get_float(optarg);
      break;
    case 'v':
      mode_verbose = 1;
      break;
    case 'V':
      mode_bdeu = 1;
      bdeu_min_smooth = get_float(optarg);
      break;
    case 'w':
      num_smooth_w = get_smooth_w(optarg);
      break;
    case 'x':
      tuple_size = get_int(optarg);
      break;
    case 'Y':
      mode_lex_order = 0;
      break;
    case 'z':
      mode_rank = 1;
      num_rank = get_int(optarg);
      break;
    case 'Z':
      mode_rank_sm = 1;
      num_rank_sm = get_int(optarg);
      break;
    default:
      fprintf(stderr,"Unknown option(s) specified\n");
      print_usage(stderr);
      exit(1);
    }
  }

  /* post-processing of options */
  if (fname_base[0] == '\0') {
    fprintf(stderr,"Missing or illegal -f option.\n");
    print_usage(stderr);
    exit(1);
  }

  if (tuple_size < 2) {
    fprintf(stderr,"Missing or illegal -x option.\n");
    print_usage(stderr);
    exit(1);
  }

  if (tuple_size != num_smooth_w) {
    if (num_smooth_w == 0) {
      smooth_w = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*tuple_size);
      if (smooth_w == NULL) {
	fprintf(stderr,"Couldn't allocate memory for smooth_w\n");
	exit(1);
      }
      mode_auto_smooth_w =
        (enum smooth_auto_type *)malloc(sizeof(enum smooth_auto_type)*tuple_size);
      if (mode_auto_smooth_w == NULL) {
	fprintf(stderr,"Couldn't allocate memory for mode_auto_smooth_w\n");
	exit(1);
      }
      for (v=0; v<tuple_size; v++) {
	smooth_w[v] = 0.0;
	mode_auto_smooth_w[v] = MODE_SMOOTH_NONE;
      }
    } else if (num_smooth_w == 1) {
      /* Reallocate smooth_w[] when uniform smoothing constants are specified: */
      tmp_smooth_w = smooth_w[0];
      tmp_mode_auto_smooth_w = mode_auto_smooth_w[0];
      free(smooth_w);
      free(mode_auto_smooth_w);
      smooth_w = (FLOAT_TYPE *)malloc(sizeof(FLOAT_TYPE)*tuple_size);
      if (smooth_w == NULL) {
	fprintf(stderr,"Couldn't allocate memory for smooth_w\n");
	exit(1);
      }
      mode_auto_smooth_w =
        (enum smooth_auto_type *)malloc(sizeof(enum smooth_auto_type)*tuple_size);
      if (mode_auto_smooth_w == NULL) {
	fprintf(stderr,"Couldn't allocate memory for mode_auto_smooth_w\n");
	exit(1);
      }
      for (v=0; v<tuple_size; v++) {
	smooth_w[v] = tmp_smooth_w;
	mode_auto_smooth_w[v] = tmp_mode_auto_smooth_w;
      }
    } else {
      fprintf(stderr,"Too many or too few smoothing constants for -w\n");
      print_usage(stderr);
      exit(1);
    }
  }

  if (mode_em && mode_search_num_class) {
    if (kmin <= 1 || kmax <= 0 || kstep <= 0 || kmin > kmax) {
      fprintf(stderr,"Illegal value for -k option.\n");
      print_usage(stderr);
      exit(1);
    }
    if (num_sm > 0 && kmin <= 2) {
      fprintf(stderr,"The number of classes should be larger than 2 for SMEM.\n");
      print_usage(stderr);
      exit(1);
    }
  } else if (mode_em && num_class <= 1) {
    fprintf(stderr,"Missing or illegal -k option.\n");
    print_usage(stderr);
    exit(1);
  } else if (mode_em && num_sm > 0 && num_class <= 2) {
    fprintf(stderr,"The number of classes should be larger than 2 for SMEM.\n");
    print_usage(stderr);
    exit(1);
  }
  
  /* In the later versions, this will be more complicated */
  mode_as_is = (mode_em) ? 1 : 0;

  if (max_iterate < 0 ||
      max_init_iterate < 0) {
    fprintf(stderr,"Illegal value for the number of iterations\n");
    exit(1);
  }

  if (max_iterate > MAX_MAX_ITERATE ||
      max_iterate > MAX_MAX_ITERATE) {
    fprintf(stderr,"Maximum number of iterations cannot exceed %d\n",
	    MAX_MAX_ITERATE);
    exit(1);
  }

  if (num_init <= 0) {
    fprintf(stderr,"Illegal value for the number of initial parameter sets\n");
    exit(1);
  }

  if (max_jsplit < 0) {
    fprintf(stderr,"Illegal value for the number of split candidate to be examined\n");
    exit(1);
  }

  if (num_rank < 0) {
    fprintf(stderr,"Illegal value for the number of ranked words\n");
    exit(1);
  }

  if (num_rank_sm < 0) {
    fprintf(stderr,"Illegal value for the number of ranked words in SMEM\n");
    exit(1);
  }

  if (min_param_c < 0) {
    fprintf(stderr,
	    "Illegal value for the lower limit of prob. "
	    "of the class to be split/merge\n");
    exit(1);
  }

  if (min_new_param_c < 0) {
    fprintf(stderr,
	    "Illegal value for the lower limit of updated prob. "
	    "of the splited/merged class\n");
    exit(1);
  }

  if (mode_nomem && num_sm > 0) {
    fprintf(stderr,"Saving memory mode is not available for SMEM algorithm\n");
    print_usage(stderr);
    exit(1);
  }

#ifdef USE_ICONV
  mode_conv_in = (strcmp(charcode_in,CHARCODE_INTERNAL) == 0) ? 0 : 1;
  mode_conv_out = (strcmp(charcode_out,CHARCODE_INTERNAL) == 0) ? 0 : 1;
#else
  mode_conv_in = 0;
  mode_conv_out = 0;
#endif

  if (mode_bdeu && bdeu_min_smooth <= 0) {
    fprintf(stderr,"Minimum pseudo count for BDeu should be positive\n");
    print_usage(stderr);
    exit(1);
  }

  strcpy(fname_freq,fname_base);
  strcpy(fname_param,fname_base);
  strcpy(fname_smooth,fname_base);
  strcpy(fname_dist,fname_base);
  strcpy(fname_memp,fname_base);
  strcpy(fname_log,fname_base);
  strcpy(fname_msg,fname_base);
  strcpy(fname_cluster,fname_base);
  strcpy(fname_rank,fname_base);
  strcpy(fname_rank_sm,fname_base);
  if (mode_suffix) {
    strcat(fname_param,"_");   strcat(fname_param, fname_suffix);
    strcat(fname_smooth,"_");  strcat(fname_smooth,fname_suffix);
    strcat(fname_dist, "_");   strcat(fname_dist,  fname_suffix);
    strcat(fname_memp, "_");   strcat(fname_memp,  fname_suffix);
    strcat(fname_log,  "_");   strcat(fname_log,   fname_suffix);
    strcat(fname_msg,  "_");   strcat(fname_msg,   fname_suffix);
    strcat(fname_cluster,"_"); strcat(fname_cluster, fname_suffix);
    strcat(fname_rank,"_");    strcat(fname_rank, fname_suffix);
    strcat(fname_rank_sm,"_"); strcat(fname_rank_sm, fname_suffix);
  }
  strcat(fname_freq,  ((mode_csv_in) ?  "_" FSUFFIX_FREQ ".csv"  : "_" FSUFFIX_FREQ ".txt"));
  strcat(fname_param, (((mode_em && mode_csv_out) || (!mode_em && mode_csv_in)) ?
		       "_" FSUFFIX_PARAM ".csv" : "_" FSUFFIX_PARAM ".txt"));
  strcat(fname_smooth, (((mode_em && mode_csv_out) || (!mode_em && mode_csv_in)) ?
		       "_" FSUFFIX_SMOOTH ".csv" : "_" FSUFFIX_SMOOTH ".txt"));
  strcat(fname_dist, ((mode_csv_out) ? "_" FSUFFIX_DIST_VB ".csv"  : "_" FSUFFIX_DIST_VB ".txt"));
  strcat(fname_memp, ((mode_csv_out) ? "_" FSUFFIX_MEMP_VB ".csv"  : "_" FSUFFIX_MEMP_VB ".txt"));
  strcat(fname_log,  ((mode_csv_out) ? "_" FSUFFIX_LOG_VB ".csv"   : "_" FSUFFIX_LOG_VB ".txt"));
  strcat(fname_msg,  ((mode_csv_out) ? "_" FSUFFIX_MSG_VB ".csv"   : "_" FSUFFIX_MSG_VB ".txt"));
  strcat(fname_cluster, ((mode_csv_out) ? "_" FSUFFIX_CLUSTER_VB ".csv" : "_" FSUFFIX_CLUSTER_VB ".txt"));
  strcat(fname_rank, ((mode_csv_out) ? "_" FSUFFIX_RANK_VB ".csv" : "_" FSUFFIX_RANK_VB ".txt"));
  strcat(fname_rank_sm, ((mode_csv_out) ? "_" FSUFFIX_SMEM_VB ".csv" : "_" FSUFFIX_SMEM_VB ".txt"));

  if (mode_cluster) {
    switch (cl_type) {
    case CL_MP:
      strncpy(cl_type_str,"MOST_LIKELY",ENUM_TYPE_LEN);
      break;
    case CL_WODD:
      strncpy(cl_type_str,"W_LOG_ODD_RATIO",ENUM_TYPE_LEN);
      break;
    default:
      fprintf(stderr,"Internal error: Illegal Info. Criterion\n");
      exit(1);
    }
  }

  switch (rel_m) {
  case REL_MP:
    strncpy(rel_m_str,"MEMBERSHIP_DIST",ENUM_TYPE_LEN);
    break;
  case REL_WODD:
    strncpy(rel_m_str,"W_LOG_ODD_RATIO",ENUM_TYPE_LEN);
    break;
  default:
    fprintf(stderr,"Internal error: Illegal relevance measure between class and word\n");
    exit(1);
  }

  if (min_smem_improve < 0) {
    min_smem_improve = epsilon * MIN_SMEM_IMPROVE_RATIO;
  }

}
